﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Configuration;
using System.Security.Principal;
using System.Text;
using System.Threading.Tasks;
using Antlr4.Runtime;
using Antlr4.Runtime.Tree;
using FourqlParser.Structures.Print;
using FourQLInterpreter.Exceptions;
using NLog;

namespace FourQLInterpreter.Application
{
    public class Application
    {
        private static readonly Logger Log = LogManager.GetCurrentClassLogger();

        public Interpreter InterpreterInstance { get; private set; }



        private static void Main(string[] args)
        {
            //Console.WindowHeight = (Console.LargestWindowHeight - 20);
           // Console.WindowWidth = (Console.LargestWindowWidth - 20);
            Application application = new Application();
            application.WelcomeMessage();
            application.ShowHelp();
            while (true)
            {
                Prompt();
                
                try
                {
                    string line = Console.ReadLine();
                    application.ParseInput(line);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e.Message); 
                    Log.Error(e.StackTrace);
                }
            }
        }

        private void WelcomeMessage()
        {
            Console.WriteLine("4QL Interpreter:");
        }

        private static void Prompt()
        {
            Console.Write(">> ");
        }

        public Application()
        {
            InterpreterInstance = new Interpreter();
            InterpreterInstance.Initialize();
        }

        public void ParseInput(string userInput)
        {
            userInput = userInput.Replace(';', ' ');
            var split = userInput.Split(' ');
            switch (split[0])
            {
                case "import":
                case "load":
                    if (split.Length < 2)
                        Console.WriteLine("Please input path to module");
                    else
                        LoadModules(split[1]);
                    break;
                case "exit":
                    Exit();
                    break;
                case "help":
                    ShowHelp();
                    break;
                case "modules":
                    PrintInfo();
                    break;
                case "print":
                    if (split.Length < 2)
                        Console.WriteLine("Please input module name");
                    else
                        PrintModuleInfo(split[1]);
                    break;
                case "clear":
                    ClearAllModules();
                    break;
                default:
                    ExecuteUserQuery(userInput);
                    break;
            }
        }

        private void Exit()
        {
            Environment.Exit(0);
        }

        private void PrintModuleInfo(String moduleName)
        {
            var module = InterpreterInstance.GetModule(moduleName);
            var printer = new ProgramPrinter();
            Console.WriteLine(printer.append(module, "  ").Done());
        }

        private void ClearAllModules()
        {
            InterpreterInstance.Clear();
        }
        private void ShowHelp()
        {
            Console.WriteLine("Available instructions:");
            Console.WriteLine();
            Console.WriteLine("load path_to_file;                      => Loads module definitions from file.");
            Console.WriteLine("module.relation(X, Y) and...            => Executes a query in 4QL language.");
            Console.WriteLine("modules;                                => Lists all modules.");
            Console.WriteLine("print module_name;                      => Prints the module.");
            Console.WriteLine("clear;                                  => Unloads all modules.");
            Console.WriteLine("exit;                                   => Exits the interpreter");
            Console.WriteLine();
        }
        private void PrintInfo()
        {
            foreach (var module in InterpreterInstance.Modules.Values)
            {
                if (!module.Name.StartsWith("_"))
                {
                    Console.Write(" " + module.Name);
                    if (module.IsInternalExternal)
                        Console.Write(" *");
                    Console.WriteLine();
                }
            }
            Console.WriteLine("* - internal/plugin module.");
        }
        private void LoadModules(string path)
        {
            try
            {
                InterpreterInstance.LoadModules(path);

                Console.WriteLine("Program loaded! Execute the 'modules' command to see the list of all available modules.");
            }
            catch (Exception e)
            {
                Console.WriteLine("Failed to load " + path + " : " + ExceptionUtils.GetMessagesWithInners(e));

                Log.Error(e.StackTrace);
            }
        }

        private void ExecuteUserQuery(string query)
        {
            try
            {
                var result = InterpreterInstance.Query(query);
                TablePrettyPrinter.PrintResult(result);
            }
            catch (ParsingException e)
            {
                Console.WriteLine("'" + query + "' is not a valid query: " + ExceptionUtils.GetMessagesWithInners(e));
                Log.Error(e.StackTrace);
            }
            catch (Exception e)
            {
                Console.WriteLine("'" + query + "' is not a valid query: " + ExceptionUtils.GetMessagesWithInners(e));
                Log.Error(e.StackTrace);
            }
        }

      }
}
