﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;
using FourQLInterpreter.Exceptions;
using FourQLInterpreter.Reasoning;
using FourQLInterpreter.WSM;
using NLog;

namespace FourQLInterpreter.Application
{
    public class Interpreter
    {

        private static readonly Logger Log = LogManager.GetCurrentClassLogger();

        public IDictionary<string, Module> Modules { get; private set; }

        public IDictionary<string, WsmEvaluator> WsmEvaluators { get; private set; }

        public Evaluator ExpressionEvaluator { get; private set; }

        public Interpreter()
        {
            Modules = new Dictionary<string, Module>();
            WsmEvaluators = new Dictionary<string, WsmEvaluator>();
            ExpressionEvaluator = InterpreterUtils.CreateEvaluator(Modules.Values);
        }

        public void Initialize()
        {
            var internalModules = InterpreterUtils.LoadInternalModules();
            Modules = internalModules.ToDictionary(v => v.Name);
            this.ExpressionEvaluator = InterpreterUtils.CreateEvaluator(Modules.Values);
        }

        /// <summary>
        /// Loads modules from specified file
        /// </summary>
        /// <param name="path">path to the .4ql file</param>
        public void LoadModules(string path)
        {
            Log.Trace("Loading " + path);
            var existingModules = this.Modules;
            this.Modules = InterpreterUtils.LoadModules(path, existingModules.Values);
            this.ExpressionEvaluator = InterpreterUtils.CreateEvaluator(Modules.Values);
            WsmEvaluators = new Dictionary<string, WsmEvaluator>();
            InitializeModules();
        }

        /// <summary>
        /// Build the WSM of each module in the correct order
        /// </summary>
        private void InitializeModules()
        {
            // construct ModuleHierarchy to find the order of WSM evaluation
            var moduleHierarchy = new ModuleHierarchy(Modules.Values).ConstructHierarchy();
            foreach (var layer in moduleHierarchy)
            {
                foreach (var module in layer.Value)
                {
                    ComputeWSM(module.Module);
                }
            }
        }

        /// <summary>
        /// Build the  WSM  of the given module
        /// </summary>
        /// <param name="module"></param>
        private void ComputeWSM(Module module)
        {
            
            var wsmEval = new OptimizedWSMEvaluator(module, ExpressionEvaluator, true);
            WsmEvaluators[module.Name] = wsmEval;
            if (!module.IsInternalExternal) // internal modules dont have a model
            {
                wsmEval.ConstructWsm();
            }
        }
        /// <summary>
        /// Return the result of evaluated expression
        /// </summary>
        /// <param name="userQuery"></param>
        /// <returns></returns>
        public Result Query(string userQuery)
        {

            Expression expr = InterpreterUtils.ParseExpression(userQuery, Modules);
            return ExpressionEvaluator.Evaluate(expr);
        }

        /// <summary>
        /// Evaluate the expression, but in the result include only given variable names
        /// </summary>
        /// <param name="userQuery"></param>
        /// <param name="variableNames"></param>
        /// <returns></returns>
        public Result Query(string userQuery, ICollection<string> variableNames)
        {
            var joinedNames = Modules.Keys.Aggregate((a, b) => a + ", " + b);
            Log.Trace(userQuery + " " + joinedNames);
            Expression expr = InterpreterUtils.ParseExpression(userQuery, Modules);
            return ExpressionEvaluator.Evaluate(expr, variableNames);
        }

        public EvaluationHistory GetWSMHistory(string moduleName)
        {
            return WsmEvaluators[moduleName].History;
        }

        /// <summary>
        /// Clear all the modules in the memory
        /// </summary>
        public void Clear()
        {
            this.Modules.Clear();
            this.ExpressionEvaluator.ModulesData.Clear();
            this.WsmEvaluators.Clear();
            Initialize();
        }


        internal Module GetModule(string moduleName)
        {
            if (!Modules.ContainsKey(moduleName))
                throw new FourqlException("Module " + moduleName + " does not exist!");
            else return Modules[moduleName];
        }
    }
}
