﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Parsers;
using FourqlParser.Structures;
using FourQLInterpreter.Exceptions;
using FourQLInterpreter.InternalModules;
using FourQLInterpreter.Reasoning;
using Module = FourqlParser.Structures.Module;

namespace FourQLInterpreter.Application
{
    public class InterpreterUtils
    {
        public static string GetFileContents(string path)
        {
            path = path.Replace("\"", "");
            return ParserDispatcher.AsFile(path);
        }


        /// <summary>
        /// Loads modules from file.
        /// </summary>
        /// <param name="fileName"></param>
        /// <returns></returns>
        public static IDictionary<string, Module> LoadModules(string fileName)
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetFileContents(fileName), errors, null);

            var prog = ParserDispatcher.toProgram(progContext);
            
            if (errors.ParserErrors.Count != 0)
                throw new ParsingException(errors.ParserErrors);

            ParserDispatcher.checkProgram(prog);

            return prog.Modules.ToDictionary(module => module.Name);
        }

        public static IDictionary<string, Module> LoadModules(string fileName, ICollection<Module> extraModules)
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetFileContents(fileName), errors, null);

            var prog = ParserDispatcher.toProgram(progContext);

            foreach (var module in extraModules)
            {
                prog.Modules.Add(module);
            }

            if (errors.ParserErrors.Count != 0)
                throw new ParsingException(errors.ParserErrors);
            ParserDispatcher.checkProgram(prog);

            return prog.Modules.ToDictionary(module => module.Name);
        }

        /// <summary>
        /// Loads all the internal modules.
        /// Loads all the plugins from the plugin directory and creates wrappers for them
        /// </summary>
        /// <returns>Collection of newly created Module objects</returns>
        public static ICollection<Module> LoadInternalModules()
        {
            var modules = new List<Module>();
            modules.Add(new MathModule());
            modules.Add(new QueryEntryModule());
            foreach (var path in GetPluginFiles())
                modules.Add(LoadPlugin(path));
            return modules;
        }

        private static Module LoadPlugin(string path)
        {
            PluginModuleWrapper p = new PluginModuleWrapper(path);
            p.Initialize();
            return p;
        }


        private static string[] GetPluginFiles()
        {
            string path = Path.Combine(AppDomain.CurrentDomain.BaseDirectory, "Plugins");
            try
            {
                return Directory.GetFiles(path);
            }
            catch (DirectoryNotFoundException e)
            {
                Console.WriteLine("'Plugin' folder not found at '" + path + ". No external modules will be loaded.");
                return  new string[]{};
            }
        }

        public static Evaluator CreateEvaluator(ICollection<Module> moduleCollection)
        {
            IDictionary<string, ModuleDataWrapper> modules = new Dictionary<string, ModuleDataWrapper>();
            foreach (var module in moduleCollection)
            {
                var wrapper = ModuleDataWrapperFactory.CreateWrapper(module);
                wrapper.Initialize();
                modules[module.Name] = wrapper;
            }
            var evaluator = new Evaluator(modules);
            return evaluator;
        }

        public static Expression ParseExpression(string s, IDictionary<string, Module> modules, string moduleName)
        {
            var errors = new ParserDispatcher.Errors();
            var context = ParserDispatcher.parseExpr(s, errors);
            if (errors.ParserErrors.Count != 0)
                throw new ParsingException(errors.ParserErrors);

            var expr = ParserDispatcher.toExpr(context);

            try
            {
                ParserDispatcher.checkExpr(expr, modules.Values, modules[moduleName]);
            }
            catch (Exception e)
            {
                throw new ParsingException( e.Message, e);
            }
            return expr;
        }

        public static Expression ParseExpression(string s, IDictionary<string, Module> modules)
        {
            return ParseExpression(s, modules, "_queryEntryPoint");
        }
    }

 

    public class EntryModuleFinder : FourqlParser.Structures.Expression.Visitor<string>
    {

        public string visitLiteral(Literal literal)
        {
            return null;
        }

        public string visitNeg(NegExpression negExpr)
        {
            return null;
        }

        public string visitModule(ModuleExpression moduleExpr)
        {
            return moduleExpr.ModuleName;
        }

        public string visitAnd(AndExpression andExpr)
        {
            return null;
        }

        public string visitOr(OrExpression orExpr)
        {
            return null;
        }

        public string visitExists(ExistsExpression exExpr)
        {
            return null;
        }

        public string visitAll(AllExpression allExpr)
        {
            return null;
        }
    }
}

