﻿using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;
using FourqlParser.StaticAnalysis;
using FourQLInterpreter.Application;
using FourQLInterpreter.Reasoning.Explanation;

namespace FourQLInterpreter.Reasoning
{
    /// <summary>
    /// Evaluates the expression using the data from all the modules.
    /// </summary>
    public class Evaluator
    {
        /// <summary>
        /// Context for the expression
        /// </summary>
        private Module Module { get; set; }

        /// <summary>
        /// Data of all the modules
        /// </summary>
        public IDictionary<string, ModuleDataWrapper> ModulesData { get; set; }

        public bool Explain = false;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="modules">Knowledge base </param>
        public Evaluator(IDictionary<string, ModuleDataWrapper> modules)
        {
            ModulesData = modules;
        }


        public Result Evaluate(Expression expression)
        {
            return Evaluate(expression, ModulesData["_queryEntryPoint"].Module);
        }

        public Result Evaluate(Expression expression, ICollection<string> targetVariableNames)
        {
            return Evaluate(expression, ModulesData["_queryEntryPoint"].Module, targetVariableNames, LogicValue.FALSE);
        }

        public Result Evaluate(Expression expression, Module module, ICollection<string> targetVariableNames)
        {
            return Evaluate(expression, module, targetVariableNames, LogicValue.FALSE);
        }

        public Result Evaluate(Expression expression, Module module)
        {
            return Evaluate(expression, module, expression.Accept(new VariableGatherer(module, ModulesData)).Keys);
        }
        /// <summary>
        /// Evaluates an expression
        /// </summary>
        /// <param name="expression">Expression to be evaluated</param>
        /// <param name="module">Context of the expression</param>
        /// <param name="targetVariableNames">variable names which will be included in the result</param>
        /// <param name="minValue">lowest logical value we are interested in</param>
        /// <param name="extraVariableRefs">used by the WSM evaluator to pass the refs to variables from rule head -> used for limiting the domain values to those present in the module</param>
        /// <returns>result of the evaluation, consisting of variable values and Logic value to which the expression evaluates</returns>
        public Result Evaluate(Expression expression, Module module, ICollection<string> targetVariableNames, LogicValue minValue, IDictionary<string, ICollection<VariableRef>> extraVariableRefs=null)
        {
            Module = module;

            // gather all the variables from the expression
            var variables = expression.Accept(new VariableGatherer(Module, ModulesData, extraVariableRefs));
            // gather possible values for each variable
            var domainValues = GatherDomainValues(variables, ModulesData);

            // split the values into 'target' variables and 'free' variables - those not important to the caller
            var targetVariables = domainValues.Where(v => targetVariableNames.Contains(v.Key)).ToDictionary(v => v.Key, v => v.Value);
            var freeVariables = domainValues.Where(v => !targetVariableNames.Contains(v.Key)).ToDictionary(v => v.Key, v => v.Value);

            // create the cartesian product of all variable values, thus generating all possible combinations
            var targetVariableTable = CartesianProduct(targetVariables.Values);
            var freeVariableTable = CartesianProduct(freeVariables.Values);

            // evaluate the expression with each possible variable values
            var evaluation = from tRow in targetVariableTable
                             select EvaluateExpression(expression, targetVariables.Keys, tRow, freeVariables.Keys, freeVariableTable, minValue);

            // construct the result 
            return ConstructResult(evaluation, ConstructHeaders(variables, targetVariables.Keys, freeVariables.Keys), expression, minValue);
        }

        private List<ColumnDescription> ConstructHeaders(IDictionary<string, ICollection<VariableRef>> variables, IEnumerable<string> targetVariableNames, IEnumerable<string> freeVariableNames)
        {

            return targetVariableNames
                .Select(v => new ColumnDescription() {Name = v, Type = variables[v].First().DomainType})
                .Concat(
                    freeVariableNames.Select(
                        v => new ColumnDescription() {Name = v, Type = variables[v].First().DomainType}))
                .ToList();
        }


        private Result ConstructResult(IEnumerable<Row> evaluation, List<ColumnDescription> headers, Expression expression, LogicValue minValue)
        {
            var result = new Result() { Module = Module, Rows = evaluation.Where(e => e.Value >= minValue), Expression = expression, ColumnHeaders = headers };
            return result;
        }

        /// <summary>
        /// Evaluate the expression to create a row of the result
        /// </summary>
        /// <param name="expression">Expression to be evaluated</param>
        /// <param name="targetVariableNames">variable names to be included in the result</param>
        /// <param name="targetVariableValues"></param>
        /// <param name="freeVariableNames">free variables, only their 'best' values will be included in the result</param>
        /// <param name="freeVariableTable"></param>
        /// <param name="minValue">lowest 'important' value</param>
        /// <returns></returns>
        private Row EvaluateExpression(Expression expression, IEnumerable<string> targetVariableNames, IEnumerable<ValueHolder> targetVariableValues, IEnumerable<string> freeVariableNames, IEnumerable<IEnumerable<ValueHolder>> freeVariableTable, LogicValue minValue)
        {
            var result = LogicPrim.Min();

            IEnumerable<ValueHolder> resultFreeVarValues = freeVariableTable.FirstOrDefault() ?? new List<ValueHolder>();
            // names of all the variables
            var variableNames = targetVariableNames.Concat(freeVariableNames);
            
            Explanation.Explanation explanation = null;
            // Those values for the free variables that give the best result (best according to the order u < f < t < i )
            // will be included in the result row
            foreach (var row in freeVariableTable)
            {
                LogicValue expressionValue = LogicPrim.Min();
                if (Explain)
                { // evaluate the expression with explanation
                    var valueWithExplanation =
                        EvaluateExpressionWithExplanation(expression, variableNames, targetVariableValues.Concat(row),
                            minValue);
                    expressionValue = valueWithExplanation.Item1;
                    explanation = valueWithExplanation.Item2;
                }
                else
                {   // evaluate the expression without the explanation
                    expressionValue = EvaluateExpression(expression, variableNames, targetVariableValues.Concat(row), minValue);
                    explanation = null;
                }
                // check if the result is better than the previous one (w.r.t u < f < t < i)
                if (expressionValue.Equals(LogicPrim.OrPrim(result, expressionValue)))
                { 
                    resultFreeVarValues = row;
                    result = expressionValue;
                }
                if (result.Equals(LogicPrim.Max()))
                    break;
            }
            return new Row() { Value = result, Values = targetVariableValues.Concat(resultFreeVarValues).ToList(), RowExplanation = explanation};
        }

        

        private LogicValue EvaluateExpression(Expression expression, IEnumerable<string> variableNames, IEnumerable<ValueHolder> variableValues, LogicValue minValue)
        {
     
            var result = expression.Accept(new ExpressionEvaluator(this.ModulesData[Module.Name], ModulesData, variableNames.Zip(variableValues, (name, value) => new { name, value }).ToDictionary(x => x.name, x => x.value)){MinValue = minValue});

            return result;
        }
        /// <summary>
        /// Evaluate the expression using the explaining evaluator
        /// </summary>
        /// <param name="expression"></param>
        /// <param name="variableNames"></param>
        /// <param name="variableValues"></param>
        /// <param name="minValue"></param>
        /// <returns></returns>
        private Tuple<LogicValue, Explanation.Explanation> EvaluateExpressionWithExplanation(Expression expression, IEnumerable<string> variableNames, IEnumerable<ValueHolder> variableValues, LogicValue minValue)
        {
            var expressionEval = new ExplainExprEvaluator(this.ModulesData[Module.Name], ModulesData, variableNames.Zip(variableValues, (name, value) => new { name, value }).ToDictionary(x => x.name, x => x.value));
            var result = expression.Accept(expressionEval);

            return new Tuple<LogicValue, Explanation.Explanation>(result, expressionEval.GetExplanation(expression));
        } 
        


        public static IDictionary<string, IEnumerable<ValueHolder>> GatherDomainValues(IDictionary<string, ICollection<VariableRef>> variables, IDictionary<string, ModuleDataWrapper> modulesData)
        {
            IDictionary<string, IEnumerable<ValueHolder>> result = new Dictionary<string, IEnumerable<ValueHolder>>();
            foreach (var variable in variables)
            {
                //intersect values from all domains concerning the variable
                var domainIntersection = variable.Value
                    .Where(r => !r.Module.IsInternalExternal) // ignore references to internal modules
                    .Select(r => modulesData[r.Module.Name].GetVariableDomainValues(r)) // get all domain values for this variable
                    .Aggregate((x, y) => x.Intersect(y)) // aggregate intersect the values
                    .Distinct(); // remove duplicates
                result[variable.Key] = domainIntersection;
            }
            return result;
        }
        /// <summary>
        /// http://blogs.msdn.com/b/ericlippert/archive/2010/06/28/computing-a-cartesian-product-with-linq.aspx
        /// </summary>
        /// <param name="sequences"></param>
        /// <returns></returns>
        IEnumerable<IEnumerable<ValueHolder>> CartesianProduct(IEnumerable<IEnumerable<ValueHolder>> sequences)
        {
            IEnumerable<IEnumerable<ValueHolder>> emptyProduct = new[] { Enumerable.Empty<ValueHolder>() };
            return sequences.Aggregate(
              emptyProduct,
              (accumulator, sequence) =>
                from accseq in accumulator
                from item in sequence
                select accseq.Concat(new[] { item }));
        }

        /// <summary>
        /// Gathers vvariable references from the expression
        /// </summary>
        public class VariableGatherer : Expression.Visitor<IDictionary<string, ICollection<VariableRef>>>
        {
            private Module _currentModule = null;

            private IDictionary<string, ICollection<VariableRef>> _variables = new Dictionary<string, ICollection<VariableRef>> ();

            private IDictionary<string, ModuleDataWrapper> _modules;

            public VariableGatherer(Module module, IDictionary<string, ModuleDataWrapper> modules, IDictionary<string, ICollection<VariableRef>> extraVariableRefs=null)
            {
                _currentModule = module;
                _modules = modules;
                if (extraVariableRefs != null)
                    _variables = extraVariableRefs;
            }

            public IDictionary<string, ICollection<VariableRef>> visitLiteral(Literal literal)
            {
                FourqlParser.StaticAnalysis.ExprVariableGatherer.findVariablesOfLiteral(_currentModule, literal, _variables);
                return _variables;
            }

            public IDictionary<string, ICollection<VariableRef>> visitNeg(NegExpression negExpr)
            {
                negExpr.Next.Accept(this);
                return _variables;
            }

            public IDictionary<string, ICollection<VariableRef>> visitModule(ModuleExpression moduleExpr)
            {
                var prevModule = _currentModule;
                _currentModule = _modules[moduleExpr.ModuleName].Module;
                moduleExpr.Next.Accept(this);
                _currentModule = prevModule;
                return _variables;
            }

            public IDictionary<string, ICollection<VariableRef>> visitAnd(AndExpression andExpr)
            {
                foreach (var next in andExpr.Nexts) {
                    next.Accept(this);
                }
                return _variables;
            }

            public IDictionary<string, ICollection<VariableRef>> visitOr(OrExpression orExpr)
            {
                foreach (var next in orExpr.Nexts) {
                    next.Accept(this);
                }
                return _variables;
            }

            public IDictionary<string, ICollection<VariableRef>> visitQuantifier(QantifierExpression expr)
            {
                var variables = _variables;
                _variables = new Dictionary<string, ICollection<VariableRef>>();
                expr.Next.Accept(this);
                // we don't want to include the quantified variable as it should be 'invisible' outside the quantified expression
                if (_variables.ContainsKey(expr.VariableName))
                    _variables.Remove(expr.VariableName);

                foreach (var var in _variables)
                {
                    if (!variables.ContainsKey(var.Key))
                    {
                        variables[var.Key] = var.Value;
                    }
                    else
                    {
                        variables[var.Key] = variables[var.Key].Concat(var.Value).ToList();
                    }
                }
                _variables = variables;

                return _variables;
            } 

            public IDictionary<string, ICollection<VariableRef>> visitExists(ExistsExpression exExpr)
            {
                return visitQuantifier(exExpr);
            }

            public IDictionary<string, ICollection<VariableRef>> visitAll(AllExpression allExpr)
            {
                return visitQuantifier(allExpr);
            }
        }
    }
}
