﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;

namespace FourQLInterpreter.Reasoning.Explanation
{
    public class ExplainExprEvaluator : ExpressionEvaluator
    {
        private Explanation ResultExplanation;

        private bool finished = false;

        private IDictionary<QantifierExpression, ValueHolder> quantifierVariables;

        private IDictionary<QantifierExpression, LogicValue?> quantifierValues; 
 

        public ExplainExprEvaluator(ModuleDataWrapper startingModule, IDictionary<string, ModuleDataWrapper> modules, IDictionary<string, ValueHolder> variableValues) : base(startingModule, modules, variableValues)
        {
            ResultExplanation = new Explanation();
            quantifierValues = new Dictionary<QantifierExpression, LogicValue?>();
            quantifierVariables = new Dictionary<QantifierExpression, ValueHolder>();
            this.Optimize = false;
        }

        public Explanation GetExplanation(Expression expr)
        {
            if (!finished) {
                FillExplanation(expr);
            }
            return ResultExplanation;
        }

        private Explanation FillExplanation(Expression expr)
        {
            expr.Accept(new ExplanationLiteralValueSetter(_currentModule, ModulesData, VariableValues.GetFirst(), ResultExplanation, quantifierValues, quantifierVariables));
            finished = true;
            return ResultExplanation;
        }

        public override LogicValue visitExists(ExistsExpression exExpr)
        {

            var domainValues = GetVariableValuesForQuantifier(exExpr);

            // default result for exist
            var result = LogicValue.FALSE;
            ValueHolder bestVal = null;
            var quantifierValuesCopy = new Dictionary<QantifierExpression, LogicValue?>(quantifierValues);
            var quantifierVarsCopy = new Dictionary<QantifierExpression, ValueHolder>(quantifierVariables);
            var prevResult = result;
            VariableValues.PushNewEnv();
            foreach (var value in domainValues)
            {
                VariableValues.Set(exExpr.VariableName, value);
                result = Logic.Or(result, exExpr.Next.Accept(this));
                if (result >= prevResult)
                {
                    quantifierValuesCopy = new Dictionary<QantifierExpression, LogicValue?>(quantifierValues);
                    quantifierVarsCopy = new Dictionary<QantifierExpression, ValueHolder>(quantifierVariables);
                    bestVal = value;
                    prevResult = result;
                }
                if (result.Equals(LogicValue.TRUE))
                    break;
            }

            quantifierValues = quantifierValuesCopy;
            quantifierVariables = quantifierVarsCopy;
            quantifierValues[exExpr] = result;
            quantifierVariables[exExpr] = bestVal;

            VariableValues.PopEnv();
            
            return result;
        }

        public override LogicValue visitAll(AllExpression allExpr)
        {
            var domainValues = GetVariableValuesForQuantifier(allExpr);
            var quantifierIndex = ResultExplanation.QuantifierValues.Count;

            VariableValues.PushNewEnv();

            // default result for all
            var result = LogicValue.TRUE;
            ValueHolder bestVal = null;
            var quantifierValuesCopy = new Dictionary<QantifierExpression, LogicValue?>(quantifierValues);
            var quantifierVarsCopy = new Dictionary<QantifierExpression, ValueHolder>(quantifierVariables);
            var prevResult = result;
            foreach (var value in domainValues)
            {
                VariableValues.Set(allExpr.VariableName, value);
                result = Logic.And(result, allExpr.Next.Accept(this));
                if (result <= prevResult)
                {
                    quantifierValuesCopy = new Dictionary<QantifierExpression, LogicValue?>(quantifierValues);
                    quantifierVarsCopy = new Dictionary<QantifierExpression, ValueHolder>(quantifierVariables);
                    bestVal = value;
                    prevResult = result;
                }
                if (result.Equals(LogicValue.FALSE))
                    break;
            }
            quantifierValues = quantifierValuesCopy;
            quantifierVariables = quantifierVarsCopy;
            quantifierValues[allExpr] = result;
            quantifierVariables[allExpr] = bestVal;

            VariableValues.PopEnv();


            return result;
        }
    }

    class ExplanationLiteralValueSetter : ExpressionEvaluator
    {
        private Explanation _explanation;
        private readonly IDictionary<QantifierExpression, LogicValue?> _quantifierValues;
        private readonly IDictionary<QantifierExpression, ValueHolder> _quantifierVariables;

        private IList<string> nullVariables = new List<string>(); 


        public ExplanationLiteralValueSetter(ModuleDataWrapper startingModule, IDictionary<string, ModuleDataWrapper> modulesData, IDictionary<string, ValueHolder> variableValues, Explanation resultExplanation, IDictionary<QantifierExpression, LogicValue?> quantifierValues, IDictionary<QantifierExpression, ValueHolder> quantifierVariables) : base(startingModule, modulesData, variableValues)
        {
            _explanation = resultExplanation;
            _quantifierValues = quantifierValues;
            _quantifierVariables = quantifierVariables;
            Optimize = false;
        }

        public override LogicValue visitLiteral(Literal literal)
        {
            var hasNullVariables = (from v in literal.Values
                where v.Type.Equals(DomainType.VARIABLE) && nullVariables.Contains(v.valueAs<string>())
                select true).FirstOrDefault();

            if (hasNullVariables)
            {
                _explanation.LiteralValues.Add(null);
                return LogicValue.UNKNOWN;
            }
            else
            {
                var result = base.visitLiteral(literal);
                _explanation.LiteralValues.Add(result);
                return result;
            }
        }

        private LogicValue visitQuantifier(QantifierExpression expr)
        {
            ValueHolder varValue = null;
            LogicValue? value = null;
            if (_quantifierVariables.ContainsKey(expr))
            {
                // get the variable  value selected for the quantifier
                varValue = _quantifierVariables[expr];
            }
            if (_quantifierValues.ContainsKey(expr))
            {
                value = _quantifierValues[expr];
            }
            // add values to explanation structure
            _explanation.QuantifierValues.Add(value);
            _explanation.QuantifierVariables.Add(varValue);

            // if it's null, add its name to the nullVariables list
            if (varValue == null)
                nullVariables.Add(expr.VariableName);


            // add chosen value to the envidronment
            VariableValues.PushNewEnv();
            VariableValues.Set(expr.VariableName, varValue);

            // proceed to the subtree
            var result = expr.Next.Accept(this);

            // restore previous environment
            VariableValues.PopEnv();

            // remove the variable name from the nullVariables list
            if (varValue == null)
                nullVariables.RemoveAt(nullVariables.Count-1);
            return result;
        }

        public override LogicValue visitExists(ExistsExpression exExpr)
        {
            return visitQuantifier(exExpr);
        }

        public override LogicValue visitAll(AllExpression allExpr)
        {
            return visitQuantifier(allExpr);
        }
    }

}
