﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.StaticAnalysis;
using FourqlParser.Structures;

namespace FourQLInterpreter.Reasoning
{
    /// <summary>
    /// Evaluates the expression and returns the LogicValue
    /// </summary>
    public class ExpressionEvaluator : Expression.Visitor<LogicValue>
    {
        protected ModuleDataWrapper _currentModule;

        public IDictionary<string, ModuleDataWrapper> ModulesData { get; private set; }

        public Environment<string, ValueHolder> VariableValues { get; protected set; }

        // minimal value user is interested in. used for optimizing the queries.
        // NOTE: expressions which evaluate to < MinValue can return wrong result! (although still < MinValue)
        public LogicValue MinValue { get; set; }

        /// <summary>
        /// If true, the evaluator will optimize the process, breaking the evaluations early if possible, thus not visiting the entire tree.
        /// <br/>
        /// When set to false, all nodes in the expression tree will be visited
        /// </summary>
        protected bool Optimize = true;



        public ExpressionEvaluator(ModuleDataWrapper startingModule, IDictionary<string, ModuleDataWrapper> modules, IDictionary<string, ValueHolder> variableValues)
        {
            this._currentModule = startingModule;
            this.ModulesData = modules;
            this.VariableValues = new Environment<string, ValueHolder>(variableValues);
            MinValue = LogicValue.FALSE;
        }


        public virtual LogicValue visitLiteral(Literal literal)
        {
            var valueList = SubstituteVariables(literal, VariableValues);
            var result = _currentModule.GetLiteraValue(literal.RelationName, valueList);
            return result;       
        }

        public static IList<ValueHolder> SubstituteVariables(Literal literal, Environment<string, ValueHolder> variableValues)
        {
            IList<ValueHolder> result = new List<ValueHolder>();
            foreach (var value in literal.Values)
            {
                if (value.Type.Equals(DomainType.VARIABLE))
                    result.Add(variableValues.Get(value.valueAs<string>()));
                else 
                    result.Add(value);
            }
            return result;
        }

        public LogicValue visitNeg(NegExpression negExpr)
        {
            return Logic.Not(negExpr.Next.Accept(this));
        }

        public LogicValue visitModule(ModuleExpression moduleExpr)
        {
            var prevModule = _currentModule;
            _currentModule = ModulesData[moduleExpr.ModuleName];
            var result = moduleExpr.Next.Accept(this);
            _currentModule = prevModule;
            //TODO z czym not wiaze?

            result = ResultOfIn(result, moduleExpr.InValues);

            if (moduleExpr.IsNegative)
                return Logic.Not(result);
            else
                return result;
        }

        public static LogicValue ResultOfIn(LogicValue result, ICollection<LogicValue> collection)
        {
            if (collection.Count <= 0)
                return result;
            if (collection.Contains(result))
                return LogicValue.TRUE;
            else
                return LogicValue.FALSE;
        }

        public LogicValue visitAnd(AndExpression andExpr)
        {
            var result = LogicValue.TRUE;
            foreach (var next in andExpr.Nexts)
            {
                result = Logic.And(result, next.Accept(this));
                if (Optimize && (result < MinValue))
                    break;
            }
            return result;
        }

        public LogicValue visitOr(OrExpression orExpr)
        {
            var result = LogicValue.FALSE;
            foreach (var next in orExpr.Nexts)
            {
                result = Logic.Or(result, next.Accept(this));
                if (Optimize && (result.Equals(LogicValue.TRUE)))
                    break;
            }
            return result;
        }

        public virtual LogicValue visitExists(ExistsExpression exExpr)
        {
            var domainValues = GetVariableValuesForQuantifier(exExpr);
            var result = LogicValue.FALSE;
            VariableValues.PushNewEnv();
            foreach (var value in domainValues)
            {
                VariableValues.Set(exExpr.VariableName, value);
                result = Logic.Or(result, exExpr.Next.Accept(this));
                if (result.Equals(LogicValue.TRUE))
                    break;
            }

            VariableValues.PopEnv();

            return result;

        }

        public virtual LogicValue visitAll(AllExpression allExpr)
        {
            var domainValues = GetVariableValuesForQuantifier(allExpr);
            var result = LogicValue.TRUE;

            VariableValues.PushNewEnv();
            foreach (var value in domainValues)
            {
                VariableValues.Set(allExpr.VariableName, value);
                result = Logic.And(result, allExpr.Next.Accept(this));
                if (result < MinValue)
                    break;
            }
            VariableValues.PopEnv();
            return result;
        }

        protected IEnumerable<ValueHolder> GetVariableValuesForQuantifier(QantifierExpression expr)
        {
            expr.DomainModule = expr.DomainModule ?? _currentModule.Module.Name;
            var mod = ModulesData[expr.DomainModule].Module;
            DomainType t;
            if (DomainType.TryParse(expr.VariableDomain, true, out t))
                ;
            else
            {
                t = mod.FindDomain(expr.VariableDomain).Type;
            }
            var qRef = new VariableRef()
            {
                DomainName = expr.VariableDomain,
                DomainType = t,
                Literal = null,
                Module = mod,
                ParamNumber = 0,
                VariableName = expr.VariableName
            };
            var dict = new Dictionary<string, ICollection<VariableRef>>() {{expr.VariableName, new List<VariableRef>(){qRef}}};
            Evaluator.VariableGatherer variableGatherer = new Evaluator.VariableGatherer(this._currentModule.Module,
                ModulesData, dict);
            var refs = expr.Next.Accept(variableGatherer);
            var relevantRefs = refs[expr.VariableName];
            var vals = Evaluator.GatherDomainValues(new Dictionary<string, ICollection<VariableRef>>(){ {expr.VariableName, relevantRefs}}, ModulesData);
            return vals[expr.VariableName];
        } 

    }

    public class Environment<TKey, TValue>
    {
        private LinkedList<IDictionary<TKey, TValue>> stack = new LinkedList<IDictionary<TKey, TValue>>();

        public Environment()
        {
            
        } 
        public Environment(IDictionary<TKey, TValue> initialData)
        {
            stack.AddLast(initialData);
        } 

        public void PushNewEnv()
        {
            stack.AddLast(new Dictionary<TKey, TValue>());
        }

        public void PopEnv()
        {
            stack.RemoveLast();
        }

        public void Set(TKey key, TValue value)
        {
            stack.Last.Value[key] = value;
        }

        public TValue Get(TKey key)
        {
            var pointer = stack.Last;
            while (pointer!=null && !pointer.Value.ContainsKey(key) )
            {
                pointer = pointer.Previous;
            }
            TValue result = default(TValue);
            if (pointer != null && pointer.Value.ContainsKey(key))
                result = pointer.Value[key];
            else
            {
                Console.WriteLine(key);
            }
            return result;
        }

        public IDictionary<TKey, TValue> GetFirst()
        {
            return stack.First.Value;
        } 
    }

}
