﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Data;
using System.Diagnostics;
using FourqlParser.Structures;
using FourQLInterpreter.InternalModules;

namespace FourQLInterpreter.Reasoning
{
    /// <summary>
    /// Knowledge base of the module.
    /// Keeps facts and domain values.
    /// </summary>
    public class ModuleDataWrapper
    {
        public Module Module { get; private set; }
        public IDictionary<string, ISet<ValueHolder>> DomainDict { get; private set; }
        public IDictionary<string, RelationData> RelationDataDict { get; private set; }

        private IDictionary<string, Relation> _relationDict = new Dictionary<string, Relation>();

        public ModuleDataWrapper(Module module)
        {
            this.Module = module;

        }

        public void Initialize()
        {
            this.DomainDict = new Dictionary<string, ISet<ValueHolder>>();
            this.RelationDataDict = new Dictionary<string, RelationData>();
            foreach (var domain in Module.Domains)
            {
                AddDomain(domain);
            }
            AddPredefinedDomains();
            foreach (var relation in Module.Relations)
            {
                AddRelation(relation);
            }
            foreach (var fact in Module.Facts)
            {
                AddFact(fact, true);
            }
            foreach (var rule in Module.Rules)
            {
                AddConstantsFromHead(rule.Head);
            }
        }

        private void AddConstantsFromHead(Fact head)
        {
            var relation = _relationDict[head.RelationName];

            // add constant values to domains (get domain nammes from relation definition)
            for (int i = 0; i < head.Values.Count; i++)
            {
                if (!head.Values[i].Type.Equals(DomainType.VARIABLE))
                {
                    if (relation.PrimitiveDomains.ContainsKey(i))
                        AddToDomain(DomainTypeUtils.getName(relation.PrimitiveDomains[i]), head.Values[i]);
                    else
                        AddToDomain(relation.DefinedDomains[i], head.Values[i]);
                }
            }
        }

        private void AddPredefinedDomains()
        {
            foreach (DomainType type in Enum.GetValues(typeof(DomainType)))
            {
                this.DomainDict[DomainTypeUtils.getName(type)] = new HashSet<ValueHolder>();
            }
            
        }

        public FactResult AddFact(Fact fact, bool addToDomain=true)
        {
            var result = AddLiteral(RelationDataDict[fact.RelationName], fact.CreateLiteralOfFact(), !fact.IsNegative);
            var relation = _relationDict[fact.RelationName];

            // add fact values to domains (get domain names from relation definition)
            if (addToDomain || DomainDict.ContainsKey(fact.RelationName))
            {
                for (int i = 0; i < fact.Values.Count; i++)
                {
                    if (relation.PrimitiveDomains.ContainsKey(i))
                        AddToDomain(DomainTypeUtils.getName(relation.PrimitiveDomains[i]), fact.Values[i]);
                    else
                        AddToDomain(relation.DefinedDomains[i], fact.Values[i]);
                }
            }
            return result;
        }

        private FactResult AddLiteral(Reasoning.RelationData relationData, Literal literal, bool value)
        {
            if (relationData.Data.ContainsKey(literal.Values))
            {
                var oldValue = relationData.Data[literal.Values];
                var newValue = DetermineValue(value, oldValue);
                relationData.Data[literal.Values] = newValue;
                return newValue.Equals(oldValue) ? FactResult.NO_CHANGE : FactResult.FACT_CHANGED;
            }
            else
            {
                relationData.Data[literal.Values] = FromBoolean(value);
                return  FactResult.FACT_ADDED;

            }
        }

        public void ClearRelation(string name, ICollection<LogicValue> valuesToLeave = null)
        {
            RelationDataDict[name].Clear(valuesToLeave);
        }

        private static LogicValue FromBoolean(bool value)
        {
            return value ? LogicValue.TRUE : LogicValue.FALSE;
        }

        private static LogicValue DetermineValue(bool v1, LogicValue v2)
        {
            switch (v2)
            {
                case LogicValue.TRUE:
                    return v1 ? LogicValue.TRUE : LogicValue.INCONSISTENT;
                case LogicValue.FALSE:
                    return !v1 ? LogicValue.FALSE : LogicValue.INCONSISTENT;
                default:
                    return LogicValue.INCONSISTENT;
            }
        }

        private void AddToDomain(string domainName, ValueHolder value)
        {
            this.DomainDict[domainName].Add(value);
        }


        private void AddRelation(Relation relation)
        {
            this.RelationDataDict[relation.Name] = new RelationData() {RelationName = relation.Name};
            this._relationDict[relation.Name] = relation;
        }

        private void AddDomain(Domain domain)
        {
            this.DomainDict[domain.Name] = new HashSet<ValueHolder>();
        }

        public IEnumerable<ValueHolder> GetVariableDomainValues(FourqlParser.StaticAnalysis.VariableRef variable)
        {
            if (variable.Module.Name != this.Module.Name)
                throw new Exception("Internal error");
            if (variable.DomainName == null)
                return DomainDict[DomainTypeUtils.getName(variable.DomainType)];
            else
                return DomainDict[variable.DomainName];
        }

        public virtual LogicValue GetLiteraValue(string literalName, IList<ValueHolder> values)
        {
            if (DomainDict.ContainsKey(literalName)) // value from domain 
                return DomainDict[literalName].Contains(values.First()) ? LogicValue.TRUE : LogicValue.FALSE;
            else // value from relation data
                return RelationDataDict[literalName].GetLogicValue(values);
        }

        public void FillImportedDomains(IDictionary<string, ModuleDataWrapper> modules)
        {
            foreach (var domain in Module.Domains)
            {
                if (domain.Imports != null)
                    foreach (var import in domain.Imports)
                    {
                        DomainDict[domain.Name].UnionWith(modules[import.Item1].DomainDict[import.Item2]);
                    }
            }
        }

    }


    public class RelationData
    {
        public string RelationName { get; set; }

        public IDictionary<IList<ValueHolder>, LogicValue> Data = new Dictionary<IList<ValueHolder>, LogicValue>(new ListComparer<ValueHolder>());

        internal LogicValue GetLogicValue(IList<ValueHolder> valueList)
        {

            if (Data.ContainsKey(valueList))
                return Data[valueList];
            else
                return LogicValue.UNKNOWN;
        }

        public void Clear(ICollection<LogicValue> valuesToLeave = null)
        {
            if (valuesToLeave == null)
                Data.Clear();
            else
            {
                Data = new Dictionary<IList<ValueHolder>, LogicValue>(Data.Where(d => valuesToLeave.Contains(d.Value)).ToDictionary(d => d.Key, d => d.Value),
                    new ListComparer<ValueHolder>());
            }
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            foreach (var lit in Data)
            {
                sb.Append(RelationName).Append("(");
                foreach (var val in lit.Key)
                {
                    sb.Append(val.ToString()).Append(",");
                }
                sb.Append(") : ").Append(lit.Value.ToString());
                sb.AppendLine();
            }
            return sb.ToString();
        }

        class ListComparer<T> : IEqualityComparer<IList<T>>
        {
            public bool Equals(IList<T> x, IList<T> y)
            {
                return x.SequenceEqual(y);
            }

            public int GetHashCode(IList<T> obj)
            {
                int hashcode = 0;
                foreach (T t in obj)
                {
                    hashcode ^= t.GetHashCode();
                }
                return hashcode;
            }
        }
    }

    public class InternalModuleWrapper : ModuleDataWrapper
    {
        private InternalModule _internalModule;
        public InternalModuleWrapper(InternalModule module) : base(module)
        {
            this._internalModule = module;
        }

        public override LogicValue GetLiteraValue(string literalName, IList<ValueHolder> values)
        {
            return _internalModule.GetLiteralValue(literalName, values);
        }
    }


    public enum FactResult
    {
        FACT_ADDED,
        FACT_CHANGED,
        NO_CHANGE
    }

}
