﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourQLInterpreter.Reasoning
{
    /// <summary>
    /// Represents the result of evaluating an expression.
    /// Each row holds the values for variables and logical value of the expression.
    /// 
    /// (Based on the FourqlModeler.Calculcations.Result class by Łukasz Makar)
    /// </summary>
    public class Result
    {
        /// <summary>
        /// Module in context of which the expression was evaluated
        /// </summary>
        public Module Module { get; set; }

        /// <summary>
        /// Evaluated expression
        /// </summary>
        public Expression Expression { get; set; }

        public List<ColumnDescription> ColumnHeaders { get; set; }
        public IEnumerable<Row> Rows { get; set; }

        public Result()
        {
            ColumnHeaders = new List<ColumnDescription>();
            Rows = new List<Row>();
        }

        public IDictionary<string, object> ToFieldMap(Row row)
        {
            var fieldMap = new Dictionary<string, object>();
            var fieldIter = ColumnHeaders.GetEnumerator();
            var valueIter = row.Values.GetEnumerator();

            while (fieldIter.MoveNext() && valueIter.MoveNext())
            {
                fieldMap.Add(fieldIter.Current.Name, valueIter.Current.Value);
            }

            fieldMap.Add("logic_value", row.Value);
            return fieldMap;
        }

        public IDictionary<string, ValueHolder> ToValueHolderMap(Row row)
        {
            var result = new Dictionary<string, ValueHolder>();
            var fieldIter = ColumnHeaders.GetEnumerator();
            var valueIter = row.Values.GetEnumerator();
            while (fieldIter.MoveNext() && valueIter.MoveNext())
            {
                result.Add(fieldIter.Current.Name, new ValueHolder(valueIter.Current.Value, fieldIter.Current.Type));
            }

            return result;
        } 


        public override string ToString()
        {

            var builder = new StringBuilder("Expression: " + new ExprPrinter().ConvertExpr(Expression).ToString());
            builder.Append("\nHeaders:");
            foreach (var field in ColumnHeaders)
            {
                builder.Append(String.Format(" {0}:{1}", field.Name, DomainTypeUtils.getName(field.Type)));
            }

            builder.Append(String.Format("\nRows ({0}):", Rows.Count()));
            foreach (var row in Rows)
            {
                builder.Append("\n\tValues=[");
                foreach (var value in row.Values)
                {
                    builder.Append(String.Format(" '{0}'", value));
                }
                builder.Append(String.Format(" ] Logical Value={0}", row.Value));
                if (row.RowExplanation != null)
                {
                    builder.Append("\n\t\tExplanation=" + row.RowExplanation);
                }
            }

            return builder.ToString();
        }
    }

    public class ColumnDescription
    {
        public string Name { get; set; }
        public DomainType Type { get; set; }

        public override string ToString()
        {
            return Name + ":" + DomainTypeUtils.getName(Type);
        }
    }

    public class Row
    {
        public List<ValueHolder> Values { get; set; }
        public LogicValue Value { get; set; }

        public Explanation.Explanation RowExplanation { get; set; }

        public Row()
        {
            Values = new List<ValueHolder>();
            RowExplanation = new Explanation.Explanation();
        }

        public override string ToString()
        {
            var builder = new StringBuilder("Values=[");
            foreach (var value in Values)
            {
                builder.Append(String.Format(" '{0}'", value));
            }
            builder.Append(String.Format(" ] Logical Value={0}", Value));
            return builder.ToString();
        }
    }

}
