﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Configuration;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;
using FourQLInterpreter.Reasoning;

namespace FourQLInterpreter.WSM
{

    public class EvaluationHistory
    {

        public Module EvaluatedModule { get; set; }

        public ModuleDataWrapper Model { get; set; }

        public SortedDictionary<int, IList<EvaluationEvent>> Iterations { get; set; }

        public EvaluationHistory()
        {
            Iterations = new SortedDictionary<int, IList<EvaluationEvent>>();
        }

        public void AddIteration(int i)
        {
            if (!Iterations.ContainsKey(i))
                Iterations[i] = new List<EvaluationEvent>();
        }

        public void AddEvent(int iteration, EvaluationEvent evaluationEvent)
        {
            AddIteration(iteration);
            Iterations[iteration].Add(evaluationEvent);
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder("WSM Evaluation history of module " + EvaluatedModule.Name);
            sb.AppendLine();
            foreach (var it in Iterations)
            {
                sb.AppendLine("Iteration " + it.Key.ToString());
                foreach (var ev in it.Value)
                {
                    sb.AppendLine(ev.ToString());
                }
            }
            sb.AppendLine("FINAL MODEL of " + EvaluatedModule.Name);

            foreach (var rel in Model.RelationDataDict.Values)
            {
                sb.AppendLine(rel.ToString());
            }

            return sb.ToString();
        }
    }

    /// <summary>
    /// Represents a single step in WSM computation process
    /// </summary>
    public class EvaluationEvent
    {
        /// <summary>
        /// Type of the event -> stage of the evaluation process
        /// </summary>
        public EventType Type { get; set; }

        /// <summary>
        /// Rule from which the facts where generated
        /// </summary>
        public Rule Rule { get; set; }

        /// <summary>
        /// Collection of facts and results of adding them to the model.
        /// Result can say that 
        /// - the fact already existed (FactResult.NO_CHANGE)
        /// - the fact did not yet exist in the model (FactResult.FACT_ADDED)
        /// - negation of the fact existed in the model (FactResult.FACT_CHANGED)
        /// </summary>
        public ICollection<Tuple<Fact, FactResult>> Facts { get; set; }


        public EvaluationEvent()
        {
            Facts = new List<Tuple<Fact, FactResult>>();
        }


        public void AddFact(Fact fact, FactResult factAdded)
        {
            Facts.Add(new Tuple<Fact, FactResult>(fact, factAdded));
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("Type : ").Append(Type.ToString());
            sb.AppendLine();
            sb.Append("Rule : ").Append(Rule);
            sb.AppendLine();
            foreach (var tuple in Facts)
            {
                sb.Append(tuple.Item1).Append(tuple.Item2);
                sb.AppendLine();
            }
            return sb.ToString();
        }
    }

    public enum EventType
    {
        /// <summary>
        /// Generate new facts from true premises
        /// </summary>
        Generate,
        /// <summary>
        /// Propagate inconsistency
        /// </summary>
        FindCorrection,
        /// <summary>
        /// Retract conclusions based on defeated premises
        /// </summary>
        Retract
    }
}
