﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.StaticAnalysis;
using FourqlParser.Structures;
using Microsoft.Win32;

namespace FourQLInterpreter.WSM
{
    public class ModuleHierarchy
    {
        private IDictionary<string, Vertex> _vertexMap;

        private IDictionary<int, ICollection<Vertex>> _hierarchyGraph;

        private ICollection<Module> _modules;

        public ModuleHierarchy(ICollection<Module> modules)
        {
            _vertexMap = new Dictionary<string, Vertex>();
            _hierarchyGraph = new SortedDictionary<int, ICollection<Vertex>>();
            _modules = modules;
            foreach (var module in modules)
            {
                _vertexMap[module.Name] = new Vertex(module);
            }
        }

        /// <summary>
        /// Constructs a module hierarchy graph - A map of layers and modules on each layer.
        /// Layer 0 consists of modules which do not have any external dependencies.
        /// Layer 1 - of modules which all the dependencies are layer 0
        /// ...etc
        /// </summary>
        /// <returns></returns>
        public IDictionary<int, ICollection<Vertex>> ConstructHierarchy()
        {
            _vertexMap = new Dictionary<string, Vertex>();
            _hierarchyGraph = new SortedDictionary<int, ICollection<Vertex>>();
            foreach (var module in _modules)
            {
                _vertexMap[module.Name] = new Vertex(module);
            }

            var layer0 = new HashSet<Vertex>();

            // Fill edges of each vertex
            foreach (var vertex in _vertexMap.Values)
            {
                var module = vertex.Module;
                foreach (var rule in module.Rules)
                {
                    // calculate references from each rule
                    var ruleReferences = new ModuleReferenceFinder().GetReferences(module.Name, rule.Body);
                    foreach (var edge in ruleReferences.Keys)
                    {
                        vertex.EdgesTo.Add(_vertexMap[edge]);
                        _vertexMap[edge].EdgesFrom.Add(vertex);
                    }
                }
                // Calculate references created by importing domains
                foreach (var domain in module.Domains)
                {
                    if (domain.Imports != null)
                    {
                        foreach (var import in domain.Imports)
                        {
                            vertex.EdgesTo.Add(_vertexMap[import.Item1]);
                            _vertexMap[import.Item1].EdgesFrom.Add(vertex);
                            
                        }
                    }
                }
                // Remove self-reference
                vertex.EdgesTo.Remove(vertex);

                // if the vertex has no outgoing edges it's in the bottom layer of the graph
                if (vertex.EdgesTo.Count == 0)
                    vertex.Layer = -1;
                    layer0.Add(vertex);
            }

            // set layer number for each vertex
            foreach (var vertex in layer0)
            {
                LayerDfs(vertex, 0);
            }

            foreach (var vertex in _vertexMap.Values)
            {
                if (!_hierarchyGraph.ContainsKey(vertex.Layer))
                    _hierarchyGraph[vertex.Layer] = new HashSet<Vertex>();
                _hierarchyGraph[vertex.Layer].Add(vertex);
            }

            return _hierarchyGraph;
        }

        //Dfs for setting layer of each vertex
        private void LayerDfs(Vertex vertex, int i)
        {
            if (vertex.Layer < i)
            {
                vertex.EdgesFrom.Remove(vertex);
                vertex.Layer = i;
                foreach (var neighbour in vertex.EdgesFrom)
                    LayerDfs(neighbour, i+1);
            }
        }

        /// <summary>
        /// Method for calculating a subgraph where given module is the root.
        /// Used when we do not want to rebuild the WSM of all the modules, but only those which need to be
        /// rebuild after a module has changed.
        /// </summary>
        /// <param name="module"></param>
        /// <returns>Subgraph - Map of layer numbers and vertices on each layer</returns>
        public SortedDictionary<int, ICollection<Vertex>> GetModuleSubgraph(String module)
        {
            var vertex = _vertexMap[module];
            var result = new SortedDictionary<int, ICollection<Vertex>> ();
            foreach (var neighbour in vertex.EdgesFrom)
            {
                PushVertex(neighbour, result);
            }
            return result;
        }

        private void PushVertex(Vertex vertex, IDictionary<int, ICollection<Vertex>> graph)
        {
            if (!graph.ContainsKey(vertex.Layer))
                graph[vertex.Layer] = new HashSet<Vertex>();
            if (!graph[vertex.Layer].Contains(vertex))
            {
                graph[vertex.Layer].Add(vertex);
                foreach (var neighbour in vertex.EdgesFrom)
                {
                    PushVertex(neighbour, graph);
                }
            }
        }



        /// <summary>
        /// Adds a new module to the hierarchy.
        /// Module has to have outgoing edges to already existing modules.
        /// </summary>
        /// <param name="module"></param>
        /// <returns>hierarchy including the new module</returns>
        public IDictionary<int, ICollection<Vertex>> AddModule(Module module)
        {
            if (!_vertexMap.ContainsKey(module.Name))
            {
                var vertex = new Vertex(module);
                _vertexMap[module.Name] = vertex;
           
                foreach (var rule in module.Rules)
                {
                    // calculate references from each rule
                    var ruleReferences = new ModuleReferenceFinder().GetReferences(module.Name, rule.Body);
                    foreach (var edge in ruleReferences.Keys)
                    {
                        vertex.EdgesTo.Add(_vertexMap[edge]);
                        _vertexMap[edge].EdgesFrom.Add(vertex);
                    }
                }
                // Calculate references created by importing domains
                foreach (var domain in module.Domains)
                {
                    if (domain.Imports != null)
                    {
                        foreach (var import in domain.Imports)
                        {
                            vertex.EdgesTo.Add(_vertexMap[import.Item1]);
                            _vertexMap[import.Item1].EdgesFrom.Add(vertex);

                        }
                    }
                }
                // Remove self-reference
                vertex.EdgesTo.Remove(vertex);
                vertex.EdgesFrom.Remove(vertex);

                //calculate layer
                int layer = 0;
                foreach (var neighbour in vertex.EdgesTo)
                {
                    if (neighbour.Layer >= layer)
                        layer = neighbour.Layer + 1;
                }
                vertex.Layer = layer;
                _hierarchyGraph[layer].Add(vertex);
            }
            return _hierarchyGraph;
        }
    }



     public class Vertex
    {
        public Module Module { get; private set; }
        public ISet<Vertex> EdgesTo { get; set; }

        public ISet<Vertex> EdgesFrom { get; set; }

        public int Layer { get; set; }

        public Vertex(Module module)
        {
            Module = module;
            EdgesTo = new HashSet<Vertex>();
            EdgesFrom = new HashSet<Vertex>();
            Layer = -2;
        }


         public override string ToString()
         {
             return Module.Name;
         }
    }
}
