﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;
using FourQLInterpreter.Reasoning;
using NLog;

namespace FourQLInterpreter.WSM
{
    public class OptimizedWSMEvaluator : WsmEvaluator
    {
        private static readonly Logger Log = LogManager.GetCurrentClassLogger();

        /// <summary>
        /// Maps relation name to set of all rules in which the relation appears in the body.
        /// </summary>
        private IDictionary<string, ISet<Rule>> _ruleRelDependencyMap = new Dictionary<string, ISet<Rule>>();
        private IDictionary<string, ISet<Rule>> _ruleDomDependencyMap = new Dictionary<string, ISet<Rule>>();


        public OptimizedWSMEvaluator(Module module, Evaluator evaluator, bool saveHistory = false) : base(module, evaluator, saveHistory)
        {
        }


        private void GenerateFrom(LogicValue value)
        {

            ISet<Rule> activeRules = new HashSet<Rule>(CurrentModule.Rules);

            while (activeRules.Count > 0)
            {
                var rules = new HashSet<Rule>(activeRules);
                activeRules.Clear();

                foreach (var rule in rules)
                {
                    Log.Debug("Generating " + value + " facts from rule: " + rule.ToString());
                    var result = GenerateFacts(rule, value);

                    if (HasModelChanged(result))
                        activeRules.UnionWith(_ruleRelDependencyMap[rule.Head.RelationName]);
                }
            }
        }

        protected override void GenerateFromTruth()
        {
            //ClearRelations(LogicValue.INCONSISTENT);

            InsertFacts();

            GenerateFrom(LogicValue.TRUE);
        }

        protected override void FindCorrection()
        {
            GenerateFrom(LogicValue.INCONSISTENT);
        }

        protected override void Prepare()
        {
            base.Prepare();
            GatherRelationRuleDependencies();
        }


        private void GatherRelationRuleDependencies()
        {
            foreach (var relation in CurrentModule.Relations)
            {
                _ruleRelDependencyMap[relation.Name] = new HashSet<Rule>();
            }
            
            foreach (var rule in CurrentModule.Rules)
            {
                foreach (var relationName in rule.Body.Accept(new RelationNameFinder()))
                {
                    _ruleRelDependencyMap[relationName].Add(rule);
                }

            }
        }

    }

    class RelationNameFinder : Expression.Visitor<ICollection<string>>
    {
        private ICollection<string> result = new HashSet<string>(); 

        public ICollection<string> visitLiteral(Literal literal)
        {
            result.Add(literal.RelationName);
            return result;
        }

        public ICollection<string> visitNeg(NegExpression negExpr)
        {
            return negExpr.Next.Accept(this);
        }

        public ICollection<string> visitModule(ModuleExpression moduleExpr)
        {
            return result;
        }

        public ICollection<string> visitAnd(AndExpression andExpr)
        {
            foreach (var next in andExpr.Nexts)
            {
                next.Accept(this);
            }
            return result;
        }

        public ICollection<string> visitOr(OrExpression orExpr)
        {
            foreach (var next in orExpr.Nexts)
            {
                next.Accept(this);
            }
            return result;
        }

        public ICollection<string> visitExists(ExistsExpression exExpr)
        {
            return exExpr.Next.Accept(this);
        }

        public ICollection<string> visitAll(AllExpression allExpr)
        {
            return allExpr.Next.Accept(this);
        }
    }


}
