﻿using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;
using FourQLInterpreter.Reasoning;
using NLog;

namespace FourQLInterpreter.WSM
{

    /// <summary>
    /// Implements the bottom-up WSM evaluation algorithm
    /// </summary>
    public class WsmEvaluator
    {

        private static readonly Logger Log = LogManager.GetCurrentClassLogger();

        public Module CurrentModule { get; set; }

        public Evaluator Evaluator { get; set; }

        public bool SaveHistory { get; set; }

        public EvaluationHistory History { get; private set; }

        private int _iteration = 0;

        private int _inconsistencies = 0;

        private ModuleDataWrapper _model;

      
        public WsmEvaluator(Module module, Evaluator evaluator, bool saveHistory = false)
        {
            CurrentModule = module;
            Evaluator = evaluator;
            SaveHistory = saveHistory;
            History = new EvaluationHistory() {EvaluatedModule = CurrentModule};
        }

        protected virtual void GenerateFromTruth()
        {
            //ClearRelations(LogicValue.INCONSISTENT);

            InsertFacts();

            var finished = false;

            while (!finished)
            {
                finished = true;
                foreach (var rule in CurrentModule.Rules)
                {
                    var result = GenerateFacts(rule, LogicValue.TRUE);
                    Log.Debug("Generating " + LogicValue.TRUE + " facts from rule: " + rule.ToString() + "\nGenerated " + result.ToString());

                    finished = !HasModelChanged(result) && finished;
                }  

            }
        }


        protected void InsertFacts()
        {
            foreach (var fact in CurrentModule.Facts)
            {
                var res = _model.AddFact(fact);
                if (SaveHistory)
                {
                    var ev = new EvaluationEvent() { Type = EventType.Generate, Rule = new Rule() { Head = fact } };
                    ev.AddFact(fact, res);
                    History.AddEvent(_iteration, ev);
                }
            }
        }

        private EvaluationEvent BuildRetractEvent(int iteration)
        {
            var ev = new EvaluationEvent() {Type = EventType.Retract};
            var leftFacts = from rel in _model.RelationDataDict.Values
                from lit in rel.Data
                select new Fact() {IsNegative = false, Values = lit.Key, RelationName = rel.RelationName};
            foreach (var fact in leftFacts)
            {
                ev.AddFact(fact, FactResult.NO_CHANGE);
                ev.AddFact(new Fact() {IsNegative = true, RelationName = fact.RelationName, Values = fact.Values} , FactResult.NO_CHANGE);
            }
            return ev;
        }

        protected void ClearRelations(LogicValue logicValue)
        {
            var moduleRelationNames =
                from rel in CurrentModule.Relations
                select rel.Name;

            ClearRelations(moduleRelationNames, new List<LogicValue>() { logicValue });

            if (SaveHistory)
                History.AddEvent(_iteration, BuildRetractEvent(_iteration));
        }

        protected bool HasModelChanged(EvaluationEvent result)
        {
            return !result.Facts.All(f => f.Item2.Equals(FactResult.NO_CHANGE));
        }

        protected EvaluationEvent GenerateFacts(Rule rule, LogicValue value)
        {
            // select variable names from rule head
            var variableNames = from val in rule.Head.Values
                where val.Type.Equals(DomainType.VARIABLE)
                select val.Value.ToString();
            var variableRefs = FourqlParser.StaticAnalysis.ExprVariableGatherer.findVariablesOfLiteral(CurrentModule,
                rule.Head);
            // evaluate the rule body
            var result = Evaluator.Evaluate(rule.Body, CurrentModule, variableNames.ToList(), value, variableRefs);
            // Debug.WriteLine(result.ToString());
            var evalEvent = new EvaluationEvent()
            {
                Rule = rule,
                Type = value.Equals(LogicValue.TRUE) ? EventType.Generate : EventType.FindCorrection
            };
            
            // create a fact from each row of the result
            foreach (var row in result.Rows)
            {
                if (row.Value.Equals(value))
                {
                    var r = result.ToValueHolderMap(row);

                    var fact = CreateFact(rule, r);

                    var factAdded = _model.AddFact(fact);

                    if (factAdded.Equals(FactResult.FACT_CHANGED))
                        _inconsistencies++;

                    evalEvent.AddFact(fact, factAdded);

                    // if the premise is inconsistent the negation also has to be added to the model
                    if (value.Equals(LogicValue.INCONSISTENT))
                    {
                        var fact2 = new Fact()
                        {
                            IsNegative = !fact.IsNegative,
                            Values = fact.Values,
                            RelationName = fact.RelationName
                        };
                        var fact2Added = _model.AddFact(fact2);
                        if (fact2Added.Equals(FactResult.FACT_CHANGED))
                            _inconsistencies++;
                        evalEvent.AddFact(fact2, fact2Added);
                    }
                }
            }

            if (SaveHistory )
                History.AddEvent(_iteration, evalEvent);

            return evalEvent;
        }

        private Fact CreateFact(Rule rule, IDictionary<string, ValueHolder> variableValues)
        {
            var fact = new Fact()
            {
                IsNegative = rule.Head.IsNegative,
                ParserContext = null,
                RelationName = rule.Head.RelationName
            };

            // in the newly created fact replace variables by evaluated values, copy constants from the rule head.
            fact.Values = (from val in rule.Head.Values
                select val.Type.Equals(DomainType.VARIABLE) ? variableValues[val.Value.ToString()] : val).ToList();

            return fact;
     ;   }

        protected virtual void FindCorrection()
        {
            var finished = false;

            while (!finished)
            {
                finished = true;
                foreach (var rule in CurrentModule.Rules)
                {
                    Log.Debug("Generating " + LogicValue.INCONSISTENT + " facts from rule: " + rule.ToString());
                   
                    var result = GenerateFacts(rule, LogicValue.INCONSISTENT);

                    Log.Debug("Generated " + result.ToString());
                    finished = !HasModelChanged(result) && finished;
                }

            }
        }

        public void ConstructWsm()
        {

            Prepare();

            //GenerateFromTruth();

            while (true)
            {
                var prevInconsistencies = _inconsistencies;

                _iteration++;

                GenerateFromTruth();

                FindCorrection();

                if (_inconsistencies == prevInconsistencies)
                    break;

                ClearRelations(LogicValue.INCONSISTENT);
            }

            if (SaveHistory)
                History.Model = _model;
        }

        protected virtual void Prepare()
        {
            _model = Evaluator.ModulesData[CurrentModule.Name];
            _model.FillImportedDomains(Evaluator.ModulesData);
            ClearAllRelations();
            _iteration = 0;
        }

        private void ClearAllRelations()
        {
            var relationNames = 
                from rel in CurrentModule.Relations
                select rel.Name;

            ClearRelations(relationNames);

        }


        // Clears all the relation data in the model
        private void ClearRelations(IEnumerable<string> relationNames, ICollection<LogicValue> valuesToLeave=null)
        {
            foreach (var rel in relationNames)
            {
                _model.ClearRelation(rel, valuesToLeave);
            }

        }



        

    }
}
