﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using FourqlParser.Parsers;
using System.Collections.Generic;
using System.Linq;
using FourqlParser.Structures;
using FourQLInterpreter.Reasoning;

namespace _4QLInterpreterTest
{
    [TestClass]
    public class ExpressionEvaluationTest : BaseModuleTests
    {

        

        [TestMethod]
        public void TestAnd()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);
           
            var expression = CreateExpression("r(a) and r(X)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(2, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                if (m["X"].Equals("a"))
                    Assert.AreEqual(LogicValue.TRUE, m["logic_value"]);
                if (m["X"].Equals("b"))
                    Assert.AreEqual(LogicValue.FALSE, m["logic_value"]);
            }

            expression = CreateExpression("r(c) and r(X)", modules, "m");
            result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(2, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                if (m["X"].Equals("a"))
                    Assert.AreEqual(LogicValue.UNKNOWN, m["logic_value"]);
                if (m["X"].Equals("b"))
                    Assert.AreEqual(LogicValue.FALSE, m["logic_value"]);
            }

        }
        [TestMethod]
        public void TestOr()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression(" r(a) or r(X)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(2, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                if (m["X"].Equals("a"))
                    Assert.AreEqual(LogicValue.TRUE, m["logic_value"]);
                if (m["X"].Equals("b"))
                    Assert.AreEqual(LogicValue.TRUE, m["logic_value"]);
            }
        }
        [TestMethod]
        public void TestNot()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression("not r(X)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(2, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                if (m["X"].Equals("a"))
                    Assert.AreEqual(LogicValue.FALSE, m["logic_value"]);
                if (m["X"].Equals("b"))
                    Assert.AreEqual(LogicValue.TRUE, m["logic_value"]);
            }
        }
        [TestMethod]
        public void TestExists()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression("EXISTS Y:dom r(Y)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                Assert.AreEqual(LogicValue.TRUE, m["logic_value"]);
            }
        }
        [TestMethod]
        public void TestAll()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression("ALL Y:dom r(Y)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                Assert.AreEqual(LogicValue.FALSE, m["logic_value"]);
            }
        }
        [TestMethod]
        public void TestIn()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression("n.r(X) in {INCONS}", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(2, result.Rows.Count());
            foreach (var row in result.Rows)
            {
                var m = result.ToFieldMap(row);
                if (m["X"].Equals("a"))
                    Assert.AreEqual(LogicValue.FALSE, m["logic_value"]);
                if (m["X"].Equals("b"))
                    Assert.AreEqual(LogicValue.TRUE, m["logic_value"]);
            }
        }
        [TestMethod]
        public void TestEmptyDomain()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression("r(X) and n.s(Y)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(0, result.Rows.Count());
        }
        [TestMethod]
        public void TestFreeVariables()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);

            var expression = CreateExpression("r(X) and r(Y)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"], new List<string>() { "X"});
            // 
            Console.WriteLine(result);
            Assert.AreEqual(2, result.Rows.Count());
        }

        [TestMethod]
        public void TestMinEvalValue()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);
            
            var expression = CreateExpression("r(c) and r(X)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"], new List<string>() { "X" }, LogicValue.INCONSISTENT);
            Console.WriteLine(result);
            Assert.AreEqual(0, result.Rows.Count());
        }


        [TestMethod]
        public void TestQuantifierDomains()
        {
            var modules = LoadModules("testModuleNoRules.4ql");
            var evaluator = CreateEvaluator(modules.Values);
            evaluator.Explain = true;

            var expression = CreateExpression("EXISTS X: dom r(X)", modules, "m");
            var result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count()); 
            
            expression = CreateExpression("EXISTS X: n.dom r(X)", modules, "m");
            result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            Assert.IsTrue(result.Rows.First().RowExplanation.QuantifierVariables.First() == null);

            expression = CreateExpression("EXISTS X: dom r(X)", modules, "n");
            result = evaluator.Evaluate(expression, modules["n"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            Assert.IsTrue(result.Rows.First().RowExplanation.QuantifierVariables.First() == null);

            expression = CreateExpression("EXISTS X: m.literal r(X)", modules, "n");
            result = evaluator.Evaluate(expression, modules["n"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            Assert.IsTrue(result.Rows.First().RowExplanation.QuantifierVariables.First() == null);
           
            expression = CreateExpression("EXISTS X: literal r(X)", modules, "m");
            result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            Assert.IsTrue(result.Rows.First().RowExplanation.QuantifierVariables.First() == null);


            expression = CreateExpression("EXISTS X: n.literal r(X)", modules, "m");
            result = evaluator.Evaluate(expression, modules["m"]);
            Console.WriteLine(result);
            Assert.AreEqual(1, result.Rows.Count());
            Assert.IsTrue(result.Rows.First().RowExplanation.QuantifierVariables.First() != null);
        }
    }
}
