﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;

namespace FourqlModeler.Calculcations
{
    public class ExprValueSetter : Expression.Visitor<Expression>
    {
        public IDictionary<string, ValueHolder> Values { get; set; }

        public ExprValueSetter(IDictionary<string, ValueHolder> values)
        {
            Values = values;
        }

        public Fact SetInFact(Fact fact)
        {
            var factWithReplacements = new Fact() { IsNegative = fact.IsNegative, RelationName = fact.RelationName, ParserContext = fact.ParserContext };
            foreach (var value in fact.Values)
            {
                if (value.Type == DomainType.VARIABLE && Values.ContainsKey(value.valueAs<string>()))
                {
                    var newValue = Values[value.valueAs<string>()];
                    factWithReplacements.Values.Add(new ValueHolder(newValue.Value, newValue.Type));
                }
                else
                {
                    factWithReplacements.Values.Add(new ValueHolder(value.Value, value.Type));
                }
            }
            return factWithReplacements;
        }

        #region Visitor<Expression> Members

        public Expression visitLiteral(Literal literal)
        {
            var newLiteral = new Literal() { ParserContext = literal.ParserContext, RelationName = literal.RelationName };
            foreach (var value in literal.Values)
            {
                if (value.Type == DomainType.VARIABLE && Values.ContainsKey(value.valueAs<string>()))
                {
                    var newValue = Values[value.valueAs<string>()] ?? new ValueHolder("?", value.Type);
                    newLiteral.Values.Add(new ValueHolder(newValue.Value, newValue.Type));
                }
                else
                {
                    newLiteral.Values.Add(new ValueHolder(value.Value, value.Type));
                }
            }
            return newLiteral;
        }

        public Expression visitNeg(NegExpression negExpr)
        {
            return new NegExpression() { ParserContext = negExpr.ParserContext, Next = negExpr.Next.Accept(this) };
        }

        public Expression visitModule(ModuleExpression moduleExpr)
        {
            return new ModuleExpression() { ParserContext = moduleExpr.ParserContext, IsNegative = moduleExpr.IsNegative, ModuleName = moduleExpr.ModuleName, InValues = new List<LogicValue>(moduleExpr.InValues), Next = moduleExpr.Next.Accept(this) };
        }

        public Expression visitAnd(AndExpression andExpr)
        {
            var newAnd = new AndExpression() { ParserContext = andExpr.ParserContext };
            foreach (var next in andExpr.Nexts)
            {
                newAnd.Nexts.Add(next.Accept(this));
            }
            return newAnd;
        }

        public Expression visitOr(OrExpression orExpr)
        {
            var newOr = new OrExpression() { ParserContext = orExpr.ParserContext };
            foreach (var next in orExpr.Nexts)
            {
                newOr.Nexts.Add(next.Accept(this));
            }
            return newOr;
        }

        public Expression visitExists(ExistsExpression exExpr)
        {
            ValueHolder prevValue = null;
            if (Values.ContainsKey(exExpr.VariableName))
            {
                prevValue = Values[exExpr.VariableName];
                Values.Remove(exExpr.VariableName);
            }

            var newEx = new ExistsExpression()
            {
                ParserContext = exExpr.ParserContext,
                VariableDomain = exExpr.VariableDomain,
                DomainModule = exExpr.DomainModule,
                VariableName = exExpr.VariableName,
                Next = exExpr.Next.Accept(this)
            };
            if (prevValue != null)
            {
                Values[exExpr.VariableName] = prevValue;
            }
            return newEx;
        }

        public Expression visitAll(AllExpression allExpr)
        {
            ValueHolder prevValue = null;
            if (Values.ContainsKey(allExpr.VariableName))
            {
                prevValue = Values[allExpr.VariableName];
                Values.Remove(allExpr.VariableName);
            }
            var newExpr = new AllExpression()
            {
                ParserContext = allExpr.ParserContext,
                VariableDomain = allExpr.VariableDomain,
                DomainModule = allExpr.DomainModule,
                VariableName = allExpr.VariableName,
                Next = allExpr.Next.Accept(this)
            };
            if (prevValue != null)
            {
                Values[allExpr.VariableName] = prevValue;
            }
            return newExpr;
        }

        #endregion
    }
}
