﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using FourqlModeler.Calculcations;
using FourqlModeler.Communication;
using FourqlModeler.Communication.Import;
using FourqlModeler.Utility;
using FourqlModeler.ViewModels;
using FourqlModeler.Views;
using Parago.Windows;

namespace FourqlModeler.Commands
{
    class BuildModulesCommand : ICommand
    {
        private ModuleStorageViewModel ModuleStorageViewModel { get; set; }

        public BuildModulesCommand(ModuleStorageViewModel moduleStorageViewModel)
        {
            ModuleStorageViewModel = moduleStorageViewModel;
        }

        #region ICommand Members

        public event EventHandler CanExecuteChanged
        {
            add { CommandManager.RequerySuggested += value; }
            remove { CommandManager.RequerySuggested -= value; }
        }

        public bool CanExecute(object parameter)
        {
            return true;
        }

        public void Execute(object parameter)
        {
            var changedModules = parameter as ICollection<ModuleViewModel>;
            if (changedModules != null)
            {
                var result = ProgressDialog.Execute(Application.Current.MainWindow, "Initiating module hierarchy build procedure...", () =>
                {
                    // Clear the selected error.
                    ProgressDialog.Current.Report("Clearing error selection...");
                    Application.Current.Dispatcher.Invoke(() => ClearErrorSelection());
                    // Parse the changed modules.
                    int i = 0;
                    foreach (var moduleViewModel in changedModules)
                    {
                        ProgressDialog.Current.Report("Parsing {0}/{1} file ...\nFile name: {2}\nFile dir: {3}", ++i, changedModules.Count, moduleViewModel.FileName, moduleViewModel.FileDir);
                        Application.Current.Dispatcher.Invoke(() =>
                        {
                            ModuleUtils.updateWithParsedContent(moduleViewModel);
                        });
                    }
                    ProgressDialog.Current.Report("Checking for semantical errors.");
                    // We assume that the modules are also in the storage.
                    Application.Current.Dispatcher.Invoke(() =>
                    {
                        // Clear the common error storage.
                        ModuleStorageViewModel.AllErrors.Clear();
                        // Run the analysis.
                        ModuleUtils.updateAndGroupModuleViewModels(ModuleStorageViewModel.Modules);
                        // Add all founds errors into the common error storage.
                        foreach (var module in ModuleStorageViewModel.Modules)
                        {
                            // Syntax errors..
                            foreach (var error in module.SyntaxErrors)
                            {
                                ModuleStorageViewModel.AllErrors.Add(error);
                            }
                            // ... as well as semantic ones.
                            foreach (var error in module.StaticAnalysisErrors)
                            {
                                ModuleStorageViewModel.AllErrors.Add(error);
                            }
                        }
                        // Update error visuals for module editors.
                        UpdateErrorVisuals();
                    });
                    // FIXME:
                    // Load the parsed structures to our module memory.
                    var builder = new ReasoningOrderBuilder();
                    Application.Current.Dispatcher.Invoke(() =>
                    {
                        foreach (var module in ModuleStorageViewModel.Modules)
                        {
                            builder.AddModel(module);
                        }
                    });
                    var reasoningOrder = builder.Done();
                    // Remember internal-external modules to exclude from the combined file.
                    var internalExternalModules = new HashSet<string>();
                    foreach (var moduleViewModel in ModuleStorageViewModel.Modules)
                    {
                        if (moduleViewModel.IsInternalExternal)
                        {
                            foreach (var module in moduleViewModel.CurrentProgram.Modules)
                            {
                                internalExternalModules.Add(module.Name);
                                //internalExternalModules.Add(DomainModuleBuilder.ToDomainModule(module.Name));
                            }
                        }
                    }
                    // Create a loadable module dictionary.
                    var loadableModules = reasoningOrder.ModuleNodes.Where(entry => !entry.Value.ErroneousDependencies).ToDictionary(entry => entry.Value.Module.Name, entry => entry.Value.Module);
                    /*foreach (var loadableModule in loadableModules.Values.Where(module => !internalExternalModules.Contains(module.Name)).ToArray())
                    {
                        var domainModule = new DomainModuleBuilderNoOpt().CreateDomainModule(loadableModule);
                        loadableModules[domainModule.Name] = domainModule;
                    }*/
                    Application.Current.Dispatcher.Invoke(() =>
                    {
                        // Create a new combined module file.
                        new ModuleReader().createCombinedModuleFile("temp.4ql", loadableModules.Where(module => !internalExternalModules.Contains(module.Key)).ToDictionary(entry => entry.Key, entry => entry.Value));
                        // Clear modules from the interpreter first.
                        ApplicationClient.Instance.ClearModules();
                        //ApplicationClient.Instance.SendCommand(MessageFactory.clearMessage(new SuccessHandler()));
                        // Ask the interpreter to load the temporary file.
                        ApplicationClient.Instance.LoadModules(Path.GetFullPath("temp.4ql"));
                        //ApplicationClient.Instance.SendCommand(MessageFactory.importMessage(Path.GetFullPath("temp.4ql"), new SuccessHandler()));
                        // Set loadable modules as a currently used hierarchy.
                        ModuleRepository.Instance.Modules = loadableModules;
                    });
                }, ProgressDialogSettings.WithSubLabel);

                if (result.OperationFailed)
                {
                    MessageBox.Show("Failed to parse and load the 4QL files: " + (result.Error != null ? result.Error.Message : result.Result), "Parse failed");
                }
            }
        }
        
        #endregion

        private void ClearErrorSelection()
        {
            ModuleStorageViewModel.SelectedError = null;
        }

        private void UpdateErrorVisuals()
        {
            foreach (var module in ModuleStorageViewModel.Modules)
            {
                var tab = MainWindowViewModel.Instance.MainWindowView.TabController.GetTab(String.Format("{0}{1}#def", module.FileDir, module.FileName));
                if (tab != null)
                {
                    var view = tab.Content as ModuleEditorView;
                    if (view != null)
                    {
                        view.EditorBox.TextArea.TextView.Redraw();
                        view.EditorBox.TextArea.TextView.InvalidateVisual();
                    }
                }
            }
        }

        private void PrintReasoningOrder(ReasoningOrder reasoningOrder)
        {
            MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand("REASONING ORDER:");
            foreach (var duplicates in reasoningOrder.DuplicatedModules)
            {
                MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand(String.Format("  DUPLICATES OF {0}:", duplicates.Key));
                foreach (var source in duplicates.Value)
                {
                    MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand(String.Format("    DUP: {0}{1}", source.FileDir, source.FileName));
                }
            }
            foreach (var module in reasoningOrder.ModuleNodes)
            {
                MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand(
                    String.Format("  MODULE {0} ({1}{2}@{3}:{4}):\n   @Erroneous={5},\n   @Group={6},\n   @SyntaxErrors={7},\n   @Errors={8},\n   @ModuleErrors={9}",
                        module.Key,
                        module.Value.ModuleViewModel.FileDir,
                        module.Value.ModuleViewModel.FileName,
                        module.Value.Module.startLine(),
                        module.Value.Module.startColumn(),
                        module.Value.ErroneousDependencies,
                        module.Value.Group,
                        module.Value.ModuleViewModel.SyntaxErrors,
                        module.Value.ModuleViewModel.StaticAnalysisErrors,
                        module.Value.ModuleViewModel.ModuleStaticAnalysisErrors.ContainsKey(module.Key) ? module.Value.ModuleViewModel.ModuleStaticAnalysisErrors[module.Key] : null));
                foreach (var node in module.Value.DependsOn)
                {
                    MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand(
                        String.Format("\n    #DependsOn={0}", node.Key.Module.Name));
                }
                foreach (var invalidModule in module.Value.DependsOnInvalid)
                {
                    MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand(
                        String.Format("\n    #DependsOnInvalid={0}, Duplicate={1}", invalidModule, reasoningOrder.DuplicatedModules.ContainsKey(invalidModule.Key)));
                }
                foreach (var node in module.Value.IsDependencyOf)
                {
                    MainWindowViewModel.Instance.ConsoleViewModel.SendDebugCommand(
                        String.Format("\n    #IsDependencyOf={0}", node.Module.Name));
                }
            }
        }
    }
}
