﻿using FourqlModeler.ViewModels;
using ICSharpCode.AvalonEdit;
using ICSharpCode.AvalonEdit.Document;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Input;
using System.Windows.Controls;
using ICSharpCode.AvalonEdit.Rendering;
using System.Windows.Media;
using ICSharpCode.AvalonEdit.Editing;

namespace FourqlModeler.Controls
{
    public class MvvmTextEditor : TextEditor, INotifyPropertyChanged
    {
        public static readonly DependencyProperty DocumentTextProperty =
            DependencyProperty.Register("Text", typeof(string), typeof(TextEditor),
                new FrameworkPropertyMetadata(String.Empty, FrameworkPropertyMetadataOptions.BindsTwoWayByDefault, (obj, args) =>
                {
                    var target = (TextEditor)obj;
                    var value = (string)args.NewValue;
                    if (!target.Text.Equals(value))
                    {
                        target.Text = (string)args.NewValue;
                    }
                })
            );
        
        protected override void OnTextChanged(EventArgs e)
        {
            base.OnTextChanged(e);

            var moduleViewModel = DataContext as ModuleViewModel;
            if (moduleViewModel != null)
            {
                moduleViewModel.Content = Text;
            }

            RaisePropertyChanged("Text");
        }

        public event PropertyChangedEventHandler PropertyChanged;
        public void RaisePropertyChanged(string info)
        {
            if (PropertyChanged != null)
            {
                PropertyChanged(this, new PropertyChangedEventArgs(info));
            }
        }

        private new ToolTip ToolTip { get; set; }
        private ModuleViewModel moduleViewModel = null;
        public ModuleViewModel ModuleViewModel { get { return moduleViewModel; } set { moduleViewModel = value; ErrorColorizer.ModuleViewModel = value; } }
        private ErrorColorizer ErrorColorizer { get; set; }

        public MvvmTextEditor()
        {
            ToolTip = new ToolTip();
            ErrorColorizer = new ErrorColorizer();

            TextArea.TextView.LineTransformers.Add(ErrorColorizer);

            MouseHover += TextEditorMouseHover;
            MouseHoverStopped += TextEditorMouseHoverStopped;
        }

        void TextEditorMouseHover(object sender, MouseEventArgs e)
        {
            if (ModuleViewModel == null)
            {
                return;
            }

            var pos = GetPositionFromPoint(e.GetPosition(this));
            if (pos != null)
            {
                var line = pos.Value.Line;
                var column = pos.Value.Column - 1;

                foreach (var error in ModuleViewModel.SyntaxErrors)
                {
                    if (IsInsideError(line, column, error))
                    {
                        ShowError(pos.Value, error);
                        e.Handled = true;
                        return;
                    }
                }

                foreach (var error in ModuleViewModel.StaticAnalysisErrors)
                {
                    if (IsInsideError(line, column, error))
                    {
                        ShowError(pos.Value, error);
                        e.Handled = true;
                        return;
                    }
                }

                e.Handled = true;
            }
        }
        
        private bool IsInsideError(int line, int column, ErrorDesc error)
        {
            if (error.EndLine < 0 || error.EndColumn < 0)
            {
                return error.StartLine == line && error.StartColumn <= column;
            }
            else if (error.StartLine <= line && error.EndLine >= line)
            {
                var startColumn = error.StartLine == line ? error.StartColumn : 0;
                var endColumn = error.EndLine == line ? error.EndColumn : column;
                return startColumn <= column && endColumn >= column;
            }
            return false;
        }

        private void ShowError(TextViewPosition pos, ErrorDesc error)
        {
            ToolTip.PlacementTarget = this;
            ToolTip.Content = pos.ToString();
            ToolTip.IsOpen = true;
            ToolTip.Content = String.Format("#{0}. {1}", error.Id, error.Description);
        }

        void TextEditorMouseHoverStopped(object sender, MouseEventArgs e)
        {
            ToolTip.IsOpen = false;
        }
    }

    class ErrorColorizer : DocumentColorizingTransformer
    {
        public ModuleViewModel ModuleViewModel { get; set; }

        protected override void ColorizeLine(DocumentLine line)
        {
            if (ModuleViewModel == null)
            {
                return;
            }

            foreach (var error in ModuleViewModel.SyntaxErrors)
            {
                ApplyError(line, error);
            }

            foreach (var error in ModuleViewModel.StaticAnalysisErrors)
            {
                ApplyError(line, error);
            }
        }

        private void ApplyError(DocumentLine line, ErrorDesc error)
        {
            ApplyError(line, error, (VisualLineElement element) =>
            {
                var underline = new TextDecoration() { Pen = App.Current.Resources["ErrorPen"] as Pen, PenThicknessUnit = TextDecorationUnit.FontRecommended };
                var decorations = new TextDecorationCollection();
                decorations.Add(underline);
                element.TextRunProperties.SetTextDecorations(decorations);
            });
        }

        private void ApplyError(DocumentLine line, ErrorDesc error, Action<VisualLineElement> applyChanges)
        {
            if (error == null || error.StartLine <= 0 || error.StartColumn < 0)
            {
                return;
            }

            if (error.EndLine <= 0 || error.EndColumn < 0)
            {
                if (line.LineNumber == error.StartLine && error.StartColumn < line.Length)
                {
                    ChangeLinePart(line.Offset + error.StartColumn, line.EndOffset, applyChanges);
                }
            }
            else if (error.StartLine <= line.LineNumber && error.EndLine >= line.LineNumber)
            {
                var startColumn = error.StartLine == line.LineNumber ? Math.Min(error.StartColumn, line.Length > 0 ? line.Length - 1 : 0) : 0;
                var endColumn = error.EndLine == line.LineNumber ? Math.Min(error.EndColumn + 1, line.Length) : line.Length;
                ChangeLinePart(line.Offset + startColumn, line.Offset + endColumn, applyChanges);
            }
        }
    }
}
