﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures;

namespace FourqlModeler.Utility
{
    public class ExprUtils
    {
        public static Expression WithReplacements(Expression expr, IDictionary<string, DomainType> valueTypes, IDictionary<string, object> values)
        {
            return expr.Accept(new ExprWalker() { ValueTypes = valueTypes, Values = values });
        }
    }

    internal class ExprWalker : Expression.Visitor<Expression>
    {
        public IDictionary<string, DomainType> ValueTypes { get; set; }
        public IDictionary<string, object> Values { get; set; }

        #region Visitor<Expression> Members

        public Expression visitLiteral(Literal literal)
        {
            var copy = new Literal() { RelationName = literal.RelationName, ParserContext = literal.ParserContext };
            foreach (var value in literal.Values)
            {
                if (value.Type == DomainType.VARIABLE)
                {
                    var variable = value.valueAs<string>();
                    if (Values.ContainsKey(variable) && ValueTypes.ContainsKey(variable))
                    {
                        copy.Values.Add(new ValueHolder(Values[variable], ValueTypes[variable]));
                        continue;
                    }
                }
                copy.Values.Add(value);
            }
            return copy;
        }

        public Expression visitNeg(NegExpression negExpr)
        {
            return new NegExpression() { Next = negExpr.Next.Accept(this), ParserContext = negExpr.ParserContext };
        }

        public Expression visitModule(ModuleExpression moduleExpr)
        {
            return new ModuleExpression() { Next = moduleExpr.Next.Accept(this), ParserContext = moduleExpr.ParserContext, ModuleName = moduleExpr.ModuleName, InValues = moduleExpr.InValues, IsNegative = moduleExpr.IsNegative };
        }

        public Expression visitAnd(AndExpression andExpr)
        {
            var copy = new AndExpression() { ParserContext = andExpr.ParserContext };
            foreach (var next in andExpr.Nexts)
            {
                copy.Nexts.Add(next.Accept(this));
            }
            return copy;
        }

        public Expression visitOr(OrExpression orExpr)
        {
            var copy = new OrExpression() { ParserContext = orExpr.ParserContext };
            foreach (var next in orExpr.Nexts)
            {
                copy.Nexts.Add(next.Accept(this));
            }
            return copy;
        }

        public Expression visitExists(ExistsExpression exExpr)
        {
            // do not substitute quantified variables!!!
            var valuesCopy = Values;
            
            Values = new Dictionary<string, object>(valuesCopy);
            Values.Remove(exExpr.VariableName);
            var res = new ExistsExpression()
            {
                Next = exExpr.Next.Accept(this),
                ParserContext = exExpr.ParserContext,
                VariableDomain = exExpr.VariableDomain,
                DomainModule = exExpr.DomainModule,
                VariableName = exExpr.VariableName
            };
            Values = valuesCopy;
            return res;
        }

        public Expression visitAll(AllExpression allExpr)
        {
            // do not substitute quantified variables!!!
            var valuesCopy = Values;
            Values = new Dictionary<string, object>(valuesCopy);
            Values.Remove(allExpr.VariableName);
            var res =  new AllExpression()
            {
                Next = allExpr.Next.Accept(this),
                ParserContext = allExpr.ParserContext,
                VariableName = allExpr.VariableName,
                DomainModule = allExpr.DomainModule,
                VariableDomain = allExpr.VariableDomain
            };
            Values = valuesCopy;
            return res;
        }

        #endregion
    }
}
