﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Antlr4.Runtime;
using FourqlModeler.ViewModels;
using FourqlParser.Parsers;
using FourqlParser.StaticAnalysis;
using FourqlParser.Structures;

namespace FourqlModeler.Utility
{
    // FIXME OPTIMIZE IN TERMS OF MODULE RELATIONS AND RECALCULATIONS!
    class ModuleUtils
    {
        /// <summary>
        /// Updates the module view model by re-parsing its <code>Content</code> and applying its details to the model.
        /// </summary>
        /// <param name="moduleViewModel">model to update</param>
        public static void updateWithParsedContent(ModuleViewModel moduleViewModel)
        {
            // Clear the errors before the update run.
            ClearStructures(moduleViewModel);
            // Check syntax.
            ParserDispatcher.Errors errors;
            moduleViewModel.CurrentProgram = ModuleUtils.readAndParseProgram(moduleViewModel.Content, out errors);
            if (errors.ErrorCount > 0)
            {
                foreach (var error in errors.LexerErrors)
                {
                    addSyntaxError(moduleViewModel, error);
                }

                foreach (var error in errors.ParserErrors)
                {
                    addSyntaxError(moduleViewModel, error);
                }
            }
        }

        private static void addSyntaxError<T>(ModuleViewModel moduleViewModel, ParserDispatcher.Error<T> error)
        {
            moduleViewModel.SyntaxErrors.Add(new ErrorDesc() { ModelSource = moduleViewModel, ModuleName = "", StartLine = error.Line, EndLine = -1, StartColumn = error.Column, EndColumn = -1, Category = "Syntax Error", Description = error.Message });
        }

        public static Program readAndParseProgram(string toParse, out ParserDispatcher.Errors errors)
        {
            errors = new ParserDispatcher.Errors();
            var programContext = ParserDispatcher.parseProgram(toParse, errors, null);
            if (errors.ErrorCount > 0)
            {
                // Return nothing, only the errors.
                return null;
            }

            return ParserDispatcher.toProgram(programContext);
        }

        public static IDictionary<string, ModuleViewModel> updateAndGroupModuleViewModels(ICollection<ModuleViewModel> moduleViewModels)
        {
            // Get modules that parsed properly.
            var moduleViewModelsWithoutErrors = moduleViewModels.Where(model => model.SyntaxErrors.Count == 0 && model.CurrentProgram != null);
            // Clear module errors.
            ClearStructures(moduleViewModelsWithoutErrors);
            // Group models by names of modules that they contain.
            var modulesByNames = getModulesByNames(moduleViewModelsWithoutErrors);
            // Remember the groups of modules that are in more than one model to push errors in the process.
            var duplicatedModules = modulesByNames.Where(group => group.Value.Count > 1).ToList();
            // Remember the groups of modules that are unique in the whole specturm of models.
            var nonDuplicatedModules = modulesByNames.Where(group => group.Value.Count == 1).Select(entry => new KeyValuePair<string, ModuleViewModel>(entry.Key, entry.Value.First())).ToDictionary(item => item.Key, item => item.Value);
            // Add errors for the duplicated modules.
            var groupId = 1;
            var errorId = pushErrorsForDuplicates(duplicatedModules, ref groupId);
            // Check non-duplicaed modules and add those that are valid and have valid references.
            validateModules(nonDuplicatedModules, new HashSet<string>(duplicatedModules.Select(module => module.Key)), errorId, groupId);
            // Return the modules as they are.
            return nonDuplicatedModules;
        }

        private static void ClearStructures(IEnumerable<ModuleViewModel> moduleViewModelsWithoutErrors)
        {
            foreach (var moduleViewModel in moduleViewModelsWithoutErrors)
            {
                ClearStructures(moduleViewModel);
            }
        }

        private static void ClearStructures(ModuleViewModel moduleViewModel)
        {
            moduleViewModel.SyntaxErrors.Clear();
            moduleViewModel.StaticAnalysisErrors.Clear();
            moduleViewModel.ModuleStaticAnalysisErrors.Clear();
            if (moduleViewModel.CurrentProgram != null)
            {
                foreach (var module in moduleViewModel.CurrentProgram.Modules)
                {
                    module.InvalidDefinitions = new ModuleInvalidDefinitions();
                }
            }
        }

        private static void validateModules(IDictionary<string, ModuleViewModel> modules, ICollection<string> duplicatedModuleNames, int errorId, int groupId)
        {
            var analysis = new ProgramAnalysisGatherErrors();
           
            var moduleMap = new Dictionary<string, Module>();
            foreach (var moduleEntry in modules)
            {
                var module = moduleEntry.Value.CurrentProgram.FindModules(moduleEntry.Key).First();
                moduleMap.Add(module.Name, module);
            }
            // check domain imports
            foreach (var module in moduleMap.Values)
            {
                analysis.checkDomainImports(module, moduleMap);
            }

            // check the definitions - domains and relations.
            foreach (var moduleEntry in modules)
            {
                analysis.checkForInvalidDefinitions(moduleEntry.Value.CurrentProgram.FindModules(moduleEntry.Key).First());
            }

            // Secondly, we run expression analysis as we know which module definitions are invalid and can also use that data.
            analysis.checkExpressions(moduleMap, duplicatedModuleNames);
            // Add errors to models.
            foreach (var moduleEntry in modules)
            {
                errorId = addErrorsToModel(moduleEntry.Value, moduleEntry.Value.CurrentProgram.FindModules(moduleEntry.Key).First(), errorId, ref groupId);
            }
            // Lastly, check for cycles and remember those problems, so we are not allowed to add the cycled modules into interpreter memory.
            var stronglyConnectedComponents = new TarjansAlgorithm().run(moduleMap);
            // Add those components that consist of two or more modules.
            foreach (var component in stronglyConnectedComponents)
            {
                // Worrisome strongly connected components are those that consist of more than one module to build a cycle that is not 1 -> 1.
                if (component.Count < 2)
                {
                    continue;
                }
                // Build an error message.
                var builder = new StringBuilder("Found a cycle of modules that consists of: ");
                var first = true;
                foreach (var module in component)
                {
                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        builder.Append(", ");
                    }
                    builder.Append(module.Name);
                }
                // Remember the error message and add it for every source.
                var error = builder.ToString();
                foreach (var module in component)
                {
                    addError(modules[module.Name], module, errorId++, groupId, module, "Module Cycle Error", error);
                }
                // Increase the error identifier.
                ++groupId;
            }
        }

        private static int addErrorsToModel(ModuleViewModel moduleViewModel, Module module, int errorId, ref int groupId)
        {
            // Add errors for duplicated domains.
            foreach (var domainEntry in module.InvalidDefinitions.DuplicatedDomains)
            {
                foreach (var source in domainEntry.Value)
                {
                    addError(moduleViewModel, module, errorId++, groupId, source, "Duplicated Domain Error", String.Format("Domain '{0}' in module '{1}' is defined more than once.", source.Name, module.Name));
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Add errors for domain definitions.
            foreach (var domainEntry in module.InvalidDefinitions.InvalidDomains)
            {
                foreach (var source in domainEntry.Value)
                {
                    addError(moduleViewModel, module, errorId++, groupId, source.ErrorSource, "Domain import definition Error", source.Message);
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Add errors for duplicated relations.
            foreach (var relationEntry in module.InvalidDefinitions.DuplicatedRelations)
            {
                foreach (var source in relationEntry.Value)
                {
                    addError(moduleViewModel, module, errorId++, groupId, source, "Duplicated Relation Error", String.Format("Relation '{0}' in module '{1}' is defined more than once.", source.Name, module.Name));
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Add errors for relation definitions.
            foreach (var relationEntry in module.InvalidDefinitions.InvalidRelations)
            {
                foreach (var source in relationEntry.Value)
                {
                    addError(moduleViewModel, module, errorId++, groupId, source.ErrorSource, "Relation Error", source.Message);
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Add errors for rule definitions.
            foreach (var ruleEntry in module.InvalidDefinitions.InvalidRules)
            {
                foreach (var source in ruleEntry.Value)
                {
                    addError(moduleViewModel, module, errorId++, groupId, source.ErrorSource, "Rule Error", source.Message);
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Add errors for fact definitions.
            foreach (var factEntry in module.InvalidDefinitions.InvalidFacts)
            {
                foreach (var source in factEntry.Value)
                {
                    addError(moduleViewModel, module, errorId++, groupId, source.ErrorSource, "Fact Error", source.Message);
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Return the new current error identifier.
            return errorId;
        }

        private static ErrorDesc addError(ModuleViewModel moduleViewModel, Module module, int errorId, int groupId, IOutputSource outputSource, string category, string description, bool addToModelErrors = true)
        {
            var error = new ErrorDesc()
            {
                Id = errorId,
                GroupId = groupId,
                ModelSource = moduleViewModel,
                ModuleName = module.Name,
                StartLine = outputSource.startLine(),
                EndLine = outputSource.endLine(),
                StartColumn = outputSource.startColumn(),
                EndColumn = outputSource.endColumn(),
                Category = category,
                Description = description
            };

            addError(moduleViewModel, module, error, addToModelErrors);
            return error;
        }

        private static void addError(ModuleViewModel moduleViewModel, Module module, ErrorDesc error, bool addToModelErrors = true)
        {
            moduleViewModel.StaticAnalysisErrors.Add(error);
            if (!moduleViewModel.ModuleStaticAnalysisErrors.ContainsKey(module.Name))
            {
                moduleViewModel.ModuleStaticAnalysisErrors.Add(module.Name, new ObservableCollection<ErrorDesc>());
            }
            moduleViewModel.ModuleStaticAnalysisErrors[module.Name].Add(error);
        }

        private static int pushErrorsForDuplicates(IEnumerable<KeyValuePair<string, ICollection<ModuleViewModel>>> duplicatedModules, ref int groupId)
        {
            // Fill module view models with errors for the duplicated modules.
            int errorId = groupId;
            foreach (var duplicatedModule in duplicatedModules)
            {
                // Prepare a builder for the group.
                var builder = new StringBuilder("Duplicated definition of module '");
                builder.Append(duplicatedModule.Key);
                builder.Append("' in:");
                // Initialize a module source identifier.
                var id = 0;
                var handledModels = new HashSet<ModuleViewModel>();
                foreach (var duplicatedModuleModel in duplicatedModule.Value)
                {
                    // We visit the model for the first time in the context?
                    if (handledModels.Add(duplicatedModuleModel))
                    {
                        // Yes - append the source data.
                        foreach (var module in duplicatedModuleModel.CurrentProgram.FindModules(duplicatedModule.Key))
                        {
                            builder.Append("\n  ");
                            builder.Append(++id);
                            builder.Append(". ");
                            builder.Append(duplicatedModuleModel.FileDir + duplicatedModuleModel.FileName);
                            // Try to append approximate coordinates of the module.
                            builder.Append(String.Format(" at {0}:{1}", module.startLine(), module.startColumn()));
                        }
                    }
                }
                // We gathered the error data - append to the models.
                var error = builder.ToString();
                handledModels.Clear();
                foreach (var duplicatedModuleModel in duplicatedModule.Value)
                {
                    // We might have already handled the model as it might have contained multiple copies of the same module.
                    if (handledModels.Contains(duplicatedModuleModel))
                    {
                        continue;
                    }
                    // Add the error data to the model for every module entry in it.
                    foreach (var module in duplicatedModuleModel.CurrentProgram.FindModules(duplicatedModule.Key))
                    {
                        addError(duplicatedModuleModel, module, errorId++, groupId, module, "Duplicated Module Error", error);
                    }
                    // Add the model to the visited set.
                    handledModels.Add(duplicatedModuleModel);
                }
                // Increase the error group identifier.
                ++groupId;
            }
            // Return the new current error identifier.
            return errorId;
        }

        private static IDictionary<string, ICollection<ModuleViewModel>> getModulesByNames(IEnumerable<ModuleViewModel> moduleViewModelsWithoutErrors)
        {
            var moduleGroups = new Dictionary<string, ICollection<ModuleViewModel>>();
            foreach (var moduleViewModel in moduleViewModelsWithoutErrors)
            {
                foreach (var module in moduleViewModel.CurrentProgram.Modules)
                {
                    if (!moduleGroups.ContainsKey(module.Name))
                    {
                        moduleGroups.Add(module.Name, new List<ModuleViewModel>());
                    }
                    moduleGroups[module.Name].Add(moduleViewModel);
                }
            }
            return moduleGroups;
        }
    }
}
