﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlModeler.ViewModels;
using FourqlParser.StaticAnalysis;
using FourqlParser.Structures;

namespace FourqlModeler.Utility
{
    /// <summary>
    /// The class that builds <code>ReasoningOrder</code> which represents a graph
    /// of modules with additional information on if they can be used and what they
    /// depend on.
    /// </summary>
    public class ReasoningOrderBuilder
    {
        /// <summary>
        /// Maps module names to corresponing node
        /// </summary>
        private IDictionary<string, ReasoningNode> ModuleNodes = new Dictionary<string, ReasoningNode>();
       /// <summary>
       /// For each module name holds a map of rules and literals from different modules the rule depends on
       /// </summary>
        private IDictionary<string, IDictionary<Rule, IDictionary<string, ICollection<string>>>> ModuleRuleDependencies = new Dictionary<string, IDictionary<Rule, IDictionary<string, ICollection<string>>>>();
       /// <summary>
       /// For each module name holds a map of domains and domains from differnet modules the rule depends on
       /// </summary>
        private IDictionary<string, IDictionary<Domain, IDictionary<string, ICollection<string>>>> ModuleDomainDependencies = new Dictionary<string, IDictionary<Domain, IDictionary<string, ICollection<string>>>>();
        private ICollection<ModuleViewModel> ModulesWithSyntaxErrors = new List<ModuleViewModel>();
        private IDictionary<string, ICollection<ModuleViewModel>> DuplicatedModules = new Dictionary<string, ICollection<ModuleViewModel>>();

        public ReasoningOrderBuilder AddModel(ModuleViewModel moduleViewModel)
        {
            if (moduleViewModel.SyntaxErrors.Count > 0)
            {
                ModulesWithSyntaxErrors.Add(moduleViewModel);
            }
            else if (moduleViewModel.CurrentProgram != null)
            {
                foreach (var module in moduleViewModel.CurrentProgram.Modules)
                {
                    if (DuplicatedModules.ContainsKey(module.Name))
                    {
                        DuplicatedModules[module.Name].Add(moduleViewModel);
                    }
                    else if (ModuleNodes.ContainsKey(module.Name))
                    {
                        DuplicatedModules.Add(module.Name, new List<ModuleViewModel>() { ModuleNodes[module.Name].ModuleViewModel, moduleViewModel });
                        ModuleNodes.Remove(module.Name);
                        ModuleRuleDependencies.Remove(module.Name);
                        ModuleDomainDependencies.Remove(module.Name);
                    }
                    else
                    {
                        ModuleNodes.Add(module.Name, new ReasoningNode() { Group = 0, ModuleViewModel = moduleViewModel, Module = module });
                        ModuleRuleDependencies.Add(module.Name, GetModuleDeps(module));
                        ModuleDomainDependencies.Add(module.Name, GetModuleDomainDeps(module));
                    }
                }
            }
            return this;
        }

        private IDictionary<Domain, IDictionary<string, ICollection<string>>> GetModuleDomainDeps(Module module)
        {
            var deps = new Dictionary<Domain, IDictionary<string, ICollection<string>>>();
            foreach (var domain in module.Domains)
            {
                var domainDeps = new Dictionary<string, ICollection<string>> ();
                foreach (var import in domain.Imports)
                {
                    if (!domainDeps.ContainsKey(import.Item1))
                        domainDeps.Add(import.Item1, new List<string>());
                    domainDeps[import.Item1].Add(import.Item2);
                }
                deps.Add(domain, domainDeps);
            }
            return deps;
        }

        private IDictionary<Rule, IDictionary<string, ICollection<string>>> GetModuleDeps(Module module)
        {
            var deps = new Dictionary<Rule, IDictionary<string, ICollection<string>>>();
            foreach (var rule in module.Rules)
            {
                var ruleDeps = new ModuleReferenceFinder().GetReferences(module.Name, rule.Body);
                deps.Add(rule, ruleDeps);
            }
            return deps;
        }

        public ReasoningOrder Done()
        {
            // For every module node...
            foreach (var moduleNode in ModuleNodes.Values)
            {
                // ... we add the modules the module depends on.
                //grab the domain dependencies
                foreach (var domainDependencies in ModuleDomainDependencies[moduleNode.Module.Name])
                {
                    // ... and for each we add module domin dependencies of modules the domain depends on.
                    foreach (var moduleDependency in domainDependencies.Value)
                    {
                        // Find the node of the rule dependency.
                        var dependencyNode = ModuleNodes.ContainsKey(moduleDependency.Key) ? ModuleNodes[moduleDependency.Key] : null;
                        if (dependencyNode != null)
                        {
                            // We found it - it means we have a parsed module of the dependency.
                            if (!moduleNode.DependsOn.ContainsKey(dependencyNode))
                            {
                                // It is our first encounter of the module dependency - add it the the dependency map.
                                moduleNode.DependsOn.Add(dependencyNode, new ReasoningDependency());
                            }
                            // Now add the dependencies.
                            moduleNode.DependsOn[dependencyNode].DomainDependency.Add(domainDependencies.Key, moduleDependency.Value);
                            dependencyNode.IsDependencyOf.Add(moduleNode);
                        }
                        else
                        {
                            // We did not find the module node - meaning it's an invalid node.
                            if (!moduleNode.DependsOnInvalid.ContainsKey(moduleDependency.Key))
                            {
                                // It is our first encounter of the module dependency - add it the the dependency map.
                                moduleNode.DependsOnInvalid.Add(moduleDependency.Key, new ReasoningDependency());
                            }
                            // Now add the dependencies.
                            moduleNode.DependsOnInvalid[moduleDependency.Key].DomainDependency.Add(domainDependencies.Key, moduleDependency.Value);
                        }
                    }
                }

                //grab the module rule dependencies...
                foreach (var ruleDependencies in ModuleRuleDependencies[moduleNode.Module.Name])
                {
                    // ... and for each we add module relation dependencies of modules the rule depends on.
                    foreach (var moduleDependency in ruleDependencies.Value)
                    {
                        // Find the node of the rule dependency.
                        var dependencyNode = ModuleNodes.ContainsKey(moduleDependency.Key) ? ModuleNodes[moduleDependency.Key] : null;
                        if (dependencyNode != null)
                        {
                            // We found it - it means we have a parsed module of the dependency.
                            if (!moduleNode.DependsOn.ContainsKey(dependencyNode))
                            {
                                // It is our first encounter of the module dependency - add it the the dependency map.
                                moduleNode.DependsOn.Add(dependencyNode, new ReasoningDependency());
                            }
                            // Now add the dependencies.
                            moduleNode.DependsOn[dependencyNode].RuleDependency.Add(ruleDependencies.Key, moduleDependency.Value);
                            dependencyNode.IsDependencyOf.Add(moduleNode);
                        }
                        else
                        {
                            // We did not find the module node - meaning it's an invalid node.
                            if (!moduleNode.DependsOnInvalid.ContainsKey(moduleDependency.Key))
                            {
                                // It is our first encounter of the module dependency - add it the the dependency map.
                                moduleNode.DependsOnInvalid.Add(moduleDependency.Key, new ReasoningDependency());
                            }
                            // Now add the dependencies.
                            moduleNode.DependsOnInvalid[moduleDependency.Key].RuleDependency.Add(ruleDependencies.Key, moduleDependency.Value);
                        }
                    }
                }
            }
            UpdateErroneousState();
            UpdateGroupState();
            return new ReasoningOrder() { ModuleNodes = ModuleNodes, ModulesWithSyntaxErrors = ModulesWithSyntaxErrors, DuplicatedModules = DuplicatedModules };
        }

        private void UpdateErroneousState()
        {
            // Set whether the module node contains errors or depends on module nodes that are erroneous.
            // We start by filling the field with independent properties only.
            foreach (var moduleNode in ModuleNodes.Values)
            {
                var errors = moduleNode.ModuleViewModel.ModuleStaticAnalysisErrors.ContainsKey(moduleNode.Module.Name) ? moduleNode.ModuleViewModel.ModuleStaticAnalysisErrors[moduleNode.Module.Name] : null;
                // The module node is erroneous when the module that it represents has any semantical errors in it.
                // It is also incorrect if it is a part of a cycle.
                // It might also be a duplicate and that also implies being incorrect.
                moduleNode.ErroneousDependencies = (errors != null && errors.Count > 0) || moduleNode.Group != 0 || moduleNode.DependsOnInvalid.Count > 0;
            }

            // We now mark all nodes that depend on the incorrect ones.
            var visitedNodes = new HashSet<ReasoningNode>();
            var nodeQueue = new HashSet<ReasoningNode>();
            // Prepare incorrect nodes to walk through.
            foreach (var errorNode in ModuleNodes.Values.Where(node => node.ErroneousDependencies))
            {
                nodeQueue.Add(errorNode);
            }
            while (nodeQueue.Count > 0)
            {
                // Grab the node with errors.
                var errorNode = nodeQueue.First();
                // Remove it from the queue.
                nodeQueue.Remove(errorNode);
                // Add it to the already handled ones, so we do not repeat steps for the node.
                visitedNodes.Add(errorNode);
                // Now mark every node that depends on it as erroneous.
                foreach (var dependentNode in errorNode.IsDependencyOf)
                {
                    dependentNode.ErroneousDependencies = true;
                    // Also, if dependent nodes have never been visited, add them to the queue.
                    if (!visitedNodes.Contains(dependentNode))
                    {
                        nodeQueue.Add(dependentNode);
                    }
                }
            }
        }

        private void UpdateGroupState()
        {
            var allModules = ModuleNodes.Values.ToDictionary(entry => entry.Module.Name, entry => entry.Module);
            var components = new TarjansAlgorithm().run(allModules);
            var group = 0;
            foreach (var component in components)
            {
                var componentGroup = component.Count < 2 ? 0 : ++group;
                foreach (var item in component)
                {
                    ModuleNodes[item.Name].Group = group;
                }
            }
        }
    }

    public class ReasoningOrder
    {
        public IDictionary<string, ReasoningNode> ModuleNodes { get; set; }
        public ICollection<ModuleViewModel> ModulesWithSyntaxErrors { get; set; }
        public IDictionary<string, ICollection<ModuleViewModel>> DuplicatedModules { get; set; }
    }

    public class ReasoningDependency
    {
        public IDictionary<Rule, ICollection<string>> RuleDependency { get; set; }

        public IDictionary<Domain, ICollection<string>> DomainDependency { get; set; }

        public ReasoningDependency() 
        {
            RuleDependency = new Dictionary<Rule, ICollection<string>>();
            DomainDependency = new Dictionary<Domain, ICollection<string>>();
        }
    }


    public class ReasoningNode
    {
        public int Group { get; set; }
        public ModuleViewModel ModuleViewModel { get; set; }
        public Module Module { get; set; }
        public IDictionary<ReasoningNode, ReasoningDependency> DependsOn { get; set; }
        public ICollection<ReasoningNode> IsDependencyOf { get; set; }
        public Dictionary<string, ReasoningDependency> DependsOnInvalid { get; set; }
        public bool ErroneousDependencies { get; set; }

        public ReasoningNode()
        {
            DependsOn = new Dictionary<ReasoningNode, ReasoningDependency>();
            IsDependencyOf = new HashSet<ReasoningNode>();
            DependsOnInvalid = new Dictionary<string, ReasoningDependency>();
        }
    }
}
