﻿using FourqlModeler.ViewModels;
using GraphX;
using GraphX.Controls;
using GraphX.GraphSharp.Algorithms.Layout.Compound.FDP;
using GraphX.GraphSharp.Algorithms.Layout.Simple.FDP;
using GraphX.GraphSharp.Algorithms.Layout.Simple.Hierarchical;
using GraphX.GraphSharp.Algorithms.Layout.Simple.Tree;
using GraphX.GraphSharp.Algorithms.OverlapRemoval;
using GraphX.Models;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Drawing;
using FourqlModeler.Views.GraphExtensions;
using FourqlModeler.Views.GraphExtensions.ModuleGraph;
using FourqlModeler.Controls;
using Microsoft.Win32;
using Parago.Windows;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using System.Xml;
using System.Xml.Linq;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;

namespace FourqlModeler.Views
{
    /// <summary>
    /// Interaction logic for ExtendedGraphView.xaml
    /// </summary>
    public partial class ExtendedGraphView : UserControl
    {
        public bool IsGraphInitialized { get; set; }
        public ExtendedGraph BaseGraph { get; set; }

        public ExtendedGraphView(ExtendedGraph graph) : this()
        {
            BaseGraph = graph;
        }

        public ExtendedGraphView()
        {
            InitializeComponent();
            AreaSetup();
            Loaded += WSMView_Loaded;
        }

        void WSMView_Loaded(object sender, RoutedEventArgs e)
        {
            RebuildGraph();
        }

        public void RebuildGraph()
        {
            if (!IsGraphInitialized && BaseGraph != null)
            {
                CreateGraph(BaseGraph);
                IsGraphInitialized = true;
            }
        }

        public void AreaSetup()
        {
            // Make the zoom control visible.
            ZoomControl.SetViewFinderVisibility(ZoomCtrl, System.Windows.Visibility.Visible);

            // Set the area's layout to its default state.
            SetAdditionalLayoutData();
        }

        private void SetAdditionalLayoutData()
        {
            GraphArea.SetEdgesDashStyle(GraphX.EdgeDashStyle.Solid);
            GraphArea.ShowAllEdgesArrows(true);
            GraphArea.SetVerticesDrag(true, true);

            // Set additional data.
            useLayoutAlgorithmEfficientSugiyama(null, null);

            // Set animations.
            GraphArea.MoveAnimation = AnimationFactory.CreateMoveAnimation(MoveAnimation.Move, TimeSpan.FromSeconds(0.5));
            GraphArea.DeleteAnimation = AnimationFactory.CreateDeleteAnimation(DeleteAnimation.Shrink);
            GraphArea.MouseOverAnimation = AnimationFactory.CreateMouseOverAnimation(MouseOverAnimation.Scale);
            GraphArea.MoveAnimation.Completed += MoveAnimation_Completed;
        }

        void MoveAnimation_Completed(object sender, EventArgs e)
        {
            ZoomCtrl.ZoomToFill();
        }

        public void CreateGraph(ExtendedGraph graph)
        {
            GraphArea.PrototypeGraph = graph;
            GraphArea.GenerateGraph(graph, true);
            ZoomCtrl.ContextMenu = CreateContextMenuForGraphArea();
            ZoomCtrl.ZoomToFill();

            GraphArea.UpdateControls();
            GraphArea.ShowAllEdgesLabels(true);
        }
        
        protected virtual ContextMenu CreateContextMenuForGraphArea()
        {
            var menu = ContextMenu != null ? ContextMenu : new ContextMenu();

            // Dumps.
            var item = new MenuItem() { Header = "Export graph as image..." };
            item.Click += ExportToImage;
            menu.Items.Add(item);

            // Layout algorithms.
            menu.Items.Add(new Separator());

            // Layout Algorithm items.
            item = new MenuItem() { Header = "Use layout algorithm: SimpleRandom" };
            item.Click += useLayoutAlgorithmSimpleRandom;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: Tree" };
            menu.Items.Add(item);

            var subitem = new MenuItem() { Header = "Direction: Top-Bottom" };
            subitem.Click += useLayoutAlgorithmTreeTB;
            item.Items.Add(subitem);

            subitem = new MenuItem() { Header = "Direction: Bottom-Top" };
            subitem.Click += useLayoutAlgorithmTreeBT;
            item.Items.Add(subitem);

            subitem = new MenuItem() { Header = "Direction: Left-Right" };
            subitem.Click += useLayoutAlgorithmTreeLR;
            item.Items.Add(subitem);

            subitem = new MenuItem() { Header = "Direction: Right-Left" };
            subitem.Click += useLayoutAlgorithmTreeRL;
            item.Items.Add(subitem);

            item = new MenuItem() { Header = "Use layout algorithm: KK" };
            item.Click += useLayoutAlgorithmKk;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: ISOM" };
            item.Click += useLayoutAlgorithmIsom;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: BoundFR" };
            item.Click += useLayoutAlgorithmBoundFr;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: LinLog" };
            item.Click += useLayoutAlgorithmLinLog;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: CompoundFDP" };
            item.Click += useLayoutAlgorithmCompoundFdp;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: Sugiyama" };
            item.Click += useLayoutAlgorithmSugiyama;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Use layout algorithm: Efficient Sugiyama" };
            item.Click += useLayoutAlgorithmEfficientSugiyama;
            menu.Items.Add(item);

            // Overlap Removal items.
            if (false)
            {
                // DISABLED: they do not change much.
                menu.Items.Add(new Separator());

                item = new MenuItem() { Header = "Use overlap removal algorithm: Default" };
                item.Click += useOverlapRemovalAlgorithmDefault;
                menu.Items.Add(item);

                item = new MenuItem() { Header = "Use overlap removal algorithm: FSA" };
                item.Click += useOverlapRemovalAlgorithmFsa;
                menu.Items.Add(item);

                item = new MenuItem() { Header = "Use overlap removal algorithm: One-Way FSA" };
                item.Click += useOverlapRemovalAlgorithmOneWayFsa;
                menu.Items.Add(item);
            }

            // Routing items.
            if (false)
            {
                // DISABLED: they break edges so heavily it's unbearable.
                menu.Items.Add(new Separator());

                item = new MenuItem() { Header = "Use edge routing algorithm: None" };
                item.Click += useEdgeRoutingAlgorithmNone;
                menu.Items.Add(item);

                item = new MenuItem() { Header = "Use edge routing algorithm: SimpleER" };
                item.Click += useEdgeRoutingAlgorithmSimpleEr;
                menu.Items.Add(item);

                item = new MenuItem() { Header = "Use edge routing algorithm: Bundling" };
                item.Click += useEdgeRoutingAlgorithmBundling;
                menu.Items.Add(item);

                item = new MenuItem() { Header = "Use edge routing algorithm: PathFinding" };
                item.Click += useEdgeRoutingAlgorithmPathFinding;
                menu.Items.Add(item);
            }

            // Edge items.
            if (false)
            {
                // DISABLES: does nothing.
                menu.Items.Add(new Separator());

                item = new MenuItem() { Header = "Set edge curving" };
                item.Click += setEdgeCurving;
                menu.Items.Add(item);
            }

            // Label items.
            menu.Items.Add(new Separator());

            item = new MenuItem() { Header = "Set labels invisible" };
            item.Click += setLabelVisibility;
            menu.Items.Add(item);

            item = new MenuItem() { Header = "Align edge labels" };
            item.Click += setLabelAlignment;
            menu.Items.Add(item);

            // Relayout.
            menu.Items.Add(new Separator());

            item = new MenuItem() { Header = "Relayout" };
            item.Click += relayoutGraph;
            menu.Items.Add(item);

            return menu;
        }

        #region Export To File

        private void ExportToImage(object sender, RoutedEventArgs e)
        {
            var dpiInputDialog = new DpiInputDialog();
            var result = dpiInputDialog.ShowDialog();
            if (result == null || !result.HasValue || result == false)
            {
                return;
            }

            var dpi = dpiInputDialog.Dpi;
            if (dpi == null || !dpi.HasValue || dpi <= 0)
            {
                MessageBox.Show("The given DPI is not a positive integer and only those are supported.", "Non-positive integer as DPI", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            string errorMessage = null;
            var saveProgressResult = ProgressDialog.Execute(App.Current.MainWindow, "Saving the graph as a bitmap to a file...", () =>
            {
                ProgressDialog.Current.Report("Saving the graph as a bitmap to a file...");
                Dispatcher.Invoke(() =>
                {
                    try
                    {
                        this.GraphArea.ExportAsImage(ImageType.PNG, true, dpi.Value);
                    }
                    catch (Exception ex)
                    {
                        errorMessage = ex.Message;
                    }
                });
            });

            if (saveProgressResult.OperationFailed || errorMessage != null)
            {
                MessageBox.Show("The save operation failed due to (DPI too high?): " + (errorMessage != null ? errorMessage : saveProgressResult.Error.Message), "Error while saving the graph.", MessageBoxButton.OK, MessageBoxImage.Error);
            }

            /*var bitmap = this.GraphArea.ExportToBitmap(dpi.Value);

            var saveDialog = new SaveFileDialog() { Title = "Select a file to write the bitmap to.", DefaultExt = ".png", Filter = "PNG Files (.png)|*.png", AddExtension = true, OverwritePrompt = true, CheckPathExists = true };
            var saveResult = saveDialog.ShowDialog();
            if (saveResult.HasValue && saveResult == true)
            {
                var saveProgressResult = ProgressDialog.Execute(App.Current.MainWindow, "Saving the graph as a bitmap to a file...", () =>
                {
                    ProgressDialog.Current.Report("Saving the graph as a bitmap to a file...");
                    bitmap.Save(saveDialog.FileName);
                });

                if (saveProgressResult.OperationFailed)
                {
                    MessageBox.Show("The save operation failed due to: " + saveProgressResult.Error.Message, "Error while saving the graph.", MessageBoxButton.OK, MessageBoxImage.Error);
                }
            }*/
        }
        
        #endregion Export To File

        #region Layout Algorithms

        private void useLayoutAlgorithmSimpleRandom(object sender, RoutedEventArgs e)
        {
            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.SimpleRandom;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = null;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmTreeTB(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.Tree) as SimpleTreeLayoutParameters;
            parameters.LayerGap = 50.0;
            parameters.VertexGap = 50.0;
            parameters.Direction = GraphX.GraphSharp.Algorithms.Layout.LayoutDirection.TopToBottom;

            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.Tree;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmTreeBT(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.Tree) as SimpleTreeLayoutParameters;
            parameters.LayerGap = 50.0;
            parameters.VertexGap = 50.0;
            parameters.Direction = GraphX.GraphSharp.Algorithms.Layout.LayoutDirection.BottomToTop;

            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.Tree;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmTreeLR(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.Tree) as SimpleTreeLayoutParameters;
            parameters.LayerGap = 50.0;
            parameters.VertexGap = 50.0;
            parameters.Direction = GraphX.GraphSharp.Algorithms.Layout.LayoutDirection.LeftToRight;

            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.Tree;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmTreeRL(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.Tree) as SimpleTreeLayoutParameters;
            parameters.LayerGap = 50.0;
            parameters.VertexGap = 50.0;
            parameters.Direction = GraphX.GraphSharp.Algorithms.Layout.LayoutDirection.RightToLeft;

            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.Tree;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmKk(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.KK) as KKLayoutParameters;
            
            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.KK;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmIsom(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.ISOM) as ISOMLayoutParameters;
            
            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.ISOM;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmBoundFr(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.BoundedFR) as BoundedFRLayoutParameters;
            
            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.BoundedFR;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmLinLog(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.LinLog) as LinLogLayoutParameters;
            
            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.LinLog;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmCompoundFdp(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.CompoundFDP) as CompoundFDPLayoutParameters;
            
            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.CompoundFDP;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmSugiyama(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.Sugiyama) as SugiyamaLayoutParameters;
            parameters.VerticalGap = 50;
            parameters.HorizontalGap = 50;

            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.Sugiyama;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useLayoutAlgorithmEfficientSugiyama(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateLayoutParameters(LayoutAlgorithmTypeEnum.EfficientSugiyama) as EfficientSugiyamaLayoutParameters;
            parameters.VertexDistance = 30;
            parameters.LayerDistance = 100;

            GraphArea.LogicCore.DefaultLayoutAlgorithm = LayoutAlgorithmTypeEnum.EfficientSugiyama;
            GraphArea.LogicCore.DefaultLayoutAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        #endregion Layout Algorithms

        #region Overlap Removal Algorithms

        private void useOverlapRemovalAlgorithmDefault(object sender, RoutedEventArgs e)
        {
            GraphArea.LogicCore.DefaultOverlapRemovalAlgorithm = GraphArea.LogicCore.DefaultOverlapRemovalAlgorithm;
            GraphArea.LogicCore.DefaultOverlapRemovalAlgorithmParams = GraphArea.LogicCore.DefaultOverlapRemovalAlgorithmParams;

            relayoutGraph(sender, e);
        }

        private void useOverlapRemovalAlgorithmFsa(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateOverlapRemovalParameters(OverlapRemovalAlgorithmTypeEnum.FSA) as OverlapRemovalParameters;
            parameters.HorizontalGap = 50;
            parameters.VerticalGap = 50;

            GraphArea.LogicCore.DefaultOverlapRemovalAlgorithm = OverlapRemovalAlgorithmTypeEnum.FSA;
            GraphArea.LogicCore.DefaultOverlapRemovalAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        private void useOverlapRemovalAlgorithmOneWayFsa(object sender, RoutedEventArgs e)
        {
            var parameters = GraphArea.LogicCore.AlgorithmFactory.CreateOverlapRemovalParameters(OverlapRemovalAlgorithmTypeEnum.OneWayFSA) as OverlapRemovalParameters;
            parameters.HorizontalGap = 50;
            parameters.VerticalGap = 50;

            GraphArea.LogicCore.DefaultOverlapRemovalAlgorithm = OverlapRemovalAlgorithmTypeEnum.OneWayFSA;
            GraphArea.LogicCore.DefaultOverlapRemovalAlgorithmParams = parameters;

            relayoutGraph(sender, e);
        }

        #endregion Overlap Removal Algorithms

        #region Edge Routing Algorithms

        private void useEdgeRoutingAlgorithmNone(object sender, RoutedEventArgs e)
        {
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithm = EdgeRoutingAlgorithmTypeEnum.None;
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithmParams = GraphArea.LogicCore.AlgorithmFactory.CreateEdgeRoutingParameters(EdgeRoutingAlgorithmTypeEnum.None);

            relayoutGraph(sender, e);
        }

        private void useEdgeRoutingAlgorithmSimpleEr(object sender, RoutedEventArgs e)
        {
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithm = EdgeRoutingAlgorithmTypeEnum.SimpleER;
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithmParams = GraphArea.LogicCore.AlgorithmFactory.CreateEdgeRoutingParameters(EdgeRoutingAlgorithmTypeEnum.SimpleER);

            relayoutGraph(sender, e);
        }

        private void useEdgeRoutingAlgorithmBundling(object sender, RoutedEventArgs e)
        {
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithm = EdgeRoutingAlgorithmTypeEnum.Bundling;
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithmParams = GraphArea.LogicCore.AlgorithmFactory.CreateEdgeRoutingParameters(EdgeRoutingAlgorithmTypeEnum.Bundling);

            relayoutGraph(sender, e);
        }

        private void useEdgeRoutingAlgorithmPathFinding(object sender, RoutedEventArgs e)
        {
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithm = EdgeRoutingAlgorithmTypeEnum.PathFinder;
            GraphArea.LogicCore.DefaultEdgeRoutingAlgorithmParams = GraphArea.LogicCore.AlgorithmFactory.CreateEdgeRoutingParameters(EdgeRoutingAlgorithmTypeEnum.PathFinder);

            relayoutGraph(sender, e);
        }

        #endregion Edge Routing Algorithms

        #region Edge Curving

        private void setEdgeCurving(object sender, RoutedEventArgs e)
        {
            var menuItem = e.OriginalSource as MenuItem;
            var curveEdges = !menuItem.Header.ToString().Equals("Unset edge curving");
            menuItem.Header = curveEdges ? "Unset edge curving" : "Set edge curving";

            GraphArea.LogicCore.EdgeCurvingEnabled = curveEdges;
            GraphArea.UpdateAllEdges();

            relayoutGraph(sender, e);
        }

        #endregion Edge Curving

        #region Label

        private void setLabelVisibility(object sender, RoutedEventArgs e)
        {
            var menuItem = e.OriginalSource as MenuItem;
            var setVisible = !menuItem.Header.ToString().Equals("Set labels invisible");
            menuItem.Header = setVisible ? "Set labels invisible" : "Set labels visible";

            GraphArea.ShowAllEdgesLabels(setVisible);
        }

        private void setLabelAlignment(object sender, RoutedEventArgs e)
        {
            var menuItem = e.OriginalSource as MenuItem;
            var alignToEdge = !menuItem.Header.ToString().Equals("Unalign edge labels");
            menuItem.Header = alignToEdge ? "Unalign edge labels" : "Align edge labels";

            GraphArea.AlignAllEdgesLabels(alignToEdge);
        }

        #endregion Label

        #region Relayout

        private void relayoutGraph(object sender, RoutedEventArgs e)
        {
            GraphArea.RelayoutGraph();
        }

        #endregion Relayout
    }
}
