﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlModeler.Calculcations;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using FourQLInterpreter.Reasoning.Explanation;

namespace FourqlModeler.Views.GraphExtensions.ExprGraph
{
    public class ExprGraphBuilder
    {

        public static ExtendedGraph Build(Expression expr, string execModule, Explanation explanation)
        {
            // Create an initial graph.
            var graph = new ExtendedGraph();
            // Proxy graph building to a walker.
            expr.Accept(new ExprWalker(execModule, graph, explanation));
            // Return the expression graph.
            return graph;
        }
    }

    internal class ExprWalker : Expression.Visitor<LogicValueVertex>
    {
        private string ExecModule { get; set; }
        private IEnumerator<LogicValue?> LiteralValueEnumerator { get; set; }
        private IEnumerator<LogicValue?> QuantifierValueEnumerator { get; set; }
        private IEnumerator<ValueHolder> QuantifierVariableEnumerator { get; set; }
        private ExtendedGraph Graph { get; set; }

        private IDictionary<string, ValueHolder> qantifiedVariableValues = new Dictionary<string, ValueHolder>(); 

        public ExprWalker(string execModule, ExtendedGraph graph, Explanation explanation)
        {
            ExecModule = execModule;
            LiteralValueEnumerator = explanation.LiteralValues.GetEnumerator();
            QuantifierValueEnumerator = explanation.QuantifierValues.GetEnumerator();
            QuantifierVariableEnumerator = explanation.QuantifierVariables.GetEnumerator();
            Graph = graph;
        }

        #region Visitor<ExtendedVertex> Members

        public LogicValueVertex visitLiteral(Literal literal)
        {
            // Move the enumerator to the proper logical value for the literal.
            LogicValue? value = LogicValue.UNKNOWN;
            if (LiteralValueEnumerator != null)
            {
                LiteralValueEnumerator.MoveNext();
                value = LiteralValueEnumerator.Current;
            }
            // Create a vertex for the literal and add it to the graph.
            var newLiteral = literal.Accept(new ExprValueSetter(qantifiedVariableValues)) as Literal ?? literal;
            var literalVertex = new LiteralVertex() { Pattern = String.Format("{0}.{1}", ExecModule, new ExprPrinter("", false).visitLiteral(newLiteral).ToString()), ModuleName = ExecModule, Literal = newLiteral, LogicValue = value };
            Graph.AddVertex(literalVertex);
            // Return the literal vertex.
            return literalVertex;
        }

        public LogicValueVertex visitNeg(NegExpression negExpr)
        {
            // Create a vertex for what is next.
            var nextVertex = negExpr.Next.Accept(this);
            // Create and add a negation vertex.
            var negVertex = new NegVertex() { Pattern = String.Format("neg({0})", ValuePattern(nextVertex.LogicValue)), LogicValue = (nextVertex.LogicValue == null ? (LogicValue?)null : Logic.not(nextVertex.LogicValue.GetValueOrDefault(LogicValue.UNKNOWN))) };
            Graph.AddVertex(negVertex);
            // Connect the negation vertex with the next vertex.
            var label = ValuePattern(nextVertex.LogicValue);
            Graph.AddEdge(new ExtendedEdge(negVertex, nextVertex) { Label = label, NotImportant = label=="?"});
            // Return the negation vertex.
            return negVertex;
        }

        public LogicValueVertex visitModule(ModuleExpression moduleExpr)
        {
            // Move calculations to the new module while keeping a backup for the current one.
            var backupExecModule = ExecModule;
            ExecModule = moduleExpr.ModuleName;
            // Create a vertex for what is next.
            var nextVertex = moduleExpr.Next.Accept(this);
            // See if negative.
            if (moduleExpr.IsNegative)
            {
                // It is - add a negative node between the module node and its next expression.
                var negVertex = new NegVertex() { Pattern = String.Format("neg({0})", ValuePattern(nextVertex.LogicValue)), LogicValue = (nextVertex.LogicValue == null ? (LogicValue?)null : Logic.not(nextVertex.LogicValue.GetValueOrDefault(LogicValue.UNKNOWN))) };
                Graph.AddVertex(negVertex);
                // Connect the negation vertex with the next vertex.
                var label = ValuePattern(nextVertex.LogicValue);
                Graph.AddEdge(new ExtendedEdge(negVertex, nextVertex) { Label = label, NotImportant = label == "?" });
                // Set the negation vertex as next.
                nextVertex = negVertex;
            }
            // Do we have an IN clause in the module expression?
            if (moduleExpr.InValues.Count > 0)
            {
                // Yes - that's the actual child of the module vertex. We have to create it and pin the nextVertex to it.
                // First of all, calculate the actual return value of the IN construction.
                var logicValue = moduleExpr.InValues.Any(inValue => inValue == nextVertex.LogicValue) ? LogicValue.TRUE : LogicValue.FALSE;
                // Create a pattern for the in construction.
                var builder = new StringBuilder(ValuePattern(nextVertex.LogicValue)).Append(" in {");
                var first = true;
                foreach (var inValue in moduleExpr.InValues)
                {
                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        builder.Append(",");
                    }
                    builder.Append(" ").Append(inValue.ToString());
                }
                // Create a vertex for the construction.
                var inVertex = new InVertex() { Pattern = builder.Append(" }").ToString(), LogicValue = logicValue };
                // Add the vertex.
                Graph.AddVertex(inVertex);
                // Add a new edge between the in construction vertex and the subexpression whose result it casts to TRUE/FALSE.

                var label = ValuePattern(nextVertex.LogicValue);
                Graph.AddEdge(new ExtendedEdge(inVertex, nextVertex) { Label = label, NotImportant = label=="?"});
                // Use the in construction vertex as a new nextVertex for the module construction.
                nextVertex = inVertex;
            }
            // Create and add a module vertex.
            var moduleVertex = new ModuleVertex() { ModuleName = ExecModule, LogicValue = nextVertex.LogicValue };
            Graph.AddVertex(moduleVertex);
            // Connect the module vertex with the next vertex.
            var label2 = ValuePattern(nextVertex.LogicValue);
            Graph.AddEdge(new ExtendedEdge(moduleVertex, nextVertex) { Label = label2, NotImportant = label2=="?"});
            // Revert the execution module with the previously created backup.
            ExecModule = backupExecModule;
            // Return the module vertex.
            return moduleVertex;
        }

        public LogicValueVertex visitAnd(AndExpression andExpr)
        {
            // Create and add a conjunction vertex.
            var andVertex = new AndVertex();
            Graph.AddVertex(andVertex);
            // Remember an initial value for the connective node.
            LogicValue? value = LogicValue.TRUE;
            // Prepare to build a pattern.
            var patternBuilder = new StringBuilder("and(");
            var firstInPattern = true;
            // We have multiple vertices that are next, so we have to operate on all and connect to the conjunction vertex.
            foreach (var next in andExpr.Nexts)
            {
                // Create a vertex for one of the next expression nodes.
                var nextVertex = next.Accept(this);
                // Connect the conjunction vertex with the next vertex.
                var label = ValuePattern(nextVertex.LogicValue);
                Graph.AddEdge(new ExtendedEdge(andVertex, nextVertex) { Label = label, NotImportant = label=="?"});
                // Calculate the connective value for the vertex.
                if (value == null || nextVertex.LogicValue == null)
                    value = null;
                else
                    value = Logic.and(value.GetValueOrDefault(LogicValue.UNKNOWN), nextVertex.LogicValue.GetValueOrDefault(LogicValue.UNKNOWN));
                // Append the value to the pattern.
                patternBuilder.Append(String.Format("{0}{1}", firstInPattern ? "" : ", ", ValuePattern(nextVertex.LogicValue)));
                firstInPattern = false;
            }
            // Add the pattern.
            andVertex.Pattern = patternBuilder.Append(")").ToString();
            // Set the connective value for the vertex.
            andVertex.LogicValue = value;
            // Return the conjunction vertex.
            return andVertex;
        }

        public LogicValueVertex visitOr(OrExpression orExpr)
        {
            // Create and add a alternative vertex.
            var orVertex = new OrVertex();
            Graph.AddVertex(orVertex);
            // Remember an initial value for the connective node.
            LogicValue? value = LogicValue.FALSE;
            // Prepare to build a pattern.
            var patternBuilder = new StringBuilder("or(");
            var firstInPattern = true;
            // We have multiple vertices that are next, so we have to operate on all and connect to the alternative vertex.
            foreach (var next in orExpr.Nexts)
            {
                // Create a vertex for one of the next expression nodes.
                var nextVertex = next.Accept(this);
                // Connect the alternative vertex with the next vertex.
                var label = ValuePattern(nextVertex.LogicValue);
                Graph.AddEdge(new ExtendedEdge(orVertex, nextVertex) { Label =  label, NotImportant = label=="?"});
                // Calculate the connective value for the vertex.

                if (value == null || nextVertex.LogicValue == null)
                    value = null;
                else
                    value = Logic.or(value.GetValueOrDefault(LogicValue.UNKNOWN), nextVertex.LogicValue.GetValueOrDefault(LogicValue.UNKNOWN));
                // Append the value to the pattern.
                patternBuilder.Append(String.Format("{0}{1}", firstInPattern ? "" : ", ", ValuePattern(nextVertex.LogicValue)));
                firstInPattern = false;
            }
            // Add the pattern.
            orVertex.Pattern = patternBuilder.Append(")").ToString();
            // Set the connective value for the vertex.
            orVertex.LogicValue = value;
            // Return the alternative vertex.
            return orVertex;
        }

        public LogicValueVertex visitExists(ExistsExpression exExpr)
        {
            // Move qunatifier-related enumerators
            QuantifierValueEnumerator.MoveNext();
            QuantifierVariableEnumerator.MoveNext();


            ValueHolder prevVal = null;
            if (qantifiedVariableValues.ContainsKey(exExpr.VariableName))
            {
                prevVal = qantifiedVariableValues[exExpr.VariableName];
            }

            qantifiedVariableValues[exExpr.VariableName] = QuantifierVariableEnumerator.Current;
            // Create and add a negation vertex.
            var exVertex = new ExistsVertex()
            {
                SubExpression = exExpr.Next,
                Pattern = QuantifierPattern(exExpr, "Exists", QuantifierValueEnumerator.Current, QuantifierVariableEnumerator.Current, exExpr.DomainModule),
                LogicValue = QuantifierValueEnumerator.Current,
                ModuleName = exExpr.DomainModule ?? ExecModule,
                DomainModule = exExpr.DomainModule,
                VariableName = exExpr.VariableName,
                VariableValue = QuantifierVariableEnumerator.Current,
                VariableDomain = exExpr.VariableDomain
            };


            

            Graph.AddVertex(exVertex);
            //Create a vertex for variable
            var varVertex = new QuantifierVariableVertex(exVertex);
            Graph.AddVertex(varVertex);

            // Create a vertex for what is next.
            var nextVertex = exExpr.Next.Accept(this);

            // Connect the quantifier vertex with the variable vertex.
            var label = ValuePattern(nextVertex.LogicValue);
            Graph.AddEdge(new ExtendedEdge(exVertex, varVertex) { Label = label, NotImportant = label=="?"});
            // Connect the variable vertex with the next vertex.
            Graph.AddEdge(new ExtendedEdge(varVertex, nextVertex) { Label = label, NotImportant = label=="?"});


            if (prevVal != null)
            {
                qantifiedVariableValues[exExpr.VariableName] = prevVal;
            }
            else
            {
                qantifiedVariableValues.Remove(exExpr.VariableName);
            }

            // Return the quantifier vertex.
            return exVertex;
        }

        public LogicValueVertex visitAll(AllExpression allExpr)
        {
            // Move qunatifier-related enumerators
            QuantifierValueEnumerator.MoveNext();
            QuantifierVariableEnumerator.MoveNext();
            // Create and add a negation vertex.

            ValueHolder prevVal  = null;
            if (qantifiedVariableValues.ContainsKey(allExpr.VariableName))
            {
                prevVal = qantifiedVariableValues[allExpr.VariableName];
            }

            qantifiedVariableValues[allExpr.VariableName] = QuantifierVariableEnumerator.Current;

            var exVertex = new AllVertex()
            {
                 SubExpression = allExpr.Next,
                 Pattern = QuantifierPattern(allExpr, "All", QuantifierValueEnumerator.Current, QuantifierVariableEnumerator.Current, allExpr.DomainModule),
                 LogicValue = QuantifierValueEnumerator.Current,
                 ModuleName = allExpr.DomainModule ?? ExecModule,
                 DomainModule = allExpr.DomainModule,
                 VariableName = allExpr.VariableName,
                 VariableValue = QuantifierVariableEnumerator.Current,
                 VariableDomain = allExpr.VariableDomain
            };


            Graph.AddVertex(exVertex);
            // Create a vertex for variable
            var varVertex = new QuantifierVariableVertex(exVertex);
            Graph.AddVertex(varVertex);

            // Create a vertex for what is next.
            var nextVertex = allExpr.Next.Accept(this);

            // Connect the quantifier vertex with the variable vertex.
            var label = ValuePattern(nextVertex.LogicValue);
            Graph.AddEdge(new ExtendedEdge(exVertex, varVertex) { Label = label, NotImportant = label=="?"});
            // Connect the variable vertex with the next vertex.
            Graph.AddEdge(new ExtendedEdge(varVertex, nextVertex) { Label = label, NotImportant = label=="?"});


            if (prevVal != null)
            {
                qantifiedVariableValues[allExpr.VariableName] = prevVal;
            }
            else
            {
                qantifiedVariableValues.Remove(allExpr.VariableName);
            }

            // Return the quantifier vertex.
            return exVertex;
        }

        #endregion

        private string QuantifierPattern(QantifierExpression expr, string quantifierName, LogicValue? quantifierValue, ValueHolder variableValue, string domainModule)
        {
            String pattern;
            if (quantifierValue != null && variableValue == null)  //case for empty domain
                pattern = String.Format("{0} {1} : {2}.{3} [domain empty]",quantifierName, expr.VariableName, domainModule, expr.VariableDomain);
            else // domain not empty
                pattern = String.Format("{0} {1} : {2}.{3} [chosen value: {4} = {5}]", quantifierName, expr.VariableName, domainModule, expr.VariableDomain, expr.VariableName, variableValue == null ? "?" : variableValue.Value);
            return pattern;
        }

        public static String ValuePattern(LogicValue? value)
        {
            return value == null ? "?" : value.ToString();
        }

    }


}
