﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using FourqlModeler.Calculcations;
using FourqlModeler.Communication.Import;
using FourqlModeler.ViewModels;
using FourqlModeler.Views.GraphExtensions.ExprGraph;
using FourqlModeler.Views.GraphExtensions.ModuleGraph;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using GraphX;
using GraphX.Controls.Models;
using QuickGraph;

namespace FourqlModeler.Views.GraphExtensions
{
    public delegate void PreContextMenuFillHandler(VertexControl vertexControl);
    public delegate void PostContextMenuFillHandler(VertexControl vertexControl);

    public class ExtendedGraphArea : GraphArea<ExtendedVertex, ExtendedEdge, BidirectionalGraph<ExtendedVertex, ExtendedEdge>>
    {
        public event PreContextMenuFillHandler PreContextMenuFill;
        public event PostContextMenuFillHandler PostContextMenuFill;

        public ICollection<ExtendedVertex> CollapsedVertices { get; set; }
        public ExtendedGraph PrototypeGraph { get; set; }

        public ExtendedGraphArea()
        {
            CollapsedVertices = new HashSet<ExtendedVertex>();
        }

        public void Expand(ExtendedVertex vertex)
        {
            CollapsedVertices.Remove(vertex);
            vertex.Collapsed = false;
            UpdateExpansion();
        }

        public void Collapse(ExtendedVertex vertex)
        {
            CollapsedVertices.Add(vertex);
            vertex.Collapsed = true;
            UpdateExpansion();
        }

        public void UpdateExpansion()
        {
            // FIXME: CYCLES! IT WILL NOT WORK WITH CYCLES AS SOME GRAPHS MIGHT NOT EVEN HAVE A ROOT! SO WE GOT TO CHECK CYCLES SEPARATELY
            // We assume there's at least one root here.
            var visited = new HashSet<ExtendedVertex>();
            foreach (var root in PrototypeGraph.Vertices)
            {
                var incomingEdges = PrototypeGraph.InEdges(root);
                if (incomingEdges == null || incomingEdges.Count() == 0)
                {
                    FindAllReachable(visited, root);
                }
            }

            // Add all the visited verticies.
            var graph = new ExtendedGraph();
            foreach (var vertex in visited)
            {
                graph.AddVertex(vertex);
            }

            // Push the edges that are visible after our action.
            // FIXME: Change it to work in linear time by remembering
            // edges for every visited vertex - just like in FIND UNION.
            // FIXME: Move it to the graph structure.
            // FIXME: Add collapsing by category (like module parent nodes
            // that collapse their whole module definitions).
            foreach (var vertex in visited)
            {
                PushEdges(graph, vertex, visited);
            }

            // Re-layout the graph.
            RelayoutGraph(graph);

            // Add the context menu to the verticies.
            foreach (var vertex in GetAllVertexControls())
            {
                CreateContextMenuForVertexNoRet(vertex);
            }

            // Update vertices and edges.
            UpdateControls();
        }

        private void FindAllReachable(ICollection<ExtendedVertex> visited, ExtendedVertex vertex)
        {
            if (visited.Contains(vertex))
            {
                return;
            }

            visited.Add(vertex);
            if (CollapsedVertices.Contains(vertex))
            {
                return;
            }

            foreach (var next in PrototypeGraph.OutEdges(vertex))
            {
                FindAllReachable(visited, next.Target);
            }
        }

        private void PushEdges(ExtendedGraph graph, ExtendedVertex vertex, ICollection<ExtendedVertex> reachable)
        {
            var outgoingEdges = PrototypeGraph.OutEdges(vertex);
            if (CollapsedVertices.Contains(vertex))
            {
                var visited = new HashSet<ExtendedVertex>();
                var closest = new HashSet<ExtendedVertex>();
                var edges = new HashSet<ExtendedEdge>();
                foreach (var edge in outgoingEdges)
                {
                    if (reachable.Contains(edge.Target))
                    {
                        edges.Add(edge);
                    }
                    else
                    {
                        FindClosestReachable(closest, visited, reachable, edge.Target);
                        foreach (var target in closest)
                        {
                            edges.Add(new ExtendedEdge(vertex, target) { ToolTip = string.Format("{0} -> {1}", vertex.Text, target.Text) });
                        }
                    }
                }
                outgoingEdges = edges;
            }

            foreach (var edge in outgoingEdges)
            {
                graph.AddEdge(edge);
            }
        }

        private void FindClosestReachable(HashSet<ExtendedVertex> closestReachable, HashSet<ExtendedVertex> visited, ICollection<ExtendedVertex> limitTo, ExtendedVertex vertex)
        {
            if (limitTo.Contains(vertex))
            {
                closestReachable.Add(vertex);
                return;
            }

            if (visited.Contains(vertex))
            {
                return;
            }

            visited.Add(vertex);
            foreach (var next in PrototypeGraph.OutEdges(vertex))
            {
                FindClosestReachable(closestReachable, visited, limitTo, next.Target);
            }
        }

        private void RelayoutGraph(ExtendedGraph graph)
        {
            ExtendedVertex dummyVertex = null;
            ExtendedEdge dummyEdge = null;
            ExtendedVertex[] vertexArray = null;
            if (graph.Vertices.Count() == 1)
            {
                dummyVertex = new ExtendedVertex() { ID = -1 };
                graph.AddVertex(dummyVertex);
                vertexArray = graph.Vertices.ToArray<ExtendedVertex>();
                dummyEdge = new ExtendedEdge(vertexArray[0], vertexArray[1]);
                graph.AddEdge(dummyEdge);
            }

            GenerateGraph(graph, true);
            RelayoutGraph();

            if (dummyVertex != null)
            {
                RemoveEdge(dummyEdge);
                RemoveVertex(dummyVertex);
            }
        }

        public void UpdateControls()
        {
            foreach (var vertexControl in VertexList.Values)
            {
                CreateContextMenuForVertexNoRet(vertexControl);
                DragBehaviour.SetUpdateEdgesOnMove(vertexControl, true);

                HighlightBehaviour.SetIsHighlightEnabled(vertexControl, true);
                HighlightBehaviour.SetHighlightControl(vertexControl, GraphControlType.VertexAndEdge);
                HighlightBehaviour.SetHighlightEdges(vertexControl, EdgesType.All);
                HighlightBehaviour.SetHighlightStrategy(vertexControl, HighlightStrategy.UseExistingControls);
            }

            foreach (var edgeControl in EdgesList.Values)
            {
                HighlightBehaviour.SetIsHighlightEnabled(edgeControl, true);
                HighlightBehaviour.SetHighlightControl(edgeControl, GraphControlType.VertexAndEdge);
                HighlightBehaviour.SetHighlightEdges(edgeControl, EdgesType.All);
                HighlightBehaviour.SetHighlightStrategy(edgeControl, HighlightStrategy.UseExistingControls);
            }

            UpdateAllEdges();
        }   
        
        public virtual void CreateContextMenuForVertexNoRet(VertexControl vertexControl)
        {
            vertexControl.ContextMenu = new ContextMenu();
            if (PreContextMenuFill != null)
            {
                PreContextMenuFill(vertexControl);
            }
            var item = new MenuItem() { DataContext = vertexControl };
            var modelVertex = vertexControl.Vertex as ExtendedVertex;

            if (CollapsedVertices.Contains(modelVertex))
            {
                item.Header = "Expand";
                item.Click += ExpandVertexHandler;
            }
            else
            {
                item.Header = "Collapse";
                item.Click += CollapseVertexHandler;
            }

            vertexControl.ContextMenu.Items.Add(item);
            if (PostContextMenuFill != null)
            {
                PostContextMenuFill(vertexControl);
            }
        }

        private void CollapseVertexHandler(object sender, RoutedEventArgs e)
        {
            var menuItem = sender as MenuItem;
            var vertexControl = menuItem.DataContext as VertexControl;
            var vertex = vertexControl.Vertex as ExtendedVertex;

            Collapse(vertex);
            CreateContextMenuForVertexNoRet(vertexControl);
        }

        private void ExpandVertexHandler(object sender, RoutedEventArgs e)
        {
            var menuItem = sender as MenuItem;
            var vertexControl = menuItem.DataContext as VertexControl;
            var vertex = vertexControl.Vertex as ExtendedVertex;

            Expand(vertex);
            CreateContextMenuForVertexNoRet(vertexControl);
        }
    }
}
