﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlModeler.Utility;
using FourqlModeler.ViewModels;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;

namespace FourqlModeler.Views.GraphExtensions.ModuleGraph
{
    public class ModuleGraphBuilder
    {

        private void AddErrorVertices(ReasoningOrder reasoningOrder, ExtendedGraph graph)
        {
            foreach (var syntaxErrorModule in reasoningOrder.ModulesWithSyntaxErrors)
            {
                // Create a tool tip.
                var builder = new StringBuilder("This 4QL file has syntax errors and thus cannot be parsed and used as a set of modules:");
                foreach (var error in syntaxErrorModule.SyntaxErrors)
                {
                    builder.Append(String.Format("\n  * {0}@{1}:{2}", error.Description, error.StartLine, error.StartColumn));
                }
                graph.AddVertex(new SyntaxErrorVertex() { ModuleViewModel = syntaxErrorModule, Errors = new List<ErrorDesc>(syntaxErrorModule.SyntaxErrors), ToolTip = builder.ToString() });
            }
        }

        private void AddDuplicatedModuleVertices(ReasoningOrder reasoningOrder, ExtendedGraph graph, IDictionary<string, ExtendedVertex> moduleToVertex)
        {

            foreach (var duplicateModuleEntry in reasoningOrder.DuplicatedModules)
            {
                // Create a message for duplicated modules.
                var builder = new StringBuilder(String.Format("Aggregate of a duplicated module that has multiple instances: {0}", duplicateModuleEntry.Value.Count));
                // Add an aggregated vertex for every duplicated module.
                var aggregateVertex = new DuplicatedModuleAggregateVertex() { Name = duplicateModuleEntry.Key };
                graph.AddVertex(aggregateVertex);
                // Map aggregated vertices to the module names they represent.
                moduleToVertex.Add(duplicateModuleEntry.Key, aggregateVertex);
                // Also, attach a vertex for every module source to respective aggregated vertices.
                foreach (var moduleSource in duplicateModuleEntry.Value)
                {
                    // Append module information to the message builder.
                    var module = moduleSource.CurrentProgram.FindModules(duplicateModuleEntry.Key).First();
                    builder.Append(String.Format("\n  * {0}{1}@{2}:{3}", moduleSource.FileDir, moduleSource.FileName, module.startLine(), module.startColumn()));
                    // Create a vertex for the duplicate.
                    var moduleVertex = new DuplicatedModuleVertex() { Name = duplicateModuleEntry.Key, ModuleViewModel = moduleSource, ToolTip = String.Format("The module has multiple instances: {0}", duplicateModuleEntry.Value.Count) };
                    graph.AddVertex(moduleVertex);
                    graph.AddEdge(new ExtendedEdge(aggregateVertex, moduleVertex) { ToolTip = String.Format("Duplicate of module {0}.", duplicateModuleEntry.Key) });
                }
                // Set the tool tip.
                aggregateVertex.ToolTip = builder.ToString();
            }
        }

        private void AddCorrectModules(ReasoningOrder reasoningOrder, ExtendedGraph graph,IDictionary<string, ExtendedVertex> moduleToVertex, ISet<int> addedErrors)
        {
            foreach (var moduleEntry in reasoningOrder.ModuleNodes)
            {
                ModuleEntryVertex vertex;
                var model = moduleEntry.Value.ModuleViewModel;
                var module = moduleEntry.Value.Module;
                if (moduleEntry.Value.ErroneousDependencies)
                {
                    // Prepare a tool tip.
                    var errors = model.ModuleStaticAnalysisErrors.ContainsKey(moduleEntry.Key) ? model.ModuleStaticAnalysisErrors[moduleEntry.Key].ToList() : new List<ErrorDesc>();
                    var builder = new StringBuilder("The module has one or more errors: " + errors.Count);
                    if (errors.Count == 0)
                    {
                        builder.Append("\n  * The module itself has no errors, but it depends on erroneous modules.");
                    }
                    else
                    {
                        foreach (var error in errors)
                        {
                            if (addedErrors.Add(error.Id))
                            {
                                builder.Append(String.Format("\n  * {0}@{1}:{2}", error.Description, error.StartLine, error.StartColumn));
                            }
                        }
                    }
                    // Add erroneous module to the graph.                    
                    vertex = new ErroneousVertex() { Name = moduleEntry.Key, ModuleViewModel = model, Errors = errors, ToolTip = builder.ToString() };
                }
                else
                {
                    // This vertex is not erroneous - add it as such.
                    vertex = new ModuleEntryVertex() { Name = moduleEntry.Key, ModuleViewModel = model, ToolTip = "The module was parsed and has no errors - can be/was loaded to the interpreted space as such." };
                }
                // Add the vertex to the graph.
                graph.AddVertex(vertex);
                // Add the vertex to dictionary as well.
                moduleToVertex.Add(moduleEntry.Key, vertex);
            }
        }

        private void AddRelationVertices(ExtendedGraph graph, IDictionary<string, ExtendedVertex> moduleToVertex, IDictionary<string, IDictionary<string, ICollection<string>>> groupedRelationDependencies, KeyValuePair<string, ReasoningNode> moduleNodeEntry )
        { 
            // Remember a module vertex.
            var moduleVertex = moduleToVertex[moduleNodeEntry.Key];
                
            foreach (var dependencyRelationEntry in groupedRelationDependencies)
            {
                // Create a relation vertex.
                var relationName = dependencyRelationEntry.Key;
                var relation = moduleNodeEntry.Value.Module.FindRelation(relationName);
                var relationVertex = new RelationVertex() { Relation = moduleNodeEntry.Value.Module.FindRelation(relationName), RelationName = relationName, RelationPattern = new ProgramPrinter(false).append(relation, "", false).Done() };
                // Add the relation vertex.
                graph.AddVertex(relationVertex);
                // Add an edge between the module vertex and relation vertex.
                graph.AddEdge(new ExtendedEdge(moduleVertex, relationVertex));
                // Now we need to add edges from the relation vertex to dependency modules.
                foreach (var dependencyModuleEntry in dependencyRelationEntry.Value)
                {
                    // Add missing dependency vertex if necessary.
                    if (!moduleToVertex.ContainsKey(dependencyModuleEntry.Key))
                    {
                        // If inexisting - create a dummy vertex for it.
                        var missingModuleVertex = new MissingModuleVertex() { Name = dependencyModuleEntry.Key, ToolTip = String.Format("Required module {0} is missing.", dependencyModuleEntry.Key) };
                        // Add it to our dictionary.
                        moduleToVertex.Add(dependencyModuleEntry.Key, missingModuleVertex);
                        // And to the graph itself.
                        graph.AddVertex(missingModuleVertex);
                    }
                    // Grab the dependency module vertex.
                    var dependencyModuleVertex = moduleToVertex[dependencyModuleEntry.Key];
                    // Prepare a tool tip for the edge.
                    var builder = new StringBuilder("The relation '").Append(relationName).Append("' from '").Append(moduleNodeEntry.Key).Append("' depends on the following relations from '").Append(dependencyModuleEntry.Key).Append("':");
                    foreach (var dependencyRelation in dependencyModuleEntry.Value)
                    {
                        builder.Append("\n  * ").Append(dependencyRelation);
                    }
                    // Add an edge between the relation vertex and its target (dependency) module.
                    graph.AddEdge(new ExtendedEdge(relationVertex, dependencyModuleVertex) { Label = String.Format("{0}{1}", dependencyModuleEntry.Value.First(), dependencyModuleEntry.Value.Count > 1 ? ", ..." : ""), ToolTip = builder.ToString() });
                }
            }
        }

        private void AddDomainVertices(ExtendedGraph graph, IDictionary<string, ExtendedVertex> moduleToVertex, Dictionary<string, IDictionary<string, ICollection<string>>> groupedDomainDependencies, KeyValuePair<string, ReasoningNode> moduleNodeEntry)
        {
            // Remember a module vertex.
            var moduleVertex = moduleToVertex[moduleNodeEntry.Key];

            foreach (var dependencyDomainEntry in groupedDomainDependencies)
            {
                // Create a domain vertex.
                var domainName = dependencyDomainEntry.Key;
                var domain = moduleNodeEntry.Value.Module.FindDomain(domainName);
                var domainVertex = new DomainVertex()
                {
                    Domain = domain,
                    DomainName = domain.Name,
                    DomainType = domain.Type,
                    DomainPattern = new ProgramPrinter(false).append(domain, "", false).Done()
                };
                // Add the domain vertex.
                graph.AddVertex(domainVertex);
                // Add an edge between the module vertex and relation vertex.
                graph.AddEdge(new ExtendedEdge(moduleVertex, domainVertex));
                // Now we need to add edges from the relation vertex to dependency modules.
                foreach (var dependencyModuleEntry in dependencyDomainEntry.Value)
                {
                    // Add missing dependency vertex if necessary.
                    if (!moduleToVertex.ContainsKey(dependencyModuleEntry.Key))
                    {
                        // If inexisting - create a dummy vertex for it.
                        var missingModuleVertex = new MissingModuleVertex() { Name = dependencyModuleEntry.Key, ToolTip = String.Format("Required module {0} is missing.", dependencyModuleEntry.Key) };
                        // Add it to our dictionary.
                        moduleToVertex.Add(dependencyModuleEntry.Key, missingModuleVertex);
                        // And to the graph itself.
                        graph.AddVertex(missingModuleVertex);
                    }
                    // Grab the dependency module vertex.
                    var dependencyModuleVertex = moduleToVertex[dependencyModuleEntry.Key];
                    // Prepare a tool tip for the edge.
                    var builder = new StringBuilder("The domain '").Append(domainName).Append("' from '").Append(moduleNodeEntry.Key).Append("' depends on the following domains from '").Append(dependencyModuleEntry.Key).Append("':");
                    foreach (var dependencyRelation in dependencyModuleEntry.Value)
                    {
                        builder.Append("\n  * ").Append(dependencyRelation);
                    }
                    // Add an edge between the relation vertex and its target (dependency) module.
                    graph.AddEdge(new ExtendedEdge(domainVertex, dependencyModuleVertex) { Label = String.Format("{0}{1}", dependencyModuleEntry.Value.First(), dependencyModuleEntry.Value.Count > 1 ? ", ..." : ""), ToolTip = builder.ToString() });
                }
            }
        }

        private void ShowRelationDependencies(ReasoningOrder reasoningOrder, ExtendedGraph graph, IDictionary<string, ExtendedVertex> moduleToVertex)
        {
            // We want to have dependencies shown between a source module, its relations and target modules (the modules that hold relations that the source module depends on).
            foreach (var moduleNodeEntry in reasoningOrder.ModuleNodes)
            {
                // First of all we have to group dependencies by relations (relation => [module => [relation]]).
                var groupedRelationDependencies = new Dictionary<string, IDictionary<string, ICollection<string>>>();
                //Group dependencies by domains (domain => [module => [domain]]).
                var groupedDomainDependencies = new Dictionary<string, IDictionary<string, ICollection<string>>>();
                foreach (var moduleDependencies in moduleNodeEntry.Value.DependsOn)
                {
                    AddDependenciesOf(moduleDependencies.Key.Module.Name, moduleDependencies.Value, groupedRelationDependencies, groupedDomainDependencies);
                }
                foreach (var moduleDependencies in moduleNodeEntry.Value.DependsOnInvalid)
                {
                    AddDependenciesOf(moduleDependencies.Key, moduleDependencies.Value, groupedRelationDependencies, groupedDomainDependencies);
                }
                // Grab those relations that have at least one dependency.
                var dependantRelationGroups = groupedRelationDependencies.Where(relationEntry => relationEntry.Value.Any(moduleEntry => moduleEntry.Value.Count > 0));
               // Now create relation vertices and add edges from their mother modules to dependency modules.
               AddRelationVertices(graph, moduleToVertex, groupedRelationDependencies, moduleNodeEntry);

               // Now create domain vertices and add edges from their mother modules to dependency modules.
               AddDomainVertices(graph, moduleToVertex, groupedDomainDependencies, moduleNodeEntry); 
            }
        }

       

        private void ShowStandartModuleGraph(ReasoningOrder reasoningOrder, ExtendedGraph graph, IDictionary<string, ExtendedVertex> moduleToVertex)
        {
            // Add edges.
            foreach (var moduleEntry in reasoningOrder.ModuleNodes)
            {
                // For every correct module..
                var moduleVertex = moduleToVertex[moduleEntry.Key];
                foreach (var dependsOn in moduleEntry.Value.DependsOn)
                {
                    // ... prepare a dependency tool tip...
                    var builder = new StringBuilder(moduleEntry.Key).Append(" depends on relations from ").Append(dependsOn.Key.Module.Name).Append(":");
                    foreach (var ruleDependencies in dependsOn.Value.RuleDependency)
                    {
                        builder.Append("\n  * ").Append(new ProgramPrinter(false).append(ruleDependencies.Key, "", false).Done()).Append(":");
                        foreach (var relationName in ruleDependencies.Value)
                        {
                            builder.Append("\n    - ").Append(relationName);
                        }
                    }
                    builder.Append("\n" + moduleEntry.Key).Append(" depends on domains from ").Append(dependsOn.Key.Module.Name).Append(":");
                    foreach (var domainDependencies in dependsOn.Value.DomainDependency)
                    {
                        builder.Append("\n  * ").Append(new ProgramPrinter(false).append(domainDependencies.Key, "", false).Done()).Append(":");
                        foreach (var domainName in domainDependencies.Value)
                        {
                            builder.Append("\n    - ").Append(domainName);
                        }
                    }
                    // ... add edges from the correctly parsed modules that this module depends on.
                    graph.AddEdge(new ExtendedEdge(moduleVertex, moduleToVertex[dependsOn.Key.Module.Name]) { ToolTip = builder.ToString() });
                }
                foreach (var dependsOn in moduleEntry.Value.DependsOnInvalid)
                {
                    // ... and add edges from the modules that are invalid - duplicated, inexisting.
                    if (!moduleToVertex.ContainsKey(dependsOn.Key))
                    {
                        // If inexisting - create a dummy vertex for it.
                        var missingModuleVertex = new MissingModuleVertex() { Name = dependsOn.Key, ToolTip = String.Format("Required module {0} is missing.", dependsOn.Key) };
                        // Add it to our dictionary.
                        moduleToVertex.Add(dependsOn.Key, missingModuleVertex);
                        // And to the graph itself.
                        graph.AddVertex(missingModuleVertex);
                    }
                    // Prepare the edge tool tip.
                    var builder = new StringBuilder(moduleEntry.Key).Append(" depends on relations from ").Append(dependsOn.Key).Append(":");
                    foreach (var ruleDependencies in dependsOn.Value.RuleDependency)
                    {
                        builder.Append("\n  * ").Append(new ProgramPrinter(false).append(ruleDependencies.Key, "", false).Done()).Append(":");
                        foreach (var relationName in ruleDependencies.Value)
                        {
                            builder.Append("\n    - ").Append(relationName);
                        }
                    }
                    builder.Append("\n" + moduleEntry.Key).Append(" depends on domains from ").Append(dependsOn.Key).Append(":");
                    foreach (var domainDependencies in dependsOn.Value.DomainDependency)
                    {
                        builder.Append("\n  * ").Append(new ProgramPrinter(false).append(domainDependencies.Key, "", false).Done()).Append(":");
                        foreach (var domainName in domainDependencies.Value)
                        {
                            builder.Append("\n    - ").Append(domainName);
                        }
                    }
                    // Add an edge between the module vertex and an invalid vertex.
                    graph.AddEdge(new ExtendedEdge(moduleVertex, moduleToVertex[dependsOn.Key]) { ToolTip = builder.ToString() });
                }
            }
        }

        public ExtendedGraph Build(ReasoningOrder reasoningOrder, bool showRelationDependency)
        {
            // Create an empty graph and a map of module names to vertices.
            var graph = new ExtendedGraph();
            var moduleToVertex = new Dictionary<string, ExtendedVertex>();
            var addedErrors = new HashSet<int>();

            // Add module files with syntax errors as singleton vertices.
            AddErrorVertices(reasoningOrder, graph);
            
            // Add vertices for duplicated modules.
            AddDuplicatedModuleVertices(reasoningOrder, graph, moduleToVertex);

            // Add modules that were verified and are correct in terms of syntax, but may not be error-free or might depend on erroneous modules.
            AddCorrectModules(reasoningOrder, graph, moduleToVertex, addedErrors);

            // See what display mode we want to have.
            if (showRelationDependency)
            {
               ShowRelationDependencies(reasoningOrder, graph, moduleToVertex);
            }
            else
            {
                // We want a standard display between modules.
                ShowStandartModuleGraph(reasoningOrder, graph, moduleToVertex);
            }
            // Connect all roots to one vertex if there is more than one.
            var roots = FindRoots(graph);
            // Did we find more than one root?
            if (roots.Count > 1)
            {
                // We did! Create a dummy root.
                var root = new RootVertex() { ToolTip = "Root aggregate - additional node create to connect multiple roots." };
                // Add the root to the graph.
                graph.AddVertex(root);
                // Connect other roots to it.
                foreach (var otherRoot in roots)
                {
                    graph.AddEdge(new ExtendedEdge(root, otherRoot) { ToolTip = "Root aggregate. No relation." });
                }
            }
            // Return the graph.
            return graph;
        }

        private ICollection<ExtendedVertex> FindRoots(ExtendedGraph graph)
        {
            var roots = new List<ExtendedVertex>();
            // Find clear roots, so vertices whose in-degree is 0.
            foreach (var vertex in graph.Vertices)
            {
                if (graph.InEdges(vertex).Count() == 0)
                {
                    roots.Add(vertex);
                }
            }
            // Now walk through all found subtrees and mark them as visited.
            var visited = new HashSet<ExtendedVertex>();
            foreach (var root in roots)
            {
                VisitSubtree(graph, root, visited);
            }
            // Find all non-visited subtrees to mark them as representatives - those cases handle separated cycles.
            foreach (var vertex in graph.Vertices)
            {
                if (!visited.Contains(vertex))
                {
                    roots.Add(vertex);
                    VisitSubtree(graph, vertex, visited);
                }
            }
            // Return the roots.
            return roots;
        }

        private void VisitSubtree(ExtendedGraph graph, ExtendedVertex root, HashSet<ExtendedVertex> visited)
        {
            if (visited.Contains(root))
            {
                return;
            }
            visited.Add(root);
            foreach (var outEdge in graph.OutEdges(root))
            {
                VisitSubtree(graph, outEdge.Target, visited);
            }
        }



        private void AddDependenciesOf(string dependencyModule, ReasoningDependency moduleDependencies, Dictionary<string, IDictionary<string, ICollection<string>>> groupedRelationDependencies, Dictionary<string, IDictionary<string, ICollection<string>>> groupedDomainDependencies)
        {
            foreach (var domainDependencies in moduleDependencies.DomainDependency)
            {
                // ... we check which domains depend on the module.
                var domainName = domainDependencies.Key.Name;
                if (!groupedDomainDependencies.ContainsKey(domainName))
                {
                    // Add a new group if necessary for the domain
                    groupedDomainDependencies.Add(domainName, new Dictionary<string, ICollection<string>>());
                }
                // Remember the domain dependencies to fill.
                var domDeps = groupedDomainDependencies[domainName];
                // See if the relation dependencies contain space for module dependencies.
                if (!domDeps.ContainsKey(dependencyModule))
                {
                    // They do not - add the space for module dependencies.
                    domDeps.Add(dependencyModule, new HashSet<string>());
                }
                // Hold the module dependencies for the relation (rule).
                var moduleDomainDependencies = domDeps[dependencyModule];
                foreach (var domainDependency in domainDependencies.Value)
                {
                    // Add the module dependencies for the relation.
                    moduleDomainDependencies.Add(domainDependency);
                }
            }
            foreach (var ruleDependencies in moduleDependencies.RuleDependency)
            {
                // ... we check which rules (=> relations) depend on the module.
                var relationName = ruleDependencies.Key.Head.RelationName;
                if (!groupedRelationDependencies.ContainsKey(relationName))
                {
                    // Add a new group if necessary for the relation (of the rule).
                    groupedRelationDependencies.Add(relationName, new Dictionary<string, ICollection<string>>());
                }
                // Remember the relation dependencies to fill.
                var relationDependencies = groupedRelationDependencies[relationName];
                // See if the relation dependencies contain space for module dependencies.
                if (!relationDependencies.ContainsKey(dependencyModule))
                {
                    // They do not - add the space for module dependencies.
                    relationDependencies.Add(dependencyModule, new HashSet<string>());
                }
                // Hold the module dependencies for the relation (rule).
                var moduleRelationDependencies = relationDependencies[dependencyModule];
                foreach (var relationDependency in ruleDependencies.Value)
                {
                    // Add the module dependencies for the relation.
                    moduleRelationDependencies.Add(relationDependency);
                }
            }
        }
    }
}
