﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using FourqlModeler.Calculcations;
using FourqlModeler.Calculcations.Communication;
using FourqlModeler.Communication;
using FourqlModeler.Communication.Import;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;

namespace FourqlModeler.Views.GraphExtensions.WSMGraph
{
    public class WsmDataBuilder
    {
        public static WsmData Build(Module module)
        {
            // Create an empty WSM data structure.
            var wsmData = new WsmData() { ModuleName = module.Name };
            // Add primitive domains
            wsmData = BuildPrimitiveDomains(module, wsmData);

            // Add defined domains
            wsmData = BuildDomains(module, wsmData);

            // Add relations
            BuildRelations(module, wsmData);
            // Return the data.
            return wsmData;
        }

        private static WsmData BuildPrimitiveDomains(Module module, WsmData wsmData)
        {
            // Add primitive domain values
            foreach (DomainType type in Enum.GetValues(typeof(DomainType)))
                if (!type.Equals(DomainType.VARIABLE))
                {
                    {
                        var values = new HashSet<object>();

                        // Execute the domain expression.
                        var query = String.Format("{0}.{1}", module.Name, ModuleCalcDataUtils.ExprToString(new Literal()
                        {
                            RelationName = type.ToString().ToLower(),
                            Values = {new ValueHolder("X", DomainType.VARIABLE)}
                        }));
                        var result = ApplicationClient.Instance.Query(query, false);
                        foreach (var row in result.Rows)
                        {
                            values.Add(row.Values[0]);
                        }
                        if (values.Count > 0)
                            wsmData.PrimitiveDomainValues.Add(type, values);

                    }
                }
            return wsmData;
        }

        private static WsmData BuildDomains(Module module, WsmData wsmData)
        {
            foreach (var domain in module.Domains)
            {
                var values = new HashSet<object>();
                wsmData.DefinedDomainTypes.Add(domain, domain.Type);
                wsmData.DefinedDomainValues.Add(domain, values);

                var query = String.Format("{0}.{1}", module.Name, ModuleCalcDataUtils.ExprToString(new Literal()
                {
                    RelationName = domain.Name,
                    Values = { new ValueHolder("X", DomainType.VARIABLE) }
                }));
                var result = ApplicationClient.Instance.Query(query);
                foreach (var row in result.Rows)
                {
                    values.Add(row.Values[0]);
                }
            }
            return wsmData;
        }

        private static WsmData BuildRelations(Module module, WsmData wsmData)
        {
            // Add relations if such exist in the module.
            if (module.Relations.Count > 0)
            {
                // Add relation facts.
                foreach (var relation in module.Relations)
                {
                    // Create a structure for the relation.
                    var relationFacts = new Dictionary<Literal, LiteralFacts>();

                    wsmData.RelationFacts.Add(relation, relationFacts);
                    // Add facts.
                    foreach (var fact in module.Facts.Where(fact => fact.RelationName.Equals(relation.Name)))
                    {
                        // Create a literal for the fact.
                        var literal = fact.CreateLiteralOfFact();
                        // See if we have a fact structure for it in the relation facts.
                        if (!relationFacts.ContainsKey(literal))
                        {
                            // We do not - add a new one.
                            relationFacts.Add(literal, new LiteralFacts(literal));
                        }
                        // Add the fact to the literal fact structure.
                        relationFacts[literal].Facts.Add(fact);
                    }
                    // Ask for relation

                    var relationLit = BuildRelationLiteral(relation, module);

                    var result = new CalcExprExecutor().Execute(relationLit.Accept(new ExprPrinter()).ToString(), module, ModuleRepository.Instance.Modules, LogicValue.FALSE, LogicValue.UNKNOWN);
                    foreach (var row in result.Rows)
                    {
                        var valueIter = row.Values.GetEnumerator();
                        var fieldIter = result.Fields.GetEnumerator();
                        var values = new Dictionary<string, ValueHolder>();
                        while (valueIter.MoveNext() && fieldIter.MoveNext())
                        {
                            values.Add(fieldIter.Current.Name, new ValueHolder(valueIter.Current, fieldIter.Current.Type));
                        }
                        // ... and using those in the rule head.
                        var relFact = new ExprValueSetter(values).SetInFact(new Fact(){RelationName = relationLit.RelationName, Values = relationLit.Values});
                        relFact.Value = row.Value;

                        var literal = relFact.CreateLiteralOfFact();
                        // See if we have a fact structure for it in the relation facts.
                        if (!relationFacts.ContainsKey(literal))
                        {
                            // We do not - add a new one.
                            relationFacts.Add(literal, new LiteralFacts(literal));
                        }

                    }

                    // Add rule facts.
                    foreach (var rule in module.Rules.Where(rule => rule.Head.RelationName.Equals(relation.Name)))
                    {
                        // Add the rule.
                        wsmData.Rules.Add(rule);
                        // Now we want to know what facts we calculated via the rules.
                        var ruleresult = new CalcExprExecutor().Execute(rule, module, ModuleRepository.Instance.Modules, LogicValue.FALSE, LogicValue.UNKNOWN);
                        // Create facts for the result.
                        foreach (var row in ruleresult.Rows)
                        {
                            // Create a fact for the row by finding all variable substitutes...
                            var valueIter = row.Values.GetEnumerator();
                            var fieldIter = ruleresult.Fields.GetEnumerator();
                            var values = new Dictionary<string, ValueHolder>();
                            while (valueIter.MoveNext() && fieldIter.MoveNext())
                            {
                                values.Add(fieldIter.Current.Name, new ValueHolder(valueIter.Current, fieldIter.Current.Type));
                            }
                            // ... and using those in the rule head.
                            var ruleFact = new ExprValueSetter(values).SetInFact(rule.Head);
                            // Resolve the value of the row.
                            ruleFact.Value = row.Value == LogicValue.INCONSISTENT ? LogicValue.INCONSISTENT : (rule.Head.IsNegative ? LogicValue.FALSE : LogicValue.TRUE);
                            // Create a literal for the fact.
                            var literal = ruleFact.CreateLiteralOfFact();
                            // See if we have a fact structure for it in the relation facts.
                            if (relationFacts.ContainsKey(literal))
                            {
                                // Add the fact to the literal fact structure.
                                if (!relationFacts[literal].RuleFacts.ContainsKey(rule))
                                    relationFacts[literal].RuleFacts.Add(rule, ruleFact);
                            }
                        }
                    }

                    //delete empty 
                    var toRemove = from l in relationFacts
                        where l.Value.Facts.Count == 0 && l.Value.RuleFacts.Count == 0
                        select l.Key;
                    foreach (var r in toRemove.ToList())
                    {
                        relationFacts.Remove(r);
                    }
                }
            }
            return wsmData;
        }

        private static Literal BuildRelationLiteral(Relation relation, Module module)
        {
            var values = new List<ValueHolder>();
            for (var i = 0; i < relation.paramCount(); i++)
            {
                values.Add(new ValueHolder("X"+i, DomainType.VARIABLE));
            }
            return new Literal()
            {
                RelationName = relation.Name,
                Values = values
            };
        }

    }


    public class WsmData
    {
        public string ModuleName { get; set; }
        public IDictionary<Domain, DomainType> DefinedDomainTypes { get; set; }
        public IDictionary<DomainType, ICollection<object>> PrimitiveDomainValues { get; set; }
        public IDictionary<Domain, ICollection<object>> DefinedDomainValues { get; set; }
        public ICollection<Rule> Rules { get; set; }
        public IDictionary<Relation, IDictionary<Literal, LiteralFacts>> RelationFacts { get; set; }

        public WsmData()
        {
            DefinedDomainTypes = new Dictionary<Domain, DomainType>();
            PrimitiveDomainValues = new Dictionary<DomainType, ICollection<object>>();
            DefinedDomainValues = new Dictionary<Domain, ICollection<object>>();
            Rules = new HashSet<Rule>();
            RelationFacts = new Dictionary<Relation, IDictionary<Literal, LiteralFacts>>();
        }
    }

    public class LiteralFacts
    {
        public Literal Literal { get; set; }
        public ICollection<Fact> Facts { get; set; }
        public IDictionary<Rule, Fact> RuleFacts { get; set; }

        public LiteralFacts(Literal literal)
        {
            Literal = literal;
            Facts = new List<Fact>();
            RuleFacts = new Dictionary<Rule, Fact>();
        }
    }
}
