﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using FourqlModeler.Calculcations;
using FourqlModeler.Calculcations.Communication;
using FourqlModeler.Communication;
using FourqlModeler.Communication.Import;
using FourqlModeler.ViewModels;
using FourqlModeler.Views.GraphExtensions.ExprGraph;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;

namespace FourqlModeler.Views.GraphExtensions.WSMGraph
{
    public class WsmGraphBuilder
    {
        public static ExtendedGraph Build(Module module, WsmData wsmData, bool meldRuleVertices)
        {
            // Create a new extended graph to fill.
            var graph = new ExtendedGraph();
            // Create an initial module wsm root.
            var moduleVertex = new ModuleWsmVertex() { ModuleName = module.Name, Module = module };
            // Add the root to the graph.
            graph.AddVertex(moduleVertex);
            // Grab non-empty domains.
            var primitiveDomains = wsmData.PrimitiveDomainValues.Where(domainEntry => domainEntry.Value.Count > 0);
            var definedDomains = wsmData.DefinedDomainValues.Where(domainEntry => domainEntry.Value.Count > 0);
            // Add domains if we have any non-empty.
            if (primitiveDomains.Count() + definedDomains.Count() > 0)
            {
                // Create a root for domains.
                var domainsVertex = new DomainsVertex();
                // Add the vertex to the graph.
                graph.AddVertex(domainsVertex);
                // Add the root to the module root.
                graph.AddEdge(new ExtendedEdge(moduleVertex, domainsVertex));
                // Add primitive domains.
                foreach (var primitiveDomain in primitiveDomains)
                {
                    // Define a vertex for the domain.
                    var domainName = DomainTypeUtils.getName(primitiveDomain.Key);
                    var domainVertex = new DomainVertex() { DomainName = domainName, DomainPattern = "Base type", DomainType = primitiveDomain.Key, Domain = new Domain() { Name = domainName, ParserContext = null, Type = primitiveDomain.Key } };
                    // Add the domain vertex.
                    graph.AddVertex(domainVertex);
                    // Create an edge between the domain and the module's domains vertex.
                    graph.AddEdge(new ExtendedEdge(domainsVertex, domainVertex));
                    // Add values as vertices.
                    foreach (var value in primitiveDomain.Value)
                    {
                        // Create a value vertex.
                        var valueVertex = new DomainValueVertex() { Value = value };
                        // Add the vertex to the graph.
                        graph.AddVertex(valueVertex);
                        // Attach the vertex to the domain vertex.
                        graph.AddEdge(new ExtendedEdge(domainVertex, valueVertex));
                    }
                }
                // Add defined domains.
                foreach (var definedDomain in definedDomains)
                {
                    // Define a vertex for the domain.
                    var domainType = wsmData.DefinedDomainTypes[definedDomain.Key];
                    var domainVertex = new DomainVertex() { DomainName = definedDomain.Key.Name, DomainPattern = new ProgramPrinter(false).append(definedDomain.Key, "", false, false).Done(), DomainType = domainType, Domain = new Domain() { Name = definedDomain.Key.Name, ParserContext = definedDomain.Key.ParserContext, Type = domainType } };
                    // Add the domain vertex.
                    graph.AddVertex(domainVertex);
                    // Create an edge between the domain and the module's domains vertex.
                    graph.AddEdge(new ExtendedEdge(domainsVertex, domainVertex));
                    // Add values as vertices.
                    foreach (var value in definedDomain.Value)
                    {
                        // Create a value vertex.
                        var valueVertex = new DomainValueVertex() { Value = value };
                        // Add the vertex to the graph.
                        graph.AddVertex(valueVertex);
                        // Attach the vertex to the domain vertex.
                        graph.AddEdge(new ExtendedEdge(domainVertex, valueVertex));
                    }
                }
            }
            // Add relation vertices if relations exist in the module.
            if (wsmData.RelationFacts.Count > 0)
            {
                // Prepare to remember vertices for rules.
                var ruleVertices = new Dictionary<Rule, ExtendedVertex>();
                // Create a root vertex for relations.
                var relationsVertex = new RelationsVertex();
                // Add the root for relations to the graph.
                graph.AddVertex(relationsVertex);
                // Add an edge between the module and the relations root.
                graph.AddEdge(new ExtendedEdge(moduleVertex, relationsVertex));
                // Create vertices for relations.
                foreach (var relationEntry in wsmData.RelationFacts)
                {
                    // Create a relation vertex.
                    var relation = relationEntry.Key;
                    var relationVertex = new RelationVertex() { RelationName = relation.Name, RelationPattern = new ProgramPrinter(false).append(relation, "", false, false).Done(), Relation = relation };
                    // Add it to the graph.
                    graph.AddVertex(relationVertex);
                    // Attach it to the module root.
                    graph.AddEdge(new ExtendedEdge(relationsVertex, relationVertex));
                    // Add facts for literals of the relation.
                    foreach (var literalEntry in relationEntry.Value)
                    {
                        // First add a literal aggregate vertex.
                        var literalVertex = new LiteralVertex() { Literal = literalEntry.Key, ModuleName = module.Name, Pattern = String.Format("{0}.{1}", module.Name, new ExprPrinter("", false).visitLiteral(literalEntry.Key).ToString()) };
                        // Add the literal vertex to the graph.
                        graph.AddVertex(literalVertex);
                        // Connect the vertex with its relation.
                        graph.AddEdge(new ExtendedEdge(relationVertex, literalVertex));
                        // Evaluate the literal's combined logical value.
                        LogicValue? value = null;
                        // Add straight facts from the module definition.
                        foreach (var fact in literalEntry.Value.Facts)
                        {
                            // Create a fact vertex.
                            var factVertex = new FactVertex() { Fact = fact, Pattern = String.Format("{0}.{1}", module.Name, new ExprPrinter("", false).visitLiteral(fact).ToString()), Value = fact.IsNegative ? LogicValue.FALSE : LogicValue.TRUE };
                            // Add it to the graph.
                            graph.AddVertex(factVertex);
                            // Attach it to the relation vertex.
                            graph.AddEdge(new ExtendedEdge(literalVertex, factVertex));
                            // Update the combined value.
                            value = !value.HasValue ? fact.Value : (value.Value != fact.Value ? LogicValue.INCONSISTENT : value.Value);
                        }
                        // Add facts from rules.
                        foreach (var ruleFactEntry in literalEntry.Value.RuleFacts)
                        {
                            // Remember useful data.
                            var rule = ruleFactEntry.Key;
                            var ruleFact = ruleFactEntry.Value;
                            // Create a rule fact vertex.
                            var ruleFactVertex = new RuleFactVertex() { ModuleName = module.Name, Rule = rule, Fact = ruleFact, Pattern = String.Format("{0}.{1}", module.Name, new ExprPrinter("", false).visitLiteral(ruleFact).ToString()), Value = ruleFact.Value };
                            // Add it to the graph.
                            graph.AddVertex(ruleFactVertex);
                            // Attach it to the literal vertex.
                            graph.AddEdge(new ExtendedEdge(literalVertex, ruleFactVertex));
                            // See if we have a rule vertex from which the fact comes from.
                            if (!ruleVertices.ContainsKey(rule))
                            {
                                // We do not - create a vertex for it.
                                var ruleVertex = new RuleVertex() { HeadPattern = new ProgramPrinter(false).appendFactNoEnd(rule.Head).Done(), BodyPattern = new ExprPrinter("", false).ConvertExpr(rule.Body).ToString(), Rule = rule, ModuleName = module.Name, ToolTip = "From the rule:\n" + new ProgramPrinter().append(rule).Done() };
                                // Add it to the graph.
                                graph.AddVertex(ruleVertex);
                                // Add it to our dictionary.
                                ruleVertices.Add(rule, ruleVertex);
                            }
                            // Attach the rule fact to its rule vertex.
                            graph.AddEdge(new ExtendedEdge(ruleFactVertex, ruleVertices[rule]));
                            // See if we want rule vertices to meld.
                            if (!meldRuleVertices)
                            {
                                // We do not - remove the current rule vertex so we can recreate it for next encounters.
                                ruleVertices.Remove(rule);
                            }
                            // Update the combined value.
                            value = !value.HasValue ? ruleFact.Value : (value.Value != ruleFact.Value ? LogicValue.INCONSISTENT : value.Value);
                        }
                        // Set the combined value.
                        literalVertex.LogicValue = value.Value;
                    }
                }
            }
            // Return the graph.
            return graph;
        }
    }
}
