﻿using FourqlModeler.Commands;
using FourqlModeler.Communication;
using FourqlModeler.TabControl;
using FourqlModeler.ViewModels;
using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Net.Sockets;
using System.Windows;
using System.Windows.Controls;
using Xceed.Wpf.AvalonDock;
using Xceed.Wpf.AvalonDock.Layout;
using Parago.Windows;
using System.IO;
using System.Threading.Tasks;
using System.Threading;
using System.Windows.Media.Imaging;
using FourqlModeler.Utility;
using FourqlModeler.Views.GraphExtensions.ModuleGraph;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using FourqlModeler.Communication.Import;
using FourqlModeler.Calculcations;
using System.Collections.Generic;
using System.Linq;
using System.Xml.Serialization;
using System.Text;
using FourqlParser.Parsers;

namespace FourqlModeler.Views
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindowView : Window, IDisposable
    {

        public TabController TabController { get; private set; }

        public MainWindowView()
        {
            InitializeComponent();

            TabController = new TabController(CentralPanel, CentralPane);
            MainWindowViewModel.Instance.MainWindowView = this;
        }

        void setDockPanelProperties(LayoutAnchorable layout, bool toggleBack = true)
        {
            layout.ToggleAutoHide();
            var pane = layout.Parent as LayoutAnchorablePane;
            if (pane != null)
            {
                pane.DockHeight = new GridLength(300.0, GridUnitType.Pixel);
                pane.DockWidth = new GridLength(300.0, GridUnitType.Pixel);
            }

            if (toggleBack)
            {
                layout.ToggleAutoHide();
            }
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
        }


        private void AppLoaded(object sender, RoutedEventArgs e)
        {
            setDockPanelProperties(ConsoleLayout, false);
            setDockPanelProperties(ModuleTreeLayout, false);
            //setDockPanelProperties(DefinitionTreeLayout);

            var thread = new Thread(new ThreadStart(launchInterpreter));
            thread.SetApartmentState(ApartmentState.STA);
            thread.Start();
        }

        private void AppClosing(object sender, CancelEventArgs e)
        {
            var moduleStorage = App.Current.Resources["ModuleStorageViewModel"] as ModuleStorageViewModel;
            var dirtyModuleModels = new List<ModuleViewModel>(); 
            foreach (var moduleViewModel in moduleStorage.Modules)
            {
                if (moduleViewModel.IsDirty)
                {
                    dirtyModuleModels.Add(moduleViewModel);
                }
            }

            if (dirtyModuleModels.Count > 0)
            {
                var dialogResult = MessageBox.Show("One or more module models might be unsaved. Do you want to save them all?", "Unsaved changes", MessageBoxButton.YesNoCancel, MessageBoxImage.Question);
                if (dialogResult == MessageBoxResult.Cancel)
                {
                    e.Cancel = true;
                }
                else if (dialogResult == MessageBoxResult.Yes)
                {
                    string errorMessage = null;
                    var result = ProgressDialog.Execute(Application.Current.MainWindow, "Saving all modules", () =>
                    {
                        try
                        {
                            var i = 0;
                            var saveModuleCommand = new SaveModuleCommand(moduleStorage, false);
                            foreach (var moduleViewModel in dirtyModuleModels)
                            {
                                ProgressDialog.Current.Report("Saving {0}/{1} file ...\nFile name: {2}\nFile dir: {3}", ++i, dirtyModuleModels.Count, moduleViewModel.FileName, moduleViewModel.FileDir);
                                if (moduleViewModel.IsDirty)
                                {
                                    saveModuleCommand.Execute(moduleViewModel);
                                }
                            }
                        }
                        catch (Exception ex)
                        {
                            errorMessage = ex.Message;
                        }
                    }, ProgressDialogSettings.WithSubLabel);

                    if (errorMessage != null)
                    {
                        MessageBox.Show("Failed to save the 4QL files: " + errorMessage, "Save failed");
                    }
                }
            }
        }

        private void launchInterpreter()
        {
            Application.Current.Dispatcher.Invoke(() =>
            {
                string errorMessage = null;
                var result = ProgressDialog.Execute(Application.Current.MainWindow, "Launching Interpreter...", () =>
                {
                    ProgressDialog.Current.ReportWithCancellationCheck("Launching Interpreter...");
                   
                    try
                    {
                        ApplicationClient.Instance.connect();
                        ProgressDialog.Current.ReportWithCancellationCheck("Interpreter launched successfully!");

                        // Fetch the defined internal and external module names.
                        ProgressDialog.Current.ReportWithCancellationCheck("Fetching internal and external module names...");
                        //var modulesHandler = new ModulesHandler();
                        var modules = FetchModules();
                        //ApplicationClient.Instance.SendCommand(MessageFactory.modulesMessage(modulesHandler));
                        // Now fetch modules whose names we just got via the modules command.
                        if (modules.Count > 0)
                        {
                            ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Fetching internal and external module definitions... ({0} in total)", modules.Count));
                            var i = 0;
                            var moduleViewModels = new List<ModuleViewModel>();
                            foreach (var module in modules)
                            {
                                ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Fetching internal and external module definitions... {0} ({1}/{2})", module, ++i, modules.Count));
                                //var printHandler = new PrintHandler();
                                var moduleString = ApplicationClient.Instance.PrintModule(module);
                                //ApplicationClient.Instance.SendCommand(MessageFactory.printMessage(module, printHandler));
                                // Parse the module.
                                var errors = new ParserDispatcher.Errors();
                                var programContext = ParserDispatcher.parseProgram(moduleString, errors, null);
                                if (errors.ErrorCount > 0)
                                {
                                    throw new Exception("Internal/External module definitions should not give any parser errors while this one does: " + module);
                                }
                                var program = ParserDispatcher.toProgram(programContext);
                                moduleViewModels.Add(new ModuleViewModel() { FileName = module + ".4ql", FileDir = ModuleViewModel.InternalExternalDefinition, Content = moduleString, SavedContent = moduleString, IsDirty = false, CurrentProgram = program, LoadedProgram = program });
                                foreach (var moduleDef in program.Modules)
                                {
                                    moduleDef.IsInternalExternal = true;
                                    ModuleRepository.Instance.Modules.Add(moduleDef.Name, moduleDef);
                                }
                            }

                            Application.Current.Dispatcher.Invoke(() =>
                            {
                                var moduleStorageViewModel = Application.Current.Resources["ModuleStorageViewModel"] as ModuleStorageViewModel;
                                foreach (var moduleViewModel in moduleViewModels)
                                {
                                    moduleStorageViewModel.Modules.Add(moduleViewModel);
                                }
                            });
                        }
                    }
                    catch (Exception ex)
                    {
                        ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Exception was thrown while fetching internal and external module definitions: {0}", ex.Message));
                        throw ex;
                    }
                }, ProgressDialogSettings.WithSubLabelAndCancel);

                if (result != null && result.Cancelled)
                {
                    // The overall result of the operation was that the user cancelled it - abort the mission!
                    errorMessage = "was cancelled by the used";
                }

                if (errorMessage != null)
                {
                    // We could not close open the interpreter - return the error to close the application.
                    MessageBox.Show(String.Format("Interpreter failed to initializ: {0} - closing the application.", errorMessage), "Interpreter failed", MessageBoxButton.OK, MessageBoxImage.Error);
                    if (Application.Current != null)
                    {
                        Application.Current.Dispatcher.Invoke(() => Application.Current.Shutdown());
                    }
                }
            });
        }

        private IList<string> FetchModules()
        {
            try
            {
                var result = ApplicationClient.Instance.GetModules();
                var builder = new StringBuilder("Modules message indicated those internal and external modules are in the interpreter:");
               
                foreach (var module in result)
                {
                    builder.Append("\n\t- ").Append(module);
                }
                MainWindowViewModel.Instance.ConsoleViewModel.SendCommand(String.Format("Success of Modules:\n{0}", builder.ToString()));
                return result;
            } catch (Exception error)
            {
                var errorMessage = String.Format("Could not fetch internal and external module names from the interpreter: {0}", error.Message);
                MainWindowViewModel.Instance.ConsoleViewModel.SendCommand(errorMessage);
                throw new Exception(errorMessage);
            }
        }


        private void AppClosed(object sender, EventArgs e)
        {
            
        }

        public void OpenModuleEvent(ModuleViewModel moduleViewModel)
        {
            var filePath = moduleViewModel.FileDir + moduleViewModel.FileName;
            var tabName = filePath + "#def";
            var document = TabController.GetTab(tabName);
            if (document == null)
            {
                document = TabController.createTab(tabName, new ModuleEditorView(moduleViewModel));
            }
            TabController.selectTab(document);
        }

        public void CreateHierarchyGraphFor(ReasoningOrder reasoningOrder, bool showRelationDependency)
        {
            var tabName = "ReasoningOrder#graph";
            var document = TabController.GetTab(tabName);
            if (document != null)
            {
                document.Close();
            }
            var tab = TabController.createTab(tabName, new ModuleView(new ModuleGraphBuilder().Build(reasoningOrder, showRelationDependency)));
            tab.ContentId = tabName;
            tab.Title = tabName;
            TabController.selectTab(tab);
        }

        public LayoutDocument GetFocusedTab()
        {
            return TabController.getFocusedTab();
        }
    }

}
