﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using FourqlModeler.Calculcations;
using FourqlModeler.Communication;
using FourqlModeler.Communication.Import;
using FourqlModeler.TabControl;
using FourqlModeler.ViewModels;
using FourqlModeler.Views.GraphExtensions;
using FourqlModeler.Views.GraphExtensions.ModuleGraph;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using FourqlParser.Structures;
using GraphX;
using Parago.Windows;
using Xceed.Wpf.AvalonDock.Layout;

namespace FourqlModeler.Views
{
    /// <summary>
    /// Interaction logic for ModuleView.xaml
    /// </summary>
    public partial class ModuleView : UserControl
    {
        public ModuleView(ExtendedGraph graph)
        {
            InitializeComponent();
            ModuleGraphView.BaseGraph = graph;
            ModuleGraphView.GraphArea.PostContextMenuFill += CreateContextMenuForVertex;
            ModuleGraphView.GraphArea.VertexDoubleClick += VertexDoubleClick;
            Loaded += ModuleView_Loaded;
        }
        
        void VertexDoubleClick(object sender, GraphX.Models.VertexSelectedEventArgs args)
        {
            var vertex = args.VertexControl.Vertex as FileVertex;
            if (vertex != null)
            {
                MainWindowViewModel.Instance.MainWindowView.OpenModuleEvent(vertex.ModuleViewModel);
                DragBehaviour.SetIsDragging(args.VertexControl, false);
            }
        }

        void ModuleView_Loaded(object sender, RoutedEventArgs e)
        {
            if (!ModuleGraphView.IsGraphInitialized)
            {
                ModuleGraphView.RebuildGraph();
            }
        }

        void CreateContextMenuForVertex(VertexControl vertexControl)
        {
            var moduleEntryVertex = vertexControl.Vertex as ModuleEntryVertex;
            if (moduleEntryVertex == null)
            {
                return;
            }
            if (vertexControl.ContextMenu == null)
            {
                vertexControl.ContextMenu = new ContextMenu();
            }
            else
            {
                vertexControl.ContextMenu.Items.Add(new Separator());
            }
            var contextMenu = vertexControl.ContextMenu;
            if (vertexControl.Vertex is ModuleEntryVertex)
            {
                var openWsmMenuItem = new MenuItem() { Header = "Show WSM (meld rule vertices)", DataContext = vertexControl, IsEnabled = !moduleEntryVertex.ModuleViewModel.IsInternalExternal };
                openWsmMenuItem.Click += OpenWsmMeldRuleVertices;
                contextMenu.Items.Add(openWsmMenuItem);

                openWsmMenuItem = new MenuItem() { Header = "Show WSM (do not meld rule vertices)", DataContext = vertexControl, IsEnabled = !moduleEntryVertex.ModuleViewModel.IsInternalExternal };
                openWsmMenuItem.Click += OpenWsmDoNotMeldRuleVertices;
                contextMenu.Items.Add(openWsmMenuItem);

                openWsmMenuItem = new MenuItem() { Header = "Show WSM Evaluation History", DataContext = vertexControl, IsEnabled = !moduleEntryVertex.ModuleViewModel.IsInternalExternal };
                openWsmMenuItem.Click += ShowWsmHistory;
                contextMenu.Items.Add(openWsmMenuItem);
            }
        }

        public void ShowWsmHistory(object sender, RoutedEventArgs e)
        {
            var menuItem = sender as MenuItem;
            var vertexControl = menuItem.DataContext as VertexControl;
            var vertex = vertexControl.Vertex as ModuleEntryVertex;

            if (!ModuleRepository.Instance.Modules.ContainsKey(vertex.Name))
            {
                MessageBox.Show("The module is non-existent in the module repository - does it exist? Or maybe it was removed in the meantime or appears erroneous now?", "Module not found in the repository.", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            var tabController = MainWindowViewModel.Instance.MainWindowView.TabController;
            LayoutDocument document = null;
            HistoryView view = null;
            string errorMessage = null;
            var result = ProgressDialog.Execute(Application.Current.MainWindow, "Loading WSM History data for the module", () =>
            {
                try
                {
                    ProgressDialog.Current.Report("Building the History table...");

                    var module = ModuleRepository.Instance.Modules[vertex.Name];
                    // var domainModule = ModuleRepository.Instance.Modules.ContainsKey(domainModuleName) ? ModuleRepository.Instance.Modules[domainModuleName] : null;

                    var wsmData = ApplicationClient.Instance.InterpreterInstance.GetWSMHistory(module.Name);
                    // vertex.ModuleViewModel.FileDir + vertex.ModuleViewModel.FileName + 
                    var tabName = "#wsmhistory-" + module.Name;
                    document = tabController.GetTab(tabName);

                    Application.Current.Dispatcher.Invoke(() =>
                    {
                        if (document != null)
                        {
                            ProgressDialog.Current.Report("Closing the previous tab...");
                            document.Close();
                        }

                        ProgressDialog.Current.Report("Creating a view for the table...");
                        view = new HistoryView(wsmData, sender, this);
                        document = tabController.createTab(tabName, view);
                        document.Title = tabName;
                        document.ContentId = tabName;
                    });
                }
                catch (Exception ex)
                {
                    errorMessage = ex.Message;
                }
            }, ProgressDialogSettings.WithSubLabel);

            if (errorMessage != null)
            {
                MessageBox.Show("Failed to display the WSM history: " + errorMessage, "WSM history creation failed");
            }
            else
            {
                Application.Current.Dispatcher.Invoke(() =>
                {
                    tabController.selectTab(document);
                });
            }
        }

        public void OpenWsmMeldRuleVertices(object sender, RoutedEventArgs e)
        {
            ShowWsmGraph(sender, true);
        }

        void OpenWsmDoNotMeldRuleVertices(object sender, RoutedEventArgs e)
        {
            ShowWsmGraph(sender, false);
        }

        private void ShowWsmGraph(object sender, bool meldRuleVertices)
        {
            var menuItem = sender as MenuItem;
            var vertexControl = menuItem.DataContext as VertexControl;
            var vertex = vertexControl.Vertex as ModuleEntryVertex;

            if (!ModuleRepository.Instance.Modules.ContainsKey(vertex.Name))
            {
                MessageBox.Show("The module is non-existent in the module repository - does it exist? Or maybe it was removed in the meantime or appears erroneous now?", "Module not found in the repository.", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            var tabController = MainWindowViewModel.Instance.MainWindowView.TabController;
            LayoutDocument document = null;
            WsmView view = null;
            string errorMessage = null;
            var result = ProgressDialog.Execute(Application.Current.MainWindow, "Loading WSM data for the module", () =>
            {
                try
                {
                    ProgressDialog.Current.Report("Building the WSM graph...");

                   // var domainModuleName = DomainModuleBuilder.ToDomainModule(vertex.Name);
                    var module = ModuleRepository.Instance.Modules[vertex.Name];
                   // var domainModule = ModuleRepository.Instance.Modules.ContainsKey(domainModuleName) ? ModuleRepository.Instance.Modules[domainModuleName] : null;

                    var wsmData = WsmDataBuilder.Build(module);
                    var graph = WsmGraphBuilder.Build(module, wsmData, meldRuleVertices);
                    // vertex.ModuleViewModel.FileDir + vertex.ModuleViewModel.FileName +
                    var tabName =   "#graph-" + module.Name;
                    document = tabController.GetTab(tabName);

                    Application.Current.Dispatcher.Invoke(() =>
                    {
                        if (document != null)
                        {
                            ProgressDialog.Current.Report("Closing the previous graph...");
                            document.Close();
                        }

                        ProgressDialog.Current.Report("Creating a view for the graph...");
                        view = new WsmView(wsmData, graph);
                        document = tabController.createTab(tabName, view);
                        document.Title = tabName;
                        document.ContentId = tabName;
                    });
                }
                catch (Exception ex)
                {
                    errorMessage = ex.Message;
                }
            }, ProgressDialogSettings.WithSubLabel);

            if (errorMessage != null)
            {
                MessageBox.Show("Failed to display the WSM graph: " + errorMessage, "WSM graph creation failed");
            }
            else
            {
                Application.Current.Dispatcher.Invoke(() =>
                {
                    tabController.selectTab(document);
                });
            }
        }
    }
}
