﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Media;
using FourqlModeler.Calculcations;
using FourqlModeler.Communication.Import;
using FourqlModeler.Utility;
using FourqlModeler.ViewModels;
using FourqlModeler.Views.GraphExtensions;
using FourqlModeler.Views.GraphExtensions.ExprGraph;
using FourqlModeler.Views.GraphExtensions.ModuleGraph;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using GraphX;
using FourqlParser.StaticAnalysis;
using Parago.Windows;
using System.IO;
using System.Windows.Markup;
using System.Text;
using System.Reflection;
using Module = FourqlParser.Structures.Module;

namespace FourqlModeler.Views
{
    /// <summary>
    /// Interaction logic for ResultView.xaml
    /// </summary>
    public partial class ResultView : UserControl
    {
        private ICollection<ExtendedVertex> ConstructionVertices { get; set; }

        /// <summary>
        /// Holds a subgraph for every extendable vertex
        /// </summary>
        private IDictionary<ExtendedVertex, RowSubgraph> GraphsForVertices { get; set; }

        /// <summary>
        /// holds a subgraph for every row of the result
        /// </summary>
        private IDictionary<Row, RowSubgraph> GraphsForRows { get; set; }
        private ICollection<ExtendedVertex> ExplainedVertices { get; set; }

        public ResultView(Result result, Expression expr, string execModule)
        {
            InitializeComponent();

            if (result.Rows.Count == 0)
            {
                result.AddDummyRow();
            }

            ConstructionVertices = new HashSet<ExtendedVertex>();
            GraphsForVertices = new Dictionary<ExtendedVertex, RowSubgraph>();
            GraphsForRows = new Dictionary<Row, RowSubgraph>();
            ExplainedVertices = new HashSet<ExtendedVertex>();

            AddResult(result, expr, execModule, null);
            ApplyResultToGrid(result, expr, null);
            ExplanationGraphView.GraphArea.VertexSelected += GraphArea_VertexSelected;
            ExplanationGraphView.GraphArea.VertexDoubleClick += GraphArea_VertexDoubleClick;
        }

        /// <summary>
        /// Handles selection of a vertex in a result graph
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        void GraphArea_VertexSelected(object sender, GraphX.Models.VertexSelectedEventArgs args)
        {
            var extendedVertex = args.VertexControl.Vertex as ExtendedVertex;
            if (extendedVertex == null)
            {
                return;
            }
            var quantifierVertex = extendedVertex as QuantifierVertex;
            if (quantifierVertex != null)
            {
                QuantifierVertexSelect(quantifierVertex, args);
            }
            var variableVertex = extendedVertex as QuantifierVariableVertex;
            if (variableVertex != null)
            {
                QuantifierVariableVertexSelect(variableVertex, args);
            }
            // check if we already created a graph for selection of this vertex
            if (GraphsForVertices.ContainsKey(extendedVertex))
            {
                var rowSubgraph = GraphsForVertices[extendedVertex];
                // apply the selection of the vertex to the result table
                if (ResultViewModel.Result != rowSubgraph.ResultDesc.Result)
                {
                    ApplyResultToGrid(rowSubgraph.ResultDesc.Result, rowSubgraph.ResultDesc.Expression, rowSubgraph.Row);
                }
                //????
                var deeperDeps = new HashSet<ExtendedVertex>();
                foreach (var vertexEntry in GraphsForVertices)
                {
                    var status = ExtendedVertexStatus.None;
                    if (vertexEntry.Key == rowSubgraph.ResultDesc.Root)
                    {
                        status = ExtendedVertexStatus.Root;
                    }
                    else if (vertexEntry.Value.ResultDesc.Root == rowSubgraph.ResultDesc.Root)
                    {
                        status = ConstructionVertices.Contains(vertexEntry.Key) ? ExtendedVertexStatus.Dependency : ExtendedVertexStatus.Selected;
                        if (status == ExtendedVertexStatus.Dependency && ExplanationGraphView.GraphArea.VertexList.ContainsKey(vertexEntry.Key))
                        {
                            // Mark deeper dependencies.
                            foreach (var edge in ExplanationGraphView.GraphArea.LogicCore.Graph.OutEdges(vertexEntry.Key))
                            {
                                deeperDeps.Add(edge.Target);
                            }
                        }
                    }

                    if (status != vertexEntry.Key.Status)
                    {
                        vertexEntry.Key.Status = status;
                        if (ExplanationGraphView.GraphArea.VertexList.ContainsKey(vertexEntry.Key))
                        {
                            var vertexControl = ExplanationGraphView.GraphArea.VertexList[vertexEntry.Key];
                            vertexControl.DataContext = null;
                            vertexControl.DataContext = vertexEntry.Key;
                        }
                    }
                }
                foreach (var deeperDep in deeperDeps)
                {
                    deeperDep.Status = ExtendedVertexStatus.DeeperDependency;
                    if (ExplanationGraphView.GraphArea.VertexList.ContainsKey(deeperDep))
                    {
                        var vertexControl = ExplanationGraphView.GraphArea.VertexList[deeperDep];
                        vertexControl.DataContext = null;
                        vertexControl.DataContext = deeperDep;
                    }
                }
            }
        }

        private void QuantifierVariableVertexSelect(QuantifierVariableVertex vertex, GraphX.Models.VertexSelectedEventArgs args)
        {
            if (!ExplainedVertices.Contains(vertex.Parent))
                QuantifierVertexSelect(vertex.Parent, args);
        }



        void FileVertexDoubleClick(FileVertex fileVertex)
        {
            if (fileVertex != null)
            {
                MainWindowViewModel.Instance.MainWindowView.OpenModuleEvent(fileVertex.ModuleViewModel);
                return;
            }
        }

        /// <summary>
        /// Check if two literals 'match'. <br/>
        /// Literals match if:<br/>
        /// - they have the same name<br/>
        /// - they have the same constant values<br/>
        /// - variables always have the same value<br/>
        ///  (i.e. A(X, 2, X) matches A(1, 2, 1), but not A(3,2,1), A(1, 3, 1) or B(1,2,1)
        /// </summary>
        /// <param name="left"></param>
        /// <param name="right"></param>
        /// <returns></returns>
        private static bool LiteralsMatch(Literal left, Literal right)
        {
            if (!left.RelationName.Equals(right.RelationName))
            {
                return false;
            }

            if (left.Values.Count != right.Values.Count)
            {
                return false;
            }

            var lIter = left.Values.GetEnumerator();
            var rIter = right.Values.GetEnumerator();
            var foundValues = new Dictionary<string, ValueHolder>();
            while (lIter.MoveNext() && rIter.MoveNext())
            {
                // 
                if (lIter.Current.Type != DomainType.VARIABLE)
                {
                    if (!lIter.Current.Equals(rIter.Current))
                    {
                        return false;
                    }
                } // 
                else if (foundValues.ContainsKey(lIter.Current.valueAs<string>()))
                {
                    if (!foundValues[lIter.Current.valueAs<string>()].Equals(rIter.Current))
                    {
                        return false;
                    }
                }
                else
                {
                    foundValues.Add(lIter.Current.valueAs<string>(), rIter.Current);
                }
            }

            return true;
        }

        void CreateFactVerticesForInternalModule(LiteralVertex literalVertex, Module execModule, IList<ExtendedVertex> vertices, List<ExtendedEdge> edges)
        {
            ProgressDialog.Current.ReportWithCancellationCheck("The literal comes from an internal/external module - generating facts...");
            var fact = new Fact()
            {
                RelationName = literalVertex.Literal.RelationName, 
                ParserContext = literalVertex.Literal.ParserContext,
                Value = literalVertex.LogicValue.GetValueOrDefault(LogicValue.UNKNOWN)
            };
            foreach (var param in literalVertex.Literal.Values)
            {
                fact.Values.Add(param);
            }
            var factVertex = new FactVertex()
            {
                Pattern = String.Format("{0}.{1}", execModule.Name, new ExprPrinter("", false).visitLiteral(literalVertex.Literal).ToString()), 
                Value = literalVertex.LogicValue.GetValueOrDefault(LogicValue.UNKNOWN), 
                Fact = fact, 
                ToolTip = String.Format("This fact provides a value for the literal: {0}, but comes from an internal/external module and thus is calculated by a side program.", new ExprPrinter().visitLiteral(literalVertex.Literal).ToString())
            };
            vertices.Add(factVertex);
            edges.Add(
                new ExtendedEdge(literalVertex, factVertex)
            {
                Label = literalVertex.LogicValue.ToString(), 
                ToolTip = "The value that comes from the internal/external module fact and provides it to the literal."
            });
       }

        private void CreateFactVertices(LiteralVertex literalVertex, Module execModule, List<ExtendedVertex> constructionVertices, List<ExtendedEdge> constructionEdges)
        {
            // select facts matching the literal
            var facts = execModule.Facts.Where(fact => fact.SignatureEquals(literalVertex.Literal)).ToList();

            var constructionId = 0;
            foreach (var fact in facts)
            {
                ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Adding facts to the construction set ({0}/{1})...", ++constructionId, facts.Count));
                var factVertex = new FactVertex() { Pattern = String.Format("{0}.{1}", execModule.Name, new ExprPrinter("", false).visitLiteral(fact).ToString()), Value = fact.IsNegative ? LogicValue.FALSE : LogicValue.TRUE, Fact = fact, ToolTip = String.Format("This fact provides a value for the literal: {0}", new ExprPrinter().visitLiteral(literalVertex.Literal).ToString()) };
                constructionVertices.Add(factVertex);
                constructionEdges.Add(new ExtendedEdge(literalVertex, factVertex) { Label = fact.IsNegative ? LogicValue.FALSE.ToString() : LogicValue.TRUE.ToString(), ToolTip = "The value that comes from the fact and provides it to the literal." });
            }
        }


        private void CreateRuleFactVertices(LiteralVertex literalVertex, Module execModule, List<ExtendedVertex> constructionVertices, List<ExtendedEdge> constructionEdges)
        {
            // select rules which have head matching the literal
            var rules = execModule.Rules.Where(rule => LiteralsMatch(rule.Head.CreateLiteralOfFact(), literalVertex.Literal));

            var exprExec = new CalcExprExecutor();
            var constructionId = 0;
            foreach (var rule in rules)
            {
                ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Adding rule facts to the construction set from the found rules ({0}/{1})...", ++constructionId, rules.Count()));
                var values = new Dictionary<string, ValueHolder>();
                var literalIter = literalVertex.Literal.Values.GetEnumerator();
                var headIter = rule.Head.Values.GetEnumerator();

                // gather variable substitutions from head with values from the literal
                while (literalIter.MoveNext() && headIter.MoveNext())
                {
                    if (headIter.Current.Type == DomainType.VARIABLE && !values.ContainsKey(headIter.Current.valueAs<string>()))
                    {
                        values.Add(headIter.Current.valueAs<string>(), literalIter.Current);
                    }
                }
                var setter = new ExprValueSetter(values);
                // replace the variables in the body with previously gathered values
                var ruleWithReplacements = new Rule() { ParserContext = rule.ParserContext, Head = setter.SetInFact(rule.Head), Body = rule.Body.Accept(setter) };
                // execute the body of the rule with substitutions
                var result = exprExec.Execute(ruleWithReplacements, execModule, ModuleRepository.Instance.Modules);
                Row chosenRow = null;
                foreach (var row in result.Rows)
                {
                    if (chosenRow == null || Logic.or(chosenRow.Value, row.Value) != chosenRow.Value)
                    {
                        chosenRow = row;
                    }
                }
                if (chosenRow == null)
                {
                    chosenRow = result.AddDummyRow();
                }
                var value = chosenRow.Value == LogicValue.TRUE ? (rule.Head.IsNegative ? LogicValue.FALSE : LogicValue.TRUE) : (chosenRow.Value == LogicValue.INCONSISTENT ? LogicValue.INCONSISTENT : LogicValue.UNKNOWN);
                // check if we should mark the edge as not important to the result 
                var nonImportant = chosenRow.Value != LogicValue.TRUE && chosenRow.Value != LogicValue.INCONSISTENT;
                // create a new vertex
                var ruleVertex = new RuleFactVertex()
                {
                    ModuleName = execModule.Name, 
                    Rule = rule, 
                    Fact = ruleWithReplacements.Head, 
                    Pattern = String.Format("{0}.{1}", execModule.Name, new ExprPrinter("", false).visitLiteral(ruleWithReplacements.Head).ToString()), 
                    ToolTip = String.Format("The fact for the literal {0} from the rule\n{1}\nAnd with replacements:\n{2}", new ExprPrinter().visitLiteral(literalVertex.Literal), new ProgramPrinter(false).append(rule).Done(), new ProgramPrinter(false).append(ruleWithReplacements).Done()), 
                    Value = value
                };
                constructionVertices.Add(ruleVertex);
                // add the edge between the literal and the rule
                constructionEdges.Add(new ExtendedEdge(literalVertex, ruleVertex)
                {
                    NotImportant = nonImportant, 
                    Label = nonImportant ? LogicValue.UNKNOWN.ToString() : value.ToString(), 
                    ToolTip = "The value that comes from the rule."
                });
            }
         
        }

        void AddConstructionsToGraph(LiteralVertex literalVertex, List<ExtendedVertex> constructionVertices, List<ExtendedEdge> constructionEdges)
        {
            ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Adding new constructions to the graph..."));
            ExplainedVertices.Add(literalVertex);
            foreach (var newVertex in constructionVertices)
            {
                ExplanationGraphView.GraphArea.PrototypeGraph.AddVertex(newVertex);
            }
            foreach (var newEdge in constructionEdges)
            {
                ExplanationGraphView.GraphArea.PrototypeGraph.AddEdge(newEdge);
            }
        }

        void LiteralVertexDoubleClick(LiteralVertex literalVertex, GraphX.Models.VertexSelectedEventArgs args)
        {
            if (literalVertex == null) return;

            if (ExplainedVertices.Contains(literalVertex))
            {
                System.Windows.MessageBox.Show(String.Format("The literal has already been explained. If its state changed, you should re-create the explanation."), "Already explained", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Information);
                DragBehaviour.SetIsDragging(args.VertexControl, false);
                return;
            }

            if (!ModuleRepository.Instance.Modules.ContainsKey(literalVertex.ModuleName))
            {
                System.Windows.MessageBox.Show(String.Format("The module ({0}) for the literal does not exist in the module repository or it might be erroneous. You can check all loaded modules via TOOLS -> Show Module Graph", literalVertex.ModuleName), "Module non-existent", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Error);
                DragBehaviour.SetIsDragging(args.VertexControl, false);
                return;
            }

            if (GraphsForVertices[literalVertex].Row.IsDummyRow)
            {
                System.Windows.MessageBox.Show(String.Format("The literal was defined by a dummy row as some of its variables could not be given proper values. You cannot explore this literal any deeper.", literalVertex.ModuleName), "Literal for a dummy row cannot be explained", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Information);
                DragBehaviour.SetIsDragging(args.VertexControl, false);
                return;
            }

            System.Windows.MessageBoxImage dialogImage = System.Windows.MessageBoxImage.None;
            string errorHeader = null;
            string errorMessage = null;
            ProgressDialog.Execute(App.Current.MainWindow, "Finding related facts and rules for the literal...", () =>
            { // Create an explanation graph for the literal

                ProgressDialog.Current.ReportWithCancellationCheck("Finding related facts and rules for the literal...");

                // select the module
                var execModule = ModuleRepository.Instance.Modules[literalVertex.ModuleName];

                // prepare the lists of vertices and edges
                var constructionVertices = new List<ExtendedVertex>();
                var constructionEdges = new List<ExtendedEdge>();
                if (execModule.IsInternalExternal)
                {
                    // generate vertices for 'facts' from an internal module
                    CreateFactVerticesForInternalModule(literalVertex, execModule, constructionVertices, constructionEdges);
                }
                else
                {
                    ProgressDialog.Current.ReportWithCancellationCheck("The literal comes from a local module - looking for facts and rules...");
                    // generate vertices for facts corresponding to the literal
                    CreateFactVertices(literalVertex, execModule, constructionVertices, constructionEdges);

                    try
                    {
                        //generate vertices for rules which head matches the literal
                        CreateRuleFactVertices(literalVertex, execModule, constructionVertices, constructionEdges);
                    } catch (Exception ex)
                    {
                        errorHeader = "Error while trying to analyze a rule";
                        errorMessage = ex.Message;
                        dialogImage = System.Windows.MessageBoxImage.Error;
                        return;
                    }
                    
                }

                if (constructionVertices.Count == 0)
                {
                    errorHeader = "No construction found";
                    errorMessage = String.Format("No construction found for the literal in its module: {0}", execModule.Name);
                    dialogImage = System.Windows.MessageBoxImage.Information;
                    return;
                }
                 
                // Add new constructions.
               AddConstructionsToGraph(literalVertex, constructionVertices, constructionEdges);

                // RULES SHARE THE LITERAL RESULTS. BECAUSE WE EXPLAIN EXPRESSIONS AND IF WE WANT TO SEE HOW RULE FACTS COME AND GO,
                // WE SHOULD SIMPLY USE A LITERAL THAT REPRESENTS INTERESTING RESULTS AND EXPLAIN FROM THERE.
                ProgressDialog.Current.ReportWithCancellationCheck(String.Format("Apply the changes to the graph..."));
                App.Current.Dispatcher.Invoke(() =>
                {
                    AddedExplanationConstructions(literalVertex, execModule.Name, constructionVertices);
                    ExplanationGraphView.GraphArea.GenerateGraph(ExplanationGraphView.GraphArea.PrototypeGraph, true);
                    ExplanationGraphView.GraphArea.UpdateControls();
                    ExplanationGraphView.GraphArea.UpdateExpansion();
                });
            });

            if (errorMessage != null)
            {
                System.Windows.MessageBox.Show(errorMessage, errorHeader, System.Windows.MessageBoxButton.OK, dialogImage);
                DragBehaviour.SetIsDragging(args.VertexControl, false);
            }
            return;
        }

       
        void RuleFactVertexDoubleClick(RuleFactVertex ruleFactVertex, GraphX.Models.VertexSelectedEventArgs args)
        {
            if (ruleFactVertex != null)
            {
                if (ExplainedVertices.Contains(ruleFactVertex))
                {
                    System.Windows.MessageBox.Show(String.Format("The rule fact has already been explained. If its state changed, you should re-create the explanation."), "Already explained", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Information);
                    DragBehaviour.SetIsDragging(args.VertexControl, false);
                    return;
                }

                if (!ModuleRepository.Instance.Modules.ContainsKey(ruleFactVertex.ModuleName))
                {
                    System.Windows.MessageBox.Show(String.Format("The module ({0}) for the rule fact does not exist in the module repository or it might be erroneous. You can check all loaded modules via TOOLS -> Show Module Graph", ruleFactVertex.ModuleName), "Module non-existent", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Error);
                    DragBehaviour.SetIsDragging(args.VertexControl, false);
                    return;
                }

                string errorMessage = null;
                ProgressDialog.Execute(App.Current.MainWindow, "Explaining the rule fact...", () =>
                {
                    ProgressDialog.Current.ReportWithCancellationCheck("Preparing a map of values to explain...");
                    var execModule = ModuleRepository.Instance.Modules[ruleFactVertex.ModuleName];
                    // Make replacements in the rules according to the rule fact.
                    var values = new Dictionary<string, ValueHolder>();
                    var literalIter = ruleFactVertex.Fact.Values.GetEnumerator();
                    var headIter = ruleFactVertex.Rule.Head.Values.GetEnumerator();
                    while (literalIter.MoveNext() && headIter.MoveNext())
                    {
                        if (headIter.Current.Type == DomainType.VARIABLE && !values.ContainsKey(headIter.Current.valueAs<string>()))
                        {
                            values.Add(headIter.Current.valueAs<string>(), literalIter.Current);
                        }
                    }
                    ProgressDialog.Current.ReportWithCancellationCheck("Making proper replacements in the fact rule...");
                    var setter = new ExprValueSetter(values);
                    var ruleWithReplacements = new Rule() { ParserContext = ruleFactVertex.Rule.ParserContext, Head = setter.SetInFact(ruleFactVertex.Rule.Head), Body = ruleFactVertex.Rule.Body.Accept(setter) };

                    var exprCalc = new CalcExprExecutor();

                     
                    Result result;
                    try
                    {
                         result = exprCalc.Execute(new ExprPrinter("", false, "or", " and", "not ").ConvertExpr(ruleWithReplacements.Body).ToString(), execModule, ModuleRepository.Instance.Modules);
                   
                    }
                    catch (Exception ex)
                    {
                        errorMessage = ex.Message;
                        return;
                    }

                    ProgressDialog.Current.ReportWithCancellationCheck("Adding graph constructions for the rule result.");

                    var vertexValue = ruleFactVertex.Value;
                    if (ruleFactVertex.Rule.Head.IsNegative)
                        vertexValue = Logic.not(vertexValue);

                    Row chosenRow = null;
                    foreach (var row in result.Rows)
                    {
                        if (chosenRow == null && (row.Value.Equals(vertexValue)))
                        {
                            chosenRow = row;
                        }
                    }

                    if (chosenRow == null)
                    {
                        chosenRow = result.AddDummyRow();
                    }

                    // Add the result in form of a graph.
                    App.Current.Dispatcher.Invoke(() =>
                    {
                        HandleExpandedResult(ruleFactVertex, execModule.Name, ruleWithReplacements.Body, result, chosenRow, ruleFactVertex, ruleFactVertex);
                    });
                    //}

                    ExplainedVertices.Add(ruleFactVertex);

                    // Update the graph.
                   // ProgressDialog.Current.Report("Updating the graph after changes...");
                    App.Current.Dispatcher.Invoke(() =>
                    {
                        //ExplanationGraphView.GraphArea.GenerateGraph(ExplanationGraphView.GraphArea.PrototypeGraph, true);
                        ExplanationGraphView.GraphArea.UpdateControls();
                        ExplanationGraphView.GraphArea.UpdateExpansion();
                    });
                });

                if (errorMessage != null)
                {
                    System.Windows.MessageBox.Show(errorMessage, "Error while trying to analyze an alternative", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Error);
                    DragBehaviour.SetIsDragging(args.VertexControl, false);
                }
            }
        }

        void QuantifierVertexSelect(QuantifierVertex vertex, GraphX.Models.VertexSelectedEventArgs args)
        {
            if (vertex != null)
            {
                if (ExplainedVertices.Contains(vertex))
                {
                    return;
                }

                if ((vertex.VariableValue == null))
                {
                    return;
                }

                string errorMessage = null;
                //ProgressDialog.Execute(App.Current.MainWindow, "Explaining the quantifier...", () =>
               // {

                var exprCalc = new CalcExprExecutor();
                var result = exprCalc.Execute(new ExprPrinter("", false).ConvertExpr(vertex.SubExpression).ToString(), ModuleRepository.Instance.Modules[vertex.ModuleName], ModuleRepository.Instance.Modules);

                var domainValuesRes = exprCalc.Execute(String.Format("{0}(X)", vertex.VariableDomain),
                    ModuleRepository.Instance.Modules[vertex.ModuleName], ModuleRepository.Instance.Modules);

                var domainValues = from r in domainValuesRes.Rows
                    where !r.IsDummyRow && r.Value == LogicValue.TRUE && domainValuesRes.ToFieldMap(r).ContainsKey("X")
                    select domainValuesRes.ToFieldMap(r)["X"];

                exprCalc.FilterResult(result, vertex.VariableName, domainValues);
                    
                    Row chosenRow = null;
                    foreach (var row in result.Rows)
                    {
                        var map = result.ToFieldMap(row);
                        if ((chosenRow == null) || ((vertex.VariableValue != null) && map[vertex.VariableName].Equals(vertex.VariableValue.Value)))
                        {
                            chosenRow = row;
                        }
                    }

                    if (chosenRow == null)
                    {
                        chosenRow = result.AddDummyRow();
                    }


                    ExplainedVertices.Add(vertex);

                    App.Current.Dispatcher.Invoke(() =>
                    {
                        HandleExpandedResult(vertex, vertex.ModuleName, vertex.SubExpression, result, chosenRow, vertex, vertex);
                    });
               // });
                if (errorMessage != null)
                {
                    System.Windows.MessageBox.Show(errorMessage, "Error while trying to analyze the quantifier", System.Windows.MessageBoxButton.OK, System.Windows.MessageBoxImage.Error);
                    DragBehaviour.SetIsDragging(args.VertexControl, false);
                }
            }
        }


        void GraphArea_VertexDoubleClick(object sender, GraphX.Models.VertexSelectedEventArgs args)
        {
            // if selected vertex is a module -> open it in the editor
            var fileVertex = args.VertexControl.Vertex as FileVertex;
            if (fileVertex != null)
            {
                FileVertexDoubleClick(fileVertex);
                return;
            }

            // if selected vertex is a literal -> explain it
            var literalVertex = args.VertexControl.Vertex as LiteralVertex;
            if (literalVertex != null)
            {
                LiteralVertexDoubleClick(literalVertex, args);
                return;
            }

            // if selected vertex is a rule fact -> ???
            var ruleFactVertex = args.VertexControl.Vertex as RuleFactVertex;
            if (ruleFactVertex != null)
            {
                RuleFactVertexDoubleClick(ruleFactVertex, args);
                return;
            }
        }

        private void HandleExpandedResult(ExtendedVertex root, string moduleName, Expression expression, Result result, Row selectedRow, ExtendedVertex setResultFor, ExtendedVertex selectVertexAfter)
        {
            var resultDesc = AddResult(result, expression, moduleName, root);
            if (setResultFor != null && !GraphsForVertices.ContainsKey(setResultFor))
            {
                GraphsForVertices.Add(setResultFor, new RowSubgraph() { ResultDesc = resultDesc, Row = selectedRow });
            }
            SelectRow(selectedRow);

            if (selectVertexAfter != null && ExplanationGraphView.GraphArea.VertexList.ContainsKey(selectVertexAfter))
            {
                GraphArea_VertexSelected(null, new GraphX.Models.VertexSelectedEventArgs(ExplanationGraphView.GraphArea.VertexList[selectVertexAfter], null, System.Windows.Input.ModifierKeys.None));
            }
        }

        private void AddedExplanationConstructions(ExtendedVertex literalVertex, string moduleName, ICollection<ExtendedVertex> constructionVertices, bool updateSelection = false)
        {
            if (GraphsForVertices.ContainsKey(literalVertex))
            {
                var subgraph = GraphsForVertices[literalVertex];
                foreach (var constructionVertex in constructionVertices)
                {
                    ConstructionVertices.Add(constructionVertex);
                    GraphsForVertices.Add(constructionVertex, subgraph);
                    constructionVertex.Status = ExtendedVertexStatus.Dependency;
                    if (ExplanationGraphView.GraphArea.VertexList.ContainsKey(constructionVertex))
                    {
                        var vertexControl = ExplanationGraphView.GraphArea.VertexList[constructionVertex];
                        vertexControl.DataContext = null;
                        vertexControl.DataContext = constructionVertex;
                    }
                }

                if (updateSelection && ExplanationGraphView.GraphArea.VertexList.ContainsKey(literalVertex))
                {
                    GraphArea_VertexSelected(null, new GraphX.Models.VertexSelectedEventArgs(ExplanationGraphView.GraphArea.VertexList[literalVertex], null, System.Windows.Input.ModifierKeys.None));
                }
            }
        }

        /// <summary>
        /// Adds result to the result view
        /// </summary>
        /// <param name="result">Result to be added</param>
        /// <param name="expr">Expression which was evaluated</param>
        /// <param name="execModule">Module/context of the expression</param>
        /// <param name="root">root vertex of the result</param>
        /// <returns></returns>
        private ResultDesc AddResult(Result result, Expression expr, string execModule, ExtendedVertex root)
        {
            var resultDesc = new ResultDesc() { Result = result, ModuleName = execModule, Expression = expr, Root = root };
            foreach (var row in result.Rows)
            {
                // add a graph for each row of the result
                GraphsForRows.Add(row, new RowSubgraph() { Row = row, ResultDesc = resultDesc });
            }
            return resultDesc;
        }

        /// <summary>
        /// Shows the result in the table
        /// </summary>
        /// <param name="result"></param>
        /// <param name="expression"></param>
        /// <param name="selectedRow"></param>
        private void ApplyResultToGrid(Result result, Expression expression, Row selectedRow)
        {
            ResultViewModel.Result = result;
            ResultViewModel.CurrentExpression = new ExprPrinter("", false).ConvertExpr(expression).ToString();
            ResultGrid.SelectionChanged -= ResultGrid_SelectionChanged;
            ResultGrid.Columns.Clear();

            var iter = result.Fields.GetEnumerator();
            var fieldCount = result.Fields.Count + 1;
            var fieldId = 0;
            var isDummy = result.Rows.Count == 1 && result.Rows[0].IsDummyRow;
            while (iter.MoveNext())
            {
                var assemblyName = Assembly.GetExecutingAssembly().GetName().Name;
                var xaml = new StringBuilder().AppendLine("<DataTemplate ")
                    .AppendLine("  xmlns=\"http://schemas.microsoft.com/winfx/2006/xaml/presentation\"")
                    .AppendLine("  xmlns:x=\"http://schemas.microsoft.com/winfx/2006/xaml\"")
                    .AppendLine("  xmlns:converters=\"clr-namespace:FourqlModeler.Views.ValueConverters;assembly=" + assemblyName + "\">")
                    .AppendLine("<DataTemplate.Resources>")
                    .AppendLine("    <converters:ResultDataConverter x:Key=\"DataConverter\" />")
                    .AppendLine("</DataTemplate.Resources>")
                    .AppendLine("    <TextBlock TextAlignment=\"Left\" Padding=\"3,0,0,0\" VerticalAlignment=\"Center\" Text=\"{Binding Path=Values[" + fieldId++ + "], Converter={StaticResource DataConverter}, ConverterParameter=" + (isDummy ? DomainType.VARIABLE.ToString() : iter.Current.Type.ToString()) + "}\" />")
                    .AppendLine("</DataTemplate>").ToString();
                ResultGrid.Columns.Add(new DataGridTemplateColumn()
                {
                    Header = String.Format("{0} ({1})", iter.Current.Name, DomainTypeUtils.getName(iter.Current.Type)),
                    Width = new DataGridLength(1.0 / fieldCount, DataGridLengthUnitType.Star),
                    CellTemplate = (System.Windows.DataTemplate)XamlReader.Parse(xaml)
                });
                //ResultGrid.Columns.Add(new DataGridTextColumn() { Header = String.Format("{0} ({1})", iter.Current.Name, DomainTypeUtils.getName(iter.Current.Type)), Width = new DataGridLength(1.0 / fieldCount, DataGridLengthUnitType.Star), Binding = new Binding("Values[" + fieldId++ + "]") });
            }
            ResultGrid.Columns.Add(new DataGridTextColumn() { Header = "Value", Width = new DataGridLength(1.0, DataGridLengthUnitType.Star), Binding = new Binding("Value") });
            if (selectedRow != null)
            {
                ResultGrid.SelectedItem = selectedRow;
            }
            ResultGrid.SelectionChanged += ResultGrid_SelectionChanged;
        }

        private void ResultGrid_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (e.AddedItems.Count == 0)
            {
                return;
            }

            SelectRow(e.AddedItems[0] as Row);
        }

        private void BuildSubgraphForRow(Row selection, RowSubgraph rowSubgraph)
        {
            var values = new Dictionary<string, object>();
            var types = new Dictionary<string, DomainType>();
            var valueEnum = selection.Values.GetEnumerator();

            // fill the rowSubgraph with values from the selected row
            foreach (var field in rowSubgraph.ResultDesc.Result.Fields)
            {
                valueEnum.MoveNext();
                values.Add(field.Name, valueEnum.Current);
                types.Add(field.Name, field.Type);
            }

            //create a new expressionn for the subgraph, substituting values for the variables in the original expression
            var exprWithReplacements = ExprUtils.WithReplacements(rowSubgraph.ResultDesc.Expression, types, values);

            //build a new graph for the expression
            rowSubgraph.Graph = ExprGraphBuilder.Build(exprWithReplacements, rowSubgraph.ResultDesc.ModuleName, selection.Explanation);


            foreach (var vertex in rowSubgraph.Graph.Vertices)
            {
                GraphsForVertices.Add(vertex, rowSubgraph);
            }
            // connect the graph to the root of the result
            if (rowSubgraph.ResultDesc.Root != null)
            {
                var isImportant = IsValueImportant(rowSubgraph.ResultDesc.Root, selection.Value);
                

                if (rowSubgraph.ResultDesc.Root is QuantifierVertex)
                {
                    AddQuantifierVariableVertex(rowSubgraph, rowSubgraph.ResultDesc.Root as QuantifierVertex, selection);
                }

                rowSubgraph.Graph.AddVertex(rowSubgraph.ResultDesc.Root);

                foreach (var vertex in rowSubgraph.Graph.Vertices)
                {
                    if (vertex != rowSubgraph.ResultDesc.Root && rowSubgraph.Graph.InDegree(vertex) == 0)
                    {
                        rowSubgraph.Graph.AddEdge(new ExtendedEdge(rowSubgraph.ResultDesc.Root, vertex) { Label = selection.Value.ToString(), NotImportant = !isImportant, ToolTip = "The edge connects a vertex with the calculations that support the result." });
                    }
                }
            }
        }

        private void AddQuantifierVariableVertex(RowSubgraph rowSubgraph, QuantifierVertex quantifierVertex, Row selection)
        {
            var graph = rowSubgraph.Graph;
            var result = rowSubgraph.ResultDesc.Result;
            var variableValue = result.ToFieldMap(selection)[quantifierVertex.VariableName];
            var vertex = new QuantifierVariableVertex(quantifierVertex)
            {
                LogicValue = selection.Value,
                VariableValue = variableValue,
                Pattern = QuantifierVariableVertex.MakePattern(quantifierVertex.VariableName, quantifierVertex.DomainModule + "." + quantifierVertex.VariableDomain, variableValue)
            };
            GraphsForVertices[vertex] = rowSubgraph;
            graph.AddVertex(vertex);
            foreach (var v in graph.Vertices)
            {
                if (v != vertex && graph.InDegree(v) == 0)
                {
                    graph.AddEdge(new ExtendedEdge(vertex, v) { Label = selection.Value.ToString(), ToolTip = "The edge connects a variable value selection for the quantifier with the calculations that support the result." });
                }
            }
        }


        private bool IsValueImportant(ExtendedVertex vertex, LogicValue value)
        {
            var ruleFactVertex = vertex as RuleFactVertex;
            if (ruleFactVertex != null)
                return IsValueImportant(ruleFactVertex, value);
            var quantifierVertex = vertex as QuantifierVertex;
            if (quantifierVertex != null)
                return IsValueImportant(quantifierVertex, value);
            return true;
        }

        /// <summary>
        /// Returns true if value is important to the vertex's value
        /// </summary>
        /// <returns></returns>
        private bool IsValueImportant(QuantifierVertex vertex, LogicValue value)
        {
            var isImportant = true;
            if (vertex != null)
            {
                isImportant = vertex.LogicValue == value;
            }
            return isImportant;
        }
        /// <summary>
        /// Returns true if value is important to the vertex's value
        /// </summary>
        /// <returns></returns>
        private bool IsValueImportant(RuleFactVertex vertex, LogicValue value)
        {
            var isImportant = true;
            if (vertex != null)
            {
                isImportant =
                    (vertex.Value == LogicValue.INCONSISTENT && value == LogicValue.INCONSISTENT) ||
                    ((vertex.Value == LogicValue.FALSE || vertex.Value == LogicValue.TRUE) && value == LogicValue.TRUE) ||
                    (vertex.Value == LogicValue.UNKNOWN && (value == LogicValue.FALSE || value == LogicValue.UNKNOWN));
            }
            return isImportant;
        }

        /// <summary>
        /// Create a new tree - a subtree with given root
        /// </summary>
        /// <param name="root"></param>
        /// <param name="mainGraph"></param>
        /// <returns></returns>
        private ExtendedGraph GetSubtree(ExtendedVertex root, ExtendedGraph mainGraph)
        {
            var backupGraph = new ExtendedGraph();
            var stack = new Stack<ExtendedVertex>();
            stack.Push(root);
            while (stack.Count > 0)
            {
                var vertex = stack.Pop();
                backupGraph.AddVertex(vertex);
                foreach (var outEdge in mainGraph.OutEdges(vertex))
                {
                    stack.Push(outEdge.Target);
                }
            }
            foreach (var vertex in backupGraph.Vertices)
            {
                foreach (var edge in mainGraph.OutEdges(vertex))
                {
                    backupGraph.AddEdge(edge);
                }
            }
            return backupGraph;
        }

        /// <summary>
        /// Remove vertices and their edges from the graph
        /// </summary>
        /// <param name="vertices"></param>
        /// <param name="graph"></param>
        /// <param name="vertexToLeave"></param>
        private void RemoveFromGraph(IEnumerable<ExtendedVertex> vertices, ExtendedGraph graph, ExtendedVertex vertexToLeave)
        {
            foreach (var vertex in vertices)
            {
                graph.ClearOutEdges(vertex);
                if (vertex != vertexToLeave)
                {
                    graph.RemoveVertex(vertex);
                }
            }
        }
        private void SelectRow(Row selection)
        {
            // FIXME: VERTEX IDENTIFIERS
            if (!GraphsForRows.ContainsKey(selection))
            {
                return;
            }

            if (selection.IsDummyRow)
                return;

            var rowSubgraph = GraphsForRows[selection];
            if (rowSubgraph.Graph == null)
            {
                BuildSubgraphForRow(selection, rowSubgraph);
            }

            if (ExplanationGraphView.BaseGraph == rowSubgraph.Graph)
            {
                return;
            }

            var graph = rowSubgraph.Graph;
            if (rowSubgraph.ResultDesc.Root != null)
            {
                graph = ExplanationGraphView.BaseGraph;
                // Remove the subgrab that is currently in the root's subtree, but remember its state to re-create if necessary.
                 var backupGraph = GetSubtree(rowSubgraph.ResultDesc.Root, graph);

                //save the backup graph
                foreach (var edge in graph.OutEdges(rowSubgraph.ResultDesc.Root))
                {
                    if (GraphsForVertices.ContainsKey(edge.Target))
                    {
                        GraphsForVertices[edge.Target].Graph = backupGraph;
                    }
                }
                // Remove the subgraph.
                RemoveFromGraph(backupGraph.Vertices, graph, rowSubgraph.ResultDesc.Root);

                // Add the new subgraph.
                foreach (var vertex in rowSubgraph.Graph.Vertices)
                {
                    if (vertex != rowSubgraph.ResultDesc.Root)
                    {
                        graph.AddVertex(vertex);
                    }
                }
                foreach (var edge in rowSubgraph.Graph.Edges)
                {
                    graph.AddEdge(edge);
                }
            }

            // Recolor the graph after switch.
            var vertexToSelect = rowSubgraph.ResultDesc.Root;
            if (vertexToSelect != null && rowSubgraph.Graph.ContainsVertex(vertexToSelect))
            {
               vertexToSelect = (from e in rowSubgraph.Graph.OutEdges(rowSubgraph.ResultDesc.Root)
                    select e.Target).FirstOrDefault();
            }
            if (vertexToSelect == null) //if nothing selected then select root of the whole tree
            {
                foreach (var vertex in graph.Vertices)
                {
                    if (graph.InDegree(vertex) == 0)
                    {
                        vertexToSelect = vertex;
                        break;
                    }
                }
            }

            ExplanationGraphView.IsGraphInitialized = false;
            ExplanationGraphView.BaseGraph = graph;
            ExplanationGraphView.RebuildGraph();

            if (vertexToSelect == null)
            {
                return;
            }

            if (!ExplanationGraphView.GraphArea.VertexList.ContainsKey(vertexToSelect))
            {
                return;
            }

            GraphArea_VertexSelected(null, new GraphX.Models.VertexSelectedEventArgs(ExplanationGraphView.GraphArea.VertexList[vertexToSelect], null, System.Windows.Input.ModifierKeys.None));
        }
    }

    /// <summary>
    /// Holds a graph for a row of the result
    /// </summary>
    class RowSubgraph
    {
        public ResultDesc ResultDesc { get; set; }
        public Row Row { get; set; }
        public ExtendedGraph Graph { get; set; }
    }

    /// <summary>
    /// 
    /// </summary>
    class ResultDesc
    {
        public string ModuleName { get; set; }
        public Expression Expression { get; set; }
        public ExtendedVertex Root { get; set; }
        public Result Result { get; set; }
    }

   
}
