﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.Xml.Linq;
using FourqlModeler.Commands.Console;
using FourqlModeler.Views.GraphExtensions;
using FourqlModeler.Views.GraphExtensions.ExprGraph;
using FourqlModeler.Views.GraphExtensions.WSMGraph;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using GraphX;
using Microsoft.Win32;
using Parago.Windows;

namespace FourqlModeler.Views
{
    /// <summary>
    /// Interaction logic for WsmView.xaml
    /// </summary>
    public partial class WsmView : UserControl
    {
        private WsmData WsmData { get; set; }

        public WsmView(WsmData wsmData, ExtendedGraph graph)
        {
            InitializeComponent();
            WsmData = wsmData;

            WsmGraphView.BaseGraph = graph;
            WsmGraphView.GraphArea.PostContextMenuFill += CreateLiteralExplanationContextMenu;
            Loaded += WsmView_Loaded;
        }

        void CreateLiteralExplanationContextMenu(VertexControl vertexControl)
        {
            var literalVertex = vertexControl.Vertex as LiteralVertex;
            if (literalVertex != null)
            {
                var explainItem = new MenuItem() { Header = "Explain the fact...", DataContext = vertexControl };
                explainItem.Click += ExplainLiteral;

                if (vertexControl.ContextMenu == null)
                {
                    vertexControl.ContextMenu = new ContextMenu();
                }
                else if (vertexControl.ContextMenu.Items.Count > 0)
                {
                    vertexControl.ContextMenu.Items.Insert(0, new Separator());
                }
                vertexControl.ContextMenu.Items.Insert(0, explainItem);
            }
        }

        void WsmView_Loaded(object sender, RoutedEventArgs e)
        {
            if (!WsmGraphView.IsGraphInitialized)
            {
                WsmGraphView.RebuildGraph();

                var item = new MenuItem() { Header = "Export WSM as XML..." };
                item.Click += ExportToXml;

                WsmGraphView.ZoomCtrl.ContextMenu.Items.Insert(1, item);
            }
        }

        private void ExplainLiteral(object sender, RoutedEventArgs e)
        {
            var menuItem = sender as MenuItem;
            var vertexControl = menuItem.DataContext as VertexControl;
            var vertex = vertexControl.Vertex as LiteralVertex;

            ExecuteExpressionCommand.ExecuteExpr(new ExprPrinter("", false).ConvertExpr(new ModuleExpression() { ModuleName = vertex.ModuleName, Next = vertex.Literal }).ToString());
        }

        private void ExportToXml(object sender, RoutedEventArgs e)
        {
            string errorMessage = null;
            var saveProgressResult = ProgressDialog.Execute(App.Current.MainWindow, "Saving the graph to an xml file...", () =>
            {
                ProgressDialog.Current.Report("Creating an xml for the graph...");

                var saveDialog = new SaveFileDialog() { Title = "Select a file to write the xml to.", DefaultExt = ".xml", Filter = "XML Files (.xml)|*.xml", AddExtension = true, OverwritePrompt = true, CheckPathExists = true };
                var saveResult = saveDialog.ShowDialog();
                if (!saveResult.HasValue || saveResult == false)
                {
                    return;
                }

                ProgressDialog.Current.Report("Preparing content to save to the xml file...");
                var rules = new Dictionary<Rule, int>();
                var id = 0;
                foreach (var rule in WsmData.Rules)
                {
                    rules.Add(rule, id++);
                }

                // Root.
                var xmlDocument = new XDocument(
                    new XElement("Wsm", new XAttribute("module", WsmData.ModuleName),
                        new XElement("Domains",
                            WsmData.PrimitiveDomainValues.Select(dom =>
                                new XElement("Domain",
                                    new XAttribute("name", DomainTypeUtils.getName(dom.Key)),
                                    new XAttribute("isPrimitive", true),
                                    new XElement("Values", dom.Value.Select(val => new XElement("Value", val)))
                                )
                            ),
                            WsmData.DefinedDomainValues.Select(dom =>
                                new XElement("Domain",
                                    new XAttribute("name", dom.Key.Name),
                                    new XAttribute("type", DomainTypeUtils.getName(dom.Key.Type)),
                                    new XAttribute("isPrimitive", false),
                                    new XElement("Values", dom.Value.Select(val => new XElement("Value", val)))
                                )
                            )
                        ),
                        new XElement("Relations",
                            WsmData.RelationFacts.Keys.Select(rel =>
                                new XElement("Relation",
                                    new XAttribute("name", rel.Name),
                                    new XElement("Params",
                                        Enumerable.Range(0, rel.paramCount() - 1).Select(i =>
                                            new XElement("Param",
                                                new XAttribute("domain", rel.PrimitiveDomains.ContainsKey(i) ? DomainTypeUtils.getName(rel.PrimitiveDomains[i]) : rel.DefinedDomains[i]))
                                        )
                                    ),
                                    new XElement("Rules",
                                        rules.Where(ruleEntry => ruleEntry.Key.Head.RelationName.Equals(rel.Name)).Select(ruleEntry =>
                                            new XElement("Rule",
                                                new XAttribute("id", ruleEntry.Value),
                                                new ProgramPrinter(false).append(ruleEntry.Key, "", false).Done()
                                            )
                                        )
                                    )
                                )
                            )
                        ),
                        new XElement("Facts",
                            WsmData.RelationFacts.Select(relFactsEntry =>
                                new XElement("RelationFacts",
                                    new XAttribute("relation", relFactsEntry.Key.Name),
                                    relFactsEntry.Value.Select(literalEntry =>
                                        new XElement("RelationFact",
                                            new XElement("Literal",
                                                new XElement("RawLiteral", new ExprPrinter("", false).visitLiteral(literalEntry.Key).Append(".").ToString()),
                                                new XElement("ParamValues",
                                                    literalEntry.Key.Values.Select(param =>
                                                        new XElement("ParamValue", ValuePrinter.ToValueString(param))
                                                    )
                                                )
                                            ),
                                            literalEntry.Value.Facts.Select(fact =>
                                                new XElement("Fact",
                                                    new XAttribute("value", fact.Value)
                                                )
                                            ),
                                            literalEntry.Value.RuleFacts.Select(ruleFact =>
                                                new XElement("RuleFact",
                                                    new XAttribute("value", ruleFact.Value.Value),
                                                    new XAttribute("ruleId", rules[ruleFact.Key])
                                                )
                                            )
                                        )
                                    )
                                )
                            )
                        )
                    )
                );

                ProgressDialog.Current.Report("Saving the graph to the xml file...");
                xmlDocument.Save(saveDialog.FileName);
            });

            if (saveProgressResult.OperationFailed || errorMessage != null)
            {
                MessageBox.Show("The save operation failed due to: " + (errorMessage != null ? errorMessage : saveProgressResult.Error.Message), "Error while saving the graph to xml.", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }
    }
}
