﻿using Antlr4.Runtime;
using Antlr4.Runtime.Tree;
using FourqlParser.Parsers.ExpParser.Converters;
using FourqlParser.Parsers.ModParser.Converters;
using FourqlParser.Structures.Print;
using FourqlParser.StaticAnalysis;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Parsers.ModParser;
using FourqlParser.Parsers.ExpParser;
using FourqlParser.Parsers;
using FourqlParser.Structures;

namespace FourqlParser
{
    class Application
    {
        private static Program Program;

        private static void Main(string[] args)
        {
            (new Application()).Run(args);
        }

        public void Run(string[] args)
        {
            try
            {
                Console.WriteLine("START");
                RunParser(args);
                //RunParser2(args);
                Console.Write("DONE. Hit RETURN to exit: ");
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: " + ex);
                Console.Write("Hit RETURN to exit: ");
            }
            Console.ReadLine();
        }

        private void RunParser(string[] args)
        {
            var errors = new FourqlParser.Parsers.ParserDispatcher.Errors();
            var hidden = new List<HiddenData>();
            var program = ParserDispatcher.parseProgram(args.Length > 0 ? ParserDispatcher.AsFile(args[0]) : "/* \ntest */ //\nmodule mod: domains: literal l. relations: rel(literal, l). facts: rel(a,b). end.", errors, hidden);
            if (errors.ErrorCount > 0)
            {
                throw new Exception(String.Format("ERRORS: {0}", errors.ErrorCount));
            }

            /*System.Console.WriteLine(program.ToStringTree());            
            System.Console.WriteLine("TREE:");
            ParseTreeWalker.Default.Walk(new FourqlTreePrinter(), program);*/

            var p = ParserDispatcher.toProgram(program);
            System.Console.WriteLine("STRUCT:\n{0}", new ProgramPrinter().append(p).Done());

            ParserDispatcher.checkProgram(p);

            Program = p;
        }

        private void RunParser2(string[] args)
        {
            var errors = new FourqlParser.Parsers.ParserDispatcher.Errors();
            var expr = ParserDispatcher.parseExpr(args.Length > 1 ? ParserDispatcher.AsFile(args[1]) : "mod.rel(a, b) or not mod.rel(b, a) and mod.rel(c, d) and (not mod.rel(d, e) or not mod.rel(f, g) in {true})." /*"mod.rel(1) or not rel(true) and mod.rel(1.0) and (not rel(false) or not main.rel(incons) in {true})."*/, errors);
            if (errors.ErrorCount > 0)
            {
                throw new Exception(String.Format("ERRORS: {0}", errors.ErrorCount));
            }

            /*System.Console.WriteLine(expr.ToStringTree());            
            System.Console.WriteLine("TREE:");
            ParseTreeWalker.Default.Walk(new ExprTreePrinter(), expr);*/

            var e = ParserDispatcher.toExpr(expr);
            System.Console.WriteLine("STRUCT:\n{0}", e.Accept(new ExprPrinter("", false)).ToString());

            ParserDispatcher.checkExpr(e, Program.Modules, null);
        }

        private String getFileContent(string file)
        {
            StringBuilder sb = new StringBuilder();
            using (StreamReader sr = new StreamReader(file))
            {
                String line;
                while ((line = sr.ReadLine()) != null)
                {
                    sb.AppendLine(line);
                }
            }
            return sb.ToString();
        }
    }

    public class FourqlTreePrinter : ModuleParserBaseListener
    {
        private String Indent = " ";

        #region Module
        public override void EnterModule(ModuleParser.ModuleContext context)
        {
            printIndent();
            System.Console.WriteLine("MODULE_NAME={0}", context.ModuleName.Text);
        }
        #endregion

        #region Sections
        public override void EnterDomain(ModuleParser.DomainContext context)
        {
            printIndent();
            System.Console.WriteLine("NAME={0}", context.DomainName.Text);
        }

        public override void EnterRelation(ModuleParser.RelationContext context)
        {
            printIndent();
            System.Console.WriteLine("NAME={0}", context.RelationName.Text);
        }

        public override void EnterRelationDomainArg(ModuleParser.RelationDomainArgContext context)
        {
            printIndent();
            System.Console.WriteLine("NAME={0}", context.DomainName.Text);
        }

        public override void EnterFact(ModuleParser.FactContext context)
        {
            printIndent();
            System.Console.WriteLine("IS_NEGATIVE={0} IS_NULL={1}", context.IsNegative, context.IsNegative == null);
        }
        #endregion

        #region Value
        public override void EnterValueLiteral(ModuleParser.ValueLiteralContext context)
        {
            printIndent();
            System.Console.WriteLine("LITERAL={0}", context.LiteralValue.Text);
        }

        public override void EnterValueString(ModuleParser.ValueStringContext context)
        {
            printIndent();
            System.Console.WriteLine("STRING=\"{0}\"", context.StringValue.Text);
        }

        public override void EnterValueInteger(ModuleParser.ValueIntegerContext context)
        {
            printIndent();
            System.Console.WriteLine("INTEGER={0}", context.IntegerValue.Text);
        }

        public override void EnterValueReal(ModuleParser.ValueRealContext context)
        {
            printIndent();
            System.Console.WriteLine("REAL={0}", context.RealValue.Text);
        }

        public override void EnterValueDate(ModuleParser.ValueDateContext context)
        {
            printIndent();
            System.Console.WriteLine("DATE={0}", context.DateValue.Text);
        }

        public override void EnterValueDateTime(ModuleParser.ValueDateTimeContext context)
        {
            printIndent();
            System.Console.WriteLine("DATE_TIME={0}", context.DateTimeValue.Text);
        }
        #endregion

        #region Every Rule
        public override void EnterEveryRule(ParserRuleContext context)
        {
            moveDown();
            printIndent();
            System.Console.WriteLine("Entered={0}", context.GetType());
        }

        public override void ExitEveryRule(ParserRuleContext context)
        {
            moveUp();
        }
        #endregion

        #region Indent Functions
        private void printIndent()
        {
            System.Console.Write(Indent);
        }

        private void moveDown()
        {
            Indent = Indent + " ";
        }

        private void moveUp()
        {
            if (Indent.Length > 0)
            {
                Indent = Indent.Substring(1);
            }
        }
        #endregion
    }

    public class ExprTreePrinter : ExprParserBaseListener
    {
        private String Indent = " ";

        #region Value
        public override void EnterValueLiteral(ExprParser.ValueLiteralContext context)
        {
            printIndent();
            System.Console.WriteLine("LITERAL={0}", context.LiteralValue.Text);
        }

        public override void EnterValueString(ExprParser.ValueStringContext context)
        {
            printIndent();
            System.Console.WriteLine("STRING=\"{0}\"", context.StringValue.Text);
        }

        public override void EnterValueInteger(ExprParser.ValueIntegerContext context)
        {
            printIndent();
            System.Console.WriteLine("INTEGER={0}", context.IntegerValue.Text);
        }

        public override void EnterValueReal(ExprParser.ValueRealContext context)
        {
            printIndent();
            System.Console.WriteLine("REAL={0}", context.RealValue.Text);
        }

        public override void EnterValueDate(ExprParser.ValueDateContext context)
        {
            printIndent();
            System.Console.WriteLine("DATE={0}", context.DateValue.Text);
        }

        public override void EnterValueDateTime(ExprParser.ValueDateTimeContext context)
        {
            printIndent();
            System.Console.WriteLine("DATE_TIME={0}", context.DateTimeValue.Text);
        }
        #endregion

        #region Every Rule
        public override void EnterEveryRule(ParserRuleContext context)
        {
            moveDown();
            printIndent();
            System.Console.WriteLine("Entered={0}", context.GetType());
        }

        public override void ExitEveryRule(ParserRuleContext context)
        {
            moveUp();
        }
        #endregion

        #region Indent Functions
        private void printIndent()
        {
            System.Console.Write(Indent);
        }

        private void moveDown()
        {
            Indent = Indent + " ";
        }

        private void moveUp()
        {
            if (Indent.Length > 0)
            {
                Indent = Indent.Substring(1);
            }
        }
        #endregion
    }
}
