﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    class ExpressionAnalysis : Expression.Visitor<Object>
    {
        private Module StartingModule = null;
        private Module CurrentModule = null;
        private IDictionary<string, Module> Modules = null;
        private ICollection<InvalidDefinition> Errors = null;
        private IDictionary<string, ICollection<VariableRef>> Variables = null;
        private ICollection<string> DuplicatedModules = null;
        
        public ExpressionAnalysis(Module startingModule, IDictionary<string, Module> modules) : this(startingModule, modules, new List<string>())
        {
        }

        public ExpressionAnalysis(Module startingModule, IDictionary<string, Module> modules, ICollection<string> duplicatedModules)
        {
            this.StartingModule = startingModule;
            this.Modules = modules;
            this.DuplicatedModules = duplicatedModules;
        }

        public IDictionary<string, ICollection<VariableRef>> run(Expression expr, ICollection<InvalidDefinition> errors = null, bool checkRelatedVariableTypes = true)
        {
            Errors = errors;
            CurrentModule = StartingModule;
            Variables = new Dictionary<string, ICollection<VariableRef>>();

            try
            {
                expr.Accept(this);

                if (checkRelatedVariableTypes)
                {
                    foreach (var variableEntry in Variables)
                    {
                        var variableIter = variableEntry.Value.GetEnumerator();
                        if (!variableIter.MoveNext())
                        {
                            continue;
                        }

                        var firstVariable = variableIter.Current;
                        var hasLocalModuleDomainRef = !firstVariable.Module.IsInternalExternal;
                        while (variableIter.MoveNext())
                        {
                            var nextVariable = variableIter.Current;
                            if (firstVariable.DomainType != nextVariable.DomainType)
                            {
                                AnalysisUtils.pushError(expr, String.Format("Variable '{0}' is defined by two or more primitive types - {1} in {2} (param no. {3}) from module '{4}' and {5} in {6} (param no. {7}) from module '{8}'", firstVariable.VariableName, DomainTypeUtils.getName(firstVariable.DomainType), firstVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), firstVariable.ParamNumber, firstVariable.Module.Name, DomainTypeUtils.getName(nextVariable.DomainType), nextVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), nextVariable.ParamNumber, nextVariable.Module.Name), errors);
                            }
                            hasLocalModuleDomainRef = hasLocalModuleDomainRef || !nextVariable.Module.IsInternalExternal;
                        }

                        if (!hasLocalModuleDomainRef)
                        {
                            AnalysisUtils.pushError(expr, String.Format("Variable '{0}' is defined by proper types, but the variable's context refers to internal/external modules only that do not hold domain values and thus no value will be generated for them (sample: {1} in {2} (param no. {3}) from module '{4}')", firstVariable.VariableName, DomainTypeUtils.getName(firstVariable.DomainType), firstVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), firstVariable.ParamNumber, firstVariable.Module.Name), errors);
                        }
                    }
                }

                return Variables;
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Error while checking the expression: {0}", expr.Accept(new ExprPrinter("", false))).ToString(), ex);
            }
        }

        private void pushVariables(Literal literal)
        {
            var variables = ExprVariableGatherer.findVariablesOfLiteral(CurrentModule, literal);
            foreach (var variableRefs in variables)
            {
                if (Variables.ContainsKey(variableRefs.Key))
                {
                    var savedVariables = Variables[variableRefs.Key];
                    foreach (var variable in variableRefs.Value)
                    {
                        savedVariables.Add(variable);
                    }
                }
                else
                {
                    Variables.Add(variableRefs.Key, variableRefs.Value);
                }
            }
        }
        
        #region Visitor<object> Members

        public object visitLiteral(Literal literal)
        {
            if (CurrentModule == null)
            {
                AnalysisUtils.pushError(literal, String.Format("Module in which the expression '{0}' would be executed is non-existent or ambiguous.", literal.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else
            {
                int previousErrors = Errors != null ? Errors.Count : 0;
                LiteralAnalysis.checkLiteral(CurrentModule, literal, Errors, true);
                if (Errors == null || Errors.Count == previousErrors)
                {
                    pushVariables(literal);
                }
            }

            return null;
        }

        public object visitNeg(NegExpression negExpr)
        {
            return negExpr.Next.Accept(this);
        }

        public object visitModule(ModuleExpression moduleExpr)
        {
            if (DuplicatedModules.Contains(moduleExpr.ModuleName))
            {
                AnalysisUtils.pushError(moduleExpr, String.Format("Module '{0}' used in expression '{1}' is ambiguous.", moduleExpr.ModuleName, moduleExpr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else if (!Modules.ContainsKey(moduleExpr.ModuleName))
            {
                AnalysisUtils.pushError(moduleExpr, String.Format("Module '{0}' used in expression '{1}' is non-existent.", moduleExpr.ModuleName, moduleExpr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else if (StartingModule != null && moduleExpr.ModuleName.Equals(StartingModule.Name))
            {
                AnalysisUtils.pushError(moduleExpr, String.Format("Module '{0}' used in expression '{1}' is also a rule module and thus it cannot be used in the rule.", moduleExpr.ModuleName, moduleExpr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }

            var module = CurrentModule;
            CurrentModule = Modules.ContainsKey(moduleExpr.ModuleName) ? Modules[moduleExpr.ModuleName] : null;

            moduleExpr.Next.Accept(this);

            CurrentModule = module;
            return null;
        }

        public object visitAnd(AndExpression andExpr)
        {
            foreach (var next in andExpr.Nexts)
            {
                next.Accept(this);
            }
            return null;
        }

        public object visitOr(OrExpression orExpr)
        {
            foreach (var next in orExpr.Nexts)
            {
                next.Accept(this);
            }
            return null;
        }

        #endregion
    }
}
