﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    class ProgramAnalysis
    {
        public void run(ICollection<Module> modules)
        {
            var moduleMap = new Dictionary<string, Module>();
            foreach (var module in modules)
            {
                if (moduleMap.ContainsKey(module.Name))
                {
                    throw new Exception(String.Format("Duplicated module '{0}'.", module.Name));
                }
                checkForInvalidDefinitions(module);
                moduleMap[module.Name] = module;
            }

            checkExpressions(moduleMap);
            checkModuleCycles(moduleMap);
        }
        
        #region Invalid Domain, Relation and Fact checks

        public void checkForInvalidDefinitions(Module module)
        {
            checkForInvalidDomains(module);
            checkForInvalidRelations(module);
            checkForInvalidFacts(module);
        }

        public void checkForInvalidDomains(Module module)
        {
            var foundDomains = new HashSet<string>();
            foreach (var domain in module.Domains)
            {
                // Check domains - if their names are unique within their module.
                if (foundDomains.Contains(domain.Name))
                {
                    throw new Exception(String.Format("Duplicated definition of domain '{0}' in module '{1}'.", domain.Name, module.Name));
                }

                foundDomains.Add(domain.Name);
            }
        }

        public void checkForInvalidRelations(Module module)
        {
            var domainNames = new HashSet<string>();
            foreach (var domain in module.Domains)
            {
                domainNames.Add(domain.Name);
            }

            var foundRelations = new HashSet<string>();
            foreach (var relation in module.Relations)
            {
                // Check relations - if their names are unique within their module.
                if (foundRelations.Contains(relation.Name))
                {
                    throw new Exception(String.Format("Duplicated definition of relation '{0}' in module '{1}'.", relation.Name, module.Name));
                }

                foreach (var domainName in relation.DefinedDomains.Values)
                {
                    // Check relations - if their parameters are either primitive types or defined domains.
                    if (!domainNames.Contains(domainName))
                    {
                        throw new Exception(String.Format("Invalid definition of relation '{0}' in module '{1}' - domain '{2}' does not exist in the module.", relation.Name, module.Name, domainName));
                    }
                }

                foundRelations.Add(relation.Name);
            }
        }

        public void checkForInvalidFacts(Module module)
        {
            foreach (var fact in module.Facts)
            {
                try
                {
                    LiteralAnalysis.checkLiteral(module, fact, null);
                }
                catch (Exception ex)
                {
                    throw new Exception(String.Format("Exception while checking fact '{0}' from '{1}'.", new ProgramPrinter().appendFactNoEnd(fact).Done(), module.Name), ex);
                }
            }
        }

        #endregion

        #region Invalid Expressions check

        private void checkExpressions(IDictionary<string, Module> moduleMap)
        {
            foreach (var module in moduleMap.Values)
            {
                foreach (var rule in module.Rules)
                {
                    new RuleAnalysis(module, moduleMap).run(rule);
                }
            }
        }

        #endregion

        #region Invalid Module Hierarchy check

        private void checkModuleCycles(IDictionary<string, Module> moduleMap)
        {
            var stronglyConnectedComponents = new TarjansAlgorithm().run(moduleMap);
            var builder = new StringBuilder("Found cycles (Strongly Connected Components) that consist of modules:\n");
            var cycleNo = 0;

            foreach (var component in stronglyConnectedComponents)
            {
                if (component.Count > 1)
                {
                    builder.Append(String.Format("\t{0}. ", cycleNo++));
                    var first = true;

                    foreach (var item in component)
                    {
                        if (first)
                        {
                            first = false;
                        }
                        else
                        {
                            builder.Append(", ");
                        }

                        builder.Append(item.Name);
                    }
                }
            }

            if (cycleNo > 0)
            {
                throw new Exception(builder.ToString());
            }
        }

        #endregion
    }
}
