﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    class RuleAnalysis
    {
        private Module StartingModule = null;
        private Module CurrentModule = null;
        private IDictionary<string, Module> Modules = null;
        private Rule Rule = null;
        private ICollection<InvalidDefinition> Errors = null;
        private IDictionary<string, ICollection<VariableRef>> CurrentVariables = null;
        private ICollection<IDictionary<string, ICollection<VariableRef>>> Variables = new List<IDictionary<string, ICollection<VariableRef>>>();
        private ICollection<string> DuplicatedModules = null;
        
        public RuleAnalysis(Module startingModule, IDictionary<string, Module> modules) : this(startingModule, modules, new List<string>())
        {
        }

        public RuleAnalysis(Module startingModule, IDictionary<string, Module> modules, ICollection<string> duplicatedModules)
        {
            this.StartingModule = startingModule;
            this.Modules = modules;
            this.DuplicatedModules = duplicatedModules;
        }

        public void run(Rule rule, ICollection<InvalidDefinition> errors = null)
        {
            Errors = errors;
            Rule = rule;
            CurrentModule = StartingModule;

            try
            {
                // Remember the number of errors we have in the error list.
                int errorNo = errors != null ? errors.Count : 0;
                // Check the literal.
                LiteralAnalysis.checkLiteral(StartingModule, rule.Head, errors, true);
                // See if our error number changed.
                if (errors != null && errors.Count != errorNo)
                {
                    // It did - ignore checking the body.
                    return;
                }

                try
                {
                    // Remember the head variables.
                    var headVariables = ExprVariableGatherer.findVariablesOfLiteral(StartingModule, rule.Head);
                    // Get the alternatives.
                    var alternatives = ExprCalcUtils.SplitRuleByAlternatives(rule.Body);
                    // Check the alternatives one by one.
                    var exprAnalysis = new ExpressionAnalysis(StartingModule, Modules);
                    var alternativeLocalVariables = new Dictionary<string, IDictionary<Expression, VariableRef>>();
                    foreach (var alternative in alternatives)
                    {
                        // Analyze the alternative expression and grab its variables.
                        var alternativeVariables = exprAnalysis.run(alternative, errors, false);
                        // Check against the head.
                        checkRuleVariableExistence(headVariables, alternativeVariables, errors, alternative);
                        var localVariables = alternativeVariables.Where(entry => !headVariables.ContainsKey(entry.Key)).ToDictionary(entry => entry.Key, entry => entry.Value);
                        // Check the variables from the alternative.
                        foreach (var variableEntry in alternativeVariables)
                        {
                            try
                            {
                                var variableIter = variableEntry.Value.GetEnumerator();
                                if (!variableIter.MoveNext())
                                {
                                    continue;
                                }

                                var firstVariable = variableIter.Current;
                                var hasLocalModuleDomainRef = !firstVariable.Module.IsInternalExternal;
                                while (variableIter.MoveNext())
                                {
                                    checkVariable(firstVariable, variableIter.Current);
                                    hasLocalModuleDomainRef = hasLocalModuleDomainRef || !variableIter.Current.Module.IsInternalExternal;
                                }

                                if (!hasLocalModuleDomainRef)
                                {
                                    AnalysisUtils.pushError(alternative, String.Format("Variable '{0}' is defined by proper types, but the variable's context refers to internal/external modules only that do not hold domain values and thus no value will be generated for them (sample: {1} in {2} (param no. {3}) from module '{4}')", firstVariable.VariableName, DomainTypeUtils.getName(firstVariable.DomainType), firstVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), firstVariable.ParamNumber, firstVariable.Module.Name), errors);
                                }

                                if (!headVariables.ContainsKey(variableEntry.Key))
                                {
                                    if (!alternativeLocalVariables.ContainsKey(variableEntry.Key))
                                    {
                                        alternativeLocalVariables.Add(variableEntry.Key, new Dictionary<Expression, VariableRef>());
                                    }
                                    alternativeLocalVariables[variableEntry.Key][alternative] = firstVariable;
                                }
                            }
                            catch (Exception ex)
                            {
                                var errorMessage = String.Format("The variable {0} types vary in the alternative: {1} of the rule: {2}", variableEntry.Key, new ExprPrinter("", false).ConvertExpr(alternative).ToString(), new ProgramPrinter().append(rule, "", false).Done());
                                if (errors == null)
                                {
                                    throw new Exception(errorMessage, ex);
                                }

                                AnalysisUtils.pushError(alternative, errorMessage + "\n" + ex.Message, errors);
                            }
                        }
                    }
                    // Check local variables between alternatives.
                    foreach (var variableEntry in alternativeLocalVariables)
                    {
                        var alternativeIter = variableEntry.Value.GetEnumerator();
                        if (!alternativeIter.MoveNext())
                        {
                            continue;
                        }

                        var firstAlternative = alternativeIter.Current.Key;
                        var firstRef = alternativeIter.Current.Value;

                        while (alternativeIter.MoveNext())
                        {
                            var secondAlternative = alternativeIter.Current.Key;
                            var secondRef = alternativeIter.Current.Value;

                            try
                            {
                                checkVariable(firstRef, secondRef);
                            }
                            catch (Exception ex)
                            {
                                var errorMessage = String.Format("The variable {0} types vary between alternatives (1) {1} and (2) {2} of the rule: {3}", variableEntry.Key, new ExprPrinter("", false).ConvertExpr(firstAlternative).ToString(), new ExprPrinter("", false).ConvertExpr(secondAlternative).ToString(), new ProgramPrinter().append(rule, "", false).Done());
                                if (errors == null)
                                {
                                    throw new Exception(errorMessage, ex);
                                }

                                AnalysisUtils.pushError(rule, errorMessage + "\n" + ex.Message, errors);
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    throw new Exception(String.Format("Error while checking the body of expression: {0}", rule.Body.Accept(new ExprPrinter("", false))).ToString(), ex);
                }
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Error while checking rule: {0} of module '{1}'", new ProgramPrinter().append(rule, "", false).Done(), StartingModule.Name), ex);
            }
        }

        private void checkRuleVariableExistence(IDictionary<string, ICollection<VariableRef>> headVariables, IDictionary<string, ICollection<VariableRef>> exprVariables, ICollection<InvalidDefinition> errors = null, Expression alternative = null)
        {
            foreach (var variableList in headVariables.Values)
            {
                var variable = variableList.ElementAt(0);
                if (!exprVariables.ContainsKey(variable.VariableName))
                {
                    if (alternative == null)
                    {
                        AnalysisUtils.pushError(Rule.Body, String.Format("There is no variable '{0}' in the body of the rule: {1}", variable.VariableName, new ProgramPrinter().append(Rule, "", false).Done()), errors);
                    }
                    else
                    {
                        AnalysisUtils.pushError(alternative, String.Format("There is no variable '{0}' in the alternative: {1} of the body of the rule: {2}", variable.VariableName, alternative.Accept(new ExprPrinter("", false)).ToString(), new ProgramPrinter().append(Rule, "", false).Done()), errors);
                    }
                }
                else
                {
                    try
                    {
                        checkVariable(variable, exprVariables[variable.VariableName].ElementAt(0));
                    }
                    catch (Exception ex)
                    {
                        if (errors == null)
                        {
                            if (alternative == null)
                            {
                                throw new Exception(String.Format("Definitions of variable '{0}' in the head and body of the rule: {1} differ", variable.VariableName, new ProgramPrinter().append(Rule, "", false).Done()), ex);
                            }
                            else
                            {
                                throw new Exception(String.Format("Definitions of variable '{0}' in the head and alternative: {1} of the rule: {2} differ", variable.VariableName, alternative.Accept(new ExprPrinter("", false)).ToString(), new ProgramPrinter().append(Rule, "", false).Done()), ex);
                            }
                        }

                        if (alternative == null)
                        {
                            AnalysisUtils.pushError(Rule, String.Format("Definitions of variable '{0}' in the head and body of the rule: {1} differ: {2}", variable.VariableName, new ProgramPrinter().append(Rule, "", false).Done(), ex.Message), errors);
                        }
                        else
                        {
                            AnalysisUtils.pushError(Rule, String.Format("Definitions of variable '{0}' in the head and alternative: {1} of the rule: {2} differ: {3}", variable.VariableName, alternative.Accept(new ExprPrinter("", false)).ToString(), new ProgramPrinter().append(Rule, "", false).Done(), ex.Message), errors);
                        }
                    }
                }
            }
        }

        private void checkVariable(VariableRef first, VariableRef second)
        {
            if (first.DomainType != second.DomainType)
            {
                throw new Exception(String.Format("Variable '{0}' is defined by two or more primitive types - {1} in {2} (param no. {3}) from module '{4}' and {5} in {6} (param no. {7}) from module '{8}'", first.VariableName, DomainTypeUtils.getName(first.DomainType), first.Literal.Accept(new ExprPrinter("", false)).ToString(), first.ParamNumber, first.Module.Name, DomainTypeUtils.getName(second.DomainType), second.Literal.Accept(new ExprPrinter("", false)).ToString(), second.ParamNumber, second.Module.Name));
            }
        }
    }

    public class ExprCalcUtils
    {
        public static ICollection<Expression> SplitRuleByAlternatives(Expression body)
        {
            var subexpressions = new List<Expression>();
            SplitRuleByAlternatives(body, subexpressions);
            return subexpressions;
        }

        public static void SplitRuleByAlternatives(Expression body, ICollection<Expression> subexpressions)
        {
            if (body.Accept(new OrChecker()))
            {
                foreach (var subexpr in ((OrExpression)body).Nexts)
                {
                    SplitRuleByAlternatives(subexpr, subexpressions);
                }
            }
            else
            {
                subexpressions.Add(body);
            }
        }

        private class OrChecker : Expression.AbstractVisitor<bool>
        {
            public override bool visitOr(OrExpression orExpr)
            {
                return true;
            }

            protected override bool defaultValue()
            {
                return false;
            }
        }
    }
}
