﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.Structures.Print
{
    public class ProgramPrinter : Printer
    {
        public ProgramPrinter(Boolean printNewLinesInExpr = true) : base(printNewLinesInExpr) { }

        public ProgramPrinter append(Program program, String indent = "")
        {
            foreach (Module module in program.Modules)
            {
                append(module, indent);
            }
            return this;
        }

        public ProgramPrinter append(Module module, String indent = "")
        {
            Builder.AppendLine(String.Format("{0}module {1}:", indent, module.Name));

            appendDomains(module, indent);
            appendRelations(module, indent);
            appendRules(module, indent);
            appendFacts(module, indent);

            Builder.AppendLine(String.Format("{0}end.", indent));
            return this;
        }

        public ProgramPrinter appendDomains(Module module, String indent = "")
        {
            Builder.AppendLine(String.Format("{0}domains:", indent));
            append(module.Domains, indent + "  ");
            return this;
        }

        public ProgramPrinter append(ICollection<Domain> domains, String indent = "")
        {
            foreach (var domain in domains)
            {
                append(domain, indent);
            }
            return this;
        }

        public ProgramPrinter append(Domain domain, String indent = "", bool endWithNewLine = true, bool endWithDot = true)
        {
            AppendNewLineNoSpace(String.Format("{0}{1} {2}{3}", indent, DomainTypeUtils.getName(domain.Type), domain.Name, endWithDot ? "." : ""), endWithNewLine);
            return this;
        }

        public ProgramPrinter appendRelations(Module module, String indent = "")
        {
            Builder.AppendLine(String.Format("{0}relations:", indent));
            append(module.Relations, indent + "  ");
            return this;
        }

        public ProgramPrinter append(ICollection<Relation> relations, String indent = "")
        {
            foreach (var relation in relations)
            {
                append(relation, indent);
            }
            return this;
        }

        public ProgramPrinter append(Relation relation, String indent = "", bool endWithNewLine = true, bool endWithDot = true)
        {
            Builder.Append(String.Format("{0}{1}(", indent, relation.Name));

            for (int i = 0; i < relation.paramCount(); ++i)
            {
                if (i != 0)
                {
                    Builder.Append(", ");
                }

                if (relation.PrimitiveDomains.ContainsKey(i))
                {
                    Builder.Append(DomainTypeUtils.getName(relation.PrimitiveDomains[i]));
                }
                else
                {
                    Builder.Append(relation.DefinedDomains[i]);
                }
            }

            AppendNewLineNoSpace(String.Format("){0}", endWithDot ? "." : ""), endWithNewLine);
            return this;
        }

        public ProgramPrinter appendRules(Module module, String indent = "")
        {
            Builder.AppendLine(String.Format("{0}rules:", indent));
            append(module.Rules, indent + "  ");
            return this;
        }

        public ProgramPrinter append(ICollection<Rule> rules, String indent = "")
        {
            foreach (var rule in rules)
            {
                append(rule, indent);
            }
            return this;
        }

        public ProgramPrinter append(Rule rule, String indent = "", bool endWithNewLine = true, bool endWithDot = true)
        {
            var headString = new ProgramPrinter().appendFactNoEnd(rule.Head, indent).appendString(" :- ").Done();
            var bodyString = rule.Body.Accept(new ExprPrinter(indent + "    ", endWithNewLine)).ToString();
            Builder.Append(headString);
            Builder.Append(bodyString);
            AppendNewLineNoSpace(endWithDot ? "." : "", endWithNewLine);
            return this;
        }

        private String createIndent(int length)
        {
            var builder = new StringBuilder();
            for (int i = 0; i < length; ++i)
            {
                builder.Append(" ");
            }
            return builder.ToString();
        }

        public ProgramPrinter appendFacts(Module module, String indent = "")
        {
            Builder.AppendLine(String.Format("{0}facts:", indent));
            foreach (var fact in module.Facts)
            {
                append(fact, indent + "  ");
            }
            return this;
        }

        public ProgramPrinter append(Fact fact, String indent = "", bool endWithNewLine = true, bool endWithDot = true)
        {
            appendFactNoEnd(fact, indent);
            AppendNewLineNoSpace(endWithDot ? "." : "", endWithNewLine);
            return this;
        }

        public ProgramPrinter appendFactNoEnd(Fact fact, String indent = "")
        {
            Builder.Append(String.Format("{0}{1}", indent, fact.IsNegative ? "-" : ""));
            appendLiteralNoEnd(fact);
            return this;
        }

        public ProgramPrinter appendLiteralNoEnd(Literal fact, String indent = "")
        {
            Builder.Append(String.Format("{0}{1}(", indent, fact.RelationName));

            for (int i = 0; i < fact.Values.Count; ++i)
            {
                if (i != 0)
                {
                    Builder.Append(", ");
                }

                Builder.Append(ValuePrinter.ToValueString(fact.Values[i]));
            }

            Builder.Append(")");
            return this;
        }

        public ProgramPrinter appendString(String str)
        {
            Builder.Append(str);
            return this;
        }
        
        
        public String Done()
        {
            return Builder.ToString();
        }
    }
}
