﻿using FourqlParser.Structures;
using FourqlParser.Structures.Optimization;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.Parsers.Converters
{
    class ComplexExprConverter : ModuleParserBaseVisitor<Expression>
    {
        private Literal Literal;
        private ModuleExpression ModuleToNegate = null;
        private ModuleExpression Module;
        public String ModuleName { get; set; }

        public Expression convert(ModuleParser.ExpressionRuleContext context)
        {
            ModuleToNegate = null;
            return new ExpressionOptimizer().optimize(context.Accept(this));
        }

        #region IExprParserVisitor<Expression> Members

        public override Expression VisitExpr(ModuleParser.ExprContext context)
        {
            ModuleToNegate = null;
            return context.Expr.Accept(this);
        }

        public override Expression VisitOrExpr(ModuleParser.OrExprContext context)
        {
            var expr = new OrExpression() { ParserContext = context };
            expr.Nexts.Add(context.Head.Accept(this));
            if (context.Tail != null)
            {
                ModuleToNegate = null;
                expr.Nexts.Add(context.Tail.Accept(this));
            }
            return expr;
        }

        public override Expression VisitAndExpr(ModuleParser.AndExprContext context)
        {
            var expr = new AndExpression() { ParserContext = context };
            expr.Nexts.Add(context.Head.Accept(this));
            if (context.Tail != null)
            {
                ModuleToNegate = null;
                expr.Nexts.Add(context.Tail.Accept(this));
            }
            return expr;
        }

        public override Expression VisitNonNegExpr(ModuleParser.NonNegExprContext context)
        {
            return context.Expr.Accept(this);
        }

        public override Expression VisitNegExpr(ModuleParser.NegExprContext context)
        {
            var next = context.Expr.Accept(this);
            var nextAsModule = next as ModuleExpression;
            if (nextAsModule != null)
            {
                nextAsModule.IsNegative = true;
                return next;
            }
            return new NegExpression() { Next = next, ParserContext = context };
        }

        public override Expression VisitParentExpr(ModuleParser.ParentExprContext context)
        {
            return context.Expr.Accept(this);
        }

        public override Expression VisitModuleExpr(ModuleParser.ModuleExprContext context)
        {
            var moduleBackup = Module;
            ModuleName = context.ModuleName.Text;
            Module = new ModuleExpression() { ModuleName = context.ModuleName.Text, Next = context.Expr.Accept(this), ParserContext = context };
            if (context.InList != null)
            {
                context.InList.Accept(this);
            }
            var module = Module;
            Module = moduleBackup;
            ModuleName = moduleBackup == null ? null : moduleBackup.ModuleName;
            return module;
        }

        public override Expression VisitLiteralExpr(ModuleParser.LiteralExprContext context)
        {
            return context.Literal.Accept(this);
        }

        public override Expression VisitLiteral(ModuleParser.LiteralContext context)
        {
            Literal = new Literal()
            {
                RelationName =  context.RelationName != null? context.RelationName.Text : context.DomainRelationName.GetText().ToLower(), 
                ParserContext = context
            };
            context.LiteralArgs.Accept(this);
            return Literal;
        }

        public override Expression VisitLiteralVariableArgument(ModuleParser.LiteralVariableArgumentContext context)
        {
            Literal.Values.Add(new ValueHolder(context.VariableName.Text, DomainType.VARIABLE));
            return Literal;
        }

        public override Expression VisitLiteralArguments(ModuleParser.LiteralArgumentsContext context)
        {
            context.Head.Accept(this);
            if (context.Tail != null)
            {
                context.Tail.Accept(this);
            }
            return Literal;
        }

        public override Expression VisitInLogicArguments(ModuleParser.InLogicArgumentsContext context)
        {
            Module.InValues.Add(context.Head.Accept(new LogicValueConverter()));
            if (context.Tail != null)
            {
                context.Tail.Accept(this);
            }
            return Module;
        }

        public override Expression VisitSingleInLogicArgument(ModuleParser.SingleInLogicArgumentContext context)
        {
            Module.InValues.Add(context.Head.Accept(new LogicValueConverter()));
            return Module;
        }

        public override Expression VisitLiteralValueArgument(ModuleParser.LiteralValueArgumentContext context)
        {
            Literal.Values.Add(context.Value.Accept(new ValueConverter()));
            return Literal;
        }

        public override Expression VisitNoLiteralArguments(ModuleParser.NoLiteralArgumentsContext context)
        {
            return Literal;
        }

        public override Expression VisitAllExpr(ModuleParser.AllExprContext context)
        {
            var next = context.Expr.Accept(this);
            Expression expr;
            expr = new AllExpression()
            {
                VariableName = context.VariableName.Text,
                DomainModule = context.ModuleName == null ? (ModuleName ?? null) : context.ModuleName.Text,
                VariableDomain = (context.DomainName == null ? context.PrimitiveDomain.GetText().ToLower() : context.DomainName.Text),
                Next = next,
                ParserContext = context
            };
            return expr;
        }

        public override Expression VisitExistsExpr(ModuleParser.ExistsExprContext context)
        {
            var next = context.Expr.Accept(this);
            Expression expr;
            expr = new ExistsExpression() { 
                VariableName = context.VariableName.Text, 
                DomainModule = context.ModuleName == null ? (ModuleName ?? null) : context.ModuleName.Text,
                VariableDomain = (context.DomainName == null ? context.PrimitiveDomain.GetText().ToLower() : context.DomainName.Text), 
                Next = next, 
                ParserContext = context };
            return expr;
        }
        #endregion
    }
}
