﻿using Antlr4.Runtime;
using Antlr4.Runtime.Tree;
using FourqlParser.Parsers.Converters;
using FourqlParser.StaticAnalysis;
using FourqlParser.Structures;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.Parsers
{
    public class ParserDispatcher
    {
        public static ModuleParser.ExpressionRuleContext parseExpr(string toParse, Errors errors)
        {
            // Create an input stream.
            var input = new AntlrInputStream(toParse);

            // Create a lexer and set a new error listener for our error structure.
            var lexer = new FourqlLexer(input);
            lexer.AddErrorListener(new LexerErrorListener(errors));

            // Create a token stream for our lexer.
            var tokenStream = new CommonTokenStream(lexer);

            // Create a parser for the token stream and create a new error listener for it and our error structure.
            var parser = new ModuleParser(tokenStream);
            parser.AddErrorListener(new ErrorListener(errors));
            
            // Parse the expression.
            var expr = parser.queryRule();
            return expr.Query;
        }

        public static Expression toExpr(ModuleParser.ExpressionRuleContext expression)
        {
            return new ComplexExprConverter().convert(expression);
        }

        public static void checkExpr(Expression expression, ICollection<Module> modules, Module currentModule)
        {
            var moduleDict = new Dictionary<string, Module>();
            foreach (var module in modules)
            {
                moduleDict.Add(module.Name, module);
            }
            new ExpressionAnalysis(currentModule, moduleDict).run(expression);
        }

        public static void checkRule(Rule rule, ICollection<Module> modules, Module currentModule)
        {
            var moduleDict = new Dictionary<string, Module>();
            foreach (var module in modules)
            {
                moduleDict.Add(module.Name, module);
            }
            new RuleAnalysis(currentModule, moduleDict).run(rule);
        }

        public static ModuleParser.ProgramRuleContext parseProgram(string toParse, Errors errors, ICollection<HiddenData> hiddenDataCollection)
        {
            // Create an input stream.
            var input = new AntlrInputStream(toParse);

            // Create a lexer and set a new error listener for our error structure.
            var lexer = new FourqlLexer(input);
            lexer.AddErrorListener(new LexerErrorListener(errors));

            // Create a token stream for our lexer.
            var tokenStream = new CommonTokenStream(lexer);

            // Create a parser for the token stream and create a new error listener for it and our error structure.
            var parser = new ModuleParser(tokenStream);
            parser.AddErrorListener(new ErrorListener(errors));

            if (hiddenDataCollection != null)
            {
                // Gather hidden data (comments, white signs) if necessary.
                writeComments(toParse, hiddenDataCollection);
            }

            // Parse the program.
            var program = parser.programRule();
            return program;
        }

        private static void writeComments(string toParse, ICollection<HiddenData> hiddenDataCollection)
        {
            // Create an input stream.
            var input = new AntlrInputStream(toParse);

            // Create a lexer and set a new error listener for our error structure.
            var lexer = new FourqlLexer(input);

            // Create a token stream for our lexer.
            var tokenStream = new CommonTokenStream(lexer, TokenConstants.HiddenChannel);

            // Fill the token stream with tokens.
            tokenStream.Fill();

            if (tokenStream.Size == 0)
            {
                // We got no tokens - return.
                return;
            }

            // Enumerate over found tokens.
            foreach (var token in tokenStream.GetTokens(0, tokenStream.Size - 1))
            {
                if (token.Channel == TokenConstants.HiddenChannel)
                {
                    // Found data from the hidden channel - add.
                    hiddenDataCollection.Add(new HiddenData(token, token.Type == FourqlLexer.COMMENT));
                }
            }
        }

        public static Program toProgram(ModuleParser.ProgramRuleContext program)
        {
            return program.Accept(new ProgramConverter());
        }

        public static void checkProgram(Program program)
        {
            new ProgramAnalysis().run(program.Modules);
        }

        public static String AsFile(string file)
        {
            var stringBuilder = new StringBuilder();
            using (var reader = new StreamReader(file))
            {
                String line;
                while ((line = reader.ReadLine()) != null)
                {
                    stringBuilder.AppendLine(line);
                }
            }
            return stringBuilder.ToString();
        }

        public class Errors
        {
            public ICollection<Error<int>> LexerErrors { get; set; }
            public ICollection<Error<IToken>> ParserErrors { get; set; }
            public int ErrorCount { get { return LexerErrors.Count + ParserErrors.Count; } }

            public Errors()
            {
                 LexerErrors = new List<Error<int>>();
                 ParserErrors = new List<Error<IToken>>();
            }

            public override string ToString()
            {
                var builder = new StringBuilder("Expression parser errors: ");
                var errorNo = 0;
                builder.Append(ErrorCount);
                foreach (var lexerError in LexerErrors)
                {
                    builder.Append(String.Format("\n{0}. Lexer", errorNo++));
                    builder.Append(lexerError.ToString());
                }

                foreach (var parserError in ParserErrors)
                {
                    builder.Append(String.Format("\n{0}. Parser", errorNo++));
                    builder.Append(parserError.ToString());
                }

                return builder.ToString();
            }
        }

        public class Error<T>
        {
            public T OffendingSymbol { get; set; }
            public int Line { get; set; }
            public int Column { get; set; }
            public string Message { get; set; }
            public RecognitionException Exception { get; set; }

            public override string ToString()
            {
                return String.Format("Error=[\n  @Symbol={0}\n  @Line={1}\n  @Column={2}\n  @Message={3}\n  @Exception={4}\n]", OffendingSymbol, Line, Column, Message, Exception);
            }
        }

        private class ErrorListener : IAntlrErrorListener<IToken>
        {
            public Errors Errors { get; set; }

            public ErrorListener(Errors errors)
            {
                Errors = errors;
            }

            #region IAntlrErrorListener<int> Members

            public void SyntaxError(IRecognizer recognizer, IToken offendingSymbol, int line, int charPositionInLine, string msg, RecognitionException e)
            {
                if (Errors != null)
                {
                    Errors.ParserErrors.Add(new Error<IToken>() { OffendingSymbol = offendingSymbol, Line = line, Column = charPositionInLine, Message = msg, Exception = e });
                }
            }

            #endregion
        }

        private class LexerErrorListener : IAntlrErrorListener<int>
        {
            public Errors Errors { get; set; }

            public LexerErrorListener(Errors errors)
            {
                Errors = errors;
            }

            #region IAntlrErrorListener<int> Members

            public void SyntaxError(IRecognizer recognizer, int offendingSymbol, int line, int charPositionInLine, string msg, RecognitionException e)
            {
                if (Errors != null)
                {
                    Errors.LexerErrors.Add(new Error<int>() { OffendingSymbol = offendingSymbol, Line = line, Column = charPositionInLine, Message = msg, Exception = e });
                }
            }

            #endregion
        }
    }
}
