﻿using FourqlParser.Structures;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    public class ExprVariableGatherer
    {
        public static IDictionary<String, ICollection<VariableRef>> findVariablesOfLiteral(Module module, Literal literal)
        {
            var variables = new Dictionary<string, ICollection<VariableRef>>();
            return findVariablesOfLiteral(module, literal, variables);
        }

        public static IDictionary<String, ICollection<VariableRef>> findVariablesOfLiteral(Module module, Literal literal, IDictionary<String, ICollection<VariableRef>> variables)
        {
            var relation = module.FindRelation(literal.RelationName);
            var paramNo = 0;

            foreach (var arg in literal.Values)
            {
                if (arg.Type == DomainType.VARIABLE)
                {
                    var variableRef = new VariableRef() { Module = module, Literal = literal, ParamNumber = paramNo, VariableName = arg.valueAs<string>() };
                    if (relation.PrimitiveDomains.ContainsKey(paramNo))
                    {
                        variableRef.DomainType = relation.PrimitiveDomains[paramNo];
                    }
                    else
                    {
                        var domain = module.FindDomain(relation.DefinedDomains[paramNo]);
                        variableRef.DomainName = domain.Name;
                        variableRef.DomainType = domain.Type;
                    }

                    ICollection<VariableRef> variablesOf;
                    if (variables.ContainsKey(variableRef.VariableName))
                    {
                        variablesOf = variables[variableRef.VariableName];
                    }
                    else
                    {
                        variablesOf = new List<VariableRef>();
                        variables[variableRef.VariableName] = variablesOf;
                    }
                    variablesOf.Add(variableRef);
                }
                ++paramNo;
            }
            return variables;
        }

        public static IDictionary<String, ICollection<VariableRef>> findBoundVariables(Module module, Expression expr, IDictionary<string, Module> modules) 
        {
            return expr.Accept(new BoundVariableGatherer(module, modules));
        }
    }

    class BoundVariableGatherer : Expression.Visitor<IDictionary<string, ICollection<VariableRef>>>
    {
        IDictionary<string, ICollection<VariableRef>> variables = new Dictionary<string, ICollection<VariableRef>>();

        private Module Module;
        private IDictionary<string, Module> Modules;
        public BoundVariableGatherer(Module module, IDictionary<string, Module> modules)
        {
            this.Module = module;
            this.Modules = modules;
        }


        public IDictionary<string, ICollection<VariableRef>> visitLiteral(Literal literal)
        {
            return variables;
        }

        public IDictionary<string, ICollection<VariableRef>> visitNeg(NegExpression negExpr)
        {
            return negExpr.Next.Accept(this);
        }

        public IDictionary<string, ICollection<VariableRef>> visitModule(ModuleExpression moduleExpr)
        {
            var prevModule = Module;
            Module = Modules[moduleExpr.ModuleName];
            moduleExpr.Next.Accept(this);
            Module = prevModule;
            return variables;
        }

        public IDictionary<string, ICollection<VariableRef>> visitAnd(AndExpression andExpr)
        {
            foreach (var next in andExpr.Nexts)
            {
                next.Accept(this);
            }
            return variables;
        }

        public IDictionary<string, ICollection<VariableRef>> visitOr(OrExpression orExpr)
        {
            foreach (var next in orExpr.Nexts)
            {
                next.Accept(this);
            }
            return variables;
        }

        public IDictionary<string, ICollection<VariableRef>> visitExists(ExistsExpression exExpr)
        {
            if (!variables.ContainsKey(exExpr.VariableName))
            {
                variables.Add(exExpr.VariableName, new List<VariableRef>());

                DomainType t;
                exExpr.DomainModule = exExpr.DomainModule ?? Module.Name;
                Module mod = null;
                if (Modules.ContainsKey(exExpr.DomainModule))
                    mod = Modules[exExpr.DomainModule];
                if (DomainType.TryParse(exExpr.VariableDomain, true, out t))
                    variables[exExpr.VariableName].Add(new VariableRef()
                    {
                        DomainName = exExpr.VariableDomain,
                        Module = mod,
                        VariableName = exExpr.VariableName,
                        ParamNumber = 0,
                        DomainType = t,
                        Literal = null
                    });
                else
                {
                    Domain dom = mod == null ? null : mod.FindDomain(exExpr.VariableDomain);
                    variables[exExpr.VariableName].Add(new VariableRef()
                    {
                        DomainName = exExpr.VariableDomain,
                        Module = mod,
                        VariableName = exExpr.VariableName,
                        ParamNumber = 0,
                        DomainType = dom == null ? DomainType.VARIABLE : dom.Type,
                        Literal = null
                    });
                }
            }

            return exExpr.Next.Accept(this);
        }

        public IDictionary<string, ICollection<VariableRef>> visitAll(AllExpression allExpr)
        {
            if (!variables.ContainsKey(allExpr.VariableName))
            {
                variables.Add(allExpr.VariableName, new List<VariableRef>());
                DomainType t;
                allExpr.DomainModule = allExpr.DomainModule ?? Module.Name;
                Module mod = null;
                if (Modules.ContainsKey(allExpr.DomainModule))
                    mod = Modules[allExpr.DomainModule];
                if (DomainType.TryParse(allExpr.VariableDomain, true, out t))
                    variables[allExpr.VariableName].Add(new VariableRef()
                    {
                        DomainName = allExpr.VariableDomain,
                        Module = mod,
                        VariableName = allExpr.VariableName,
                        ParamNumber = 0,
                        DomainType = t,
                        Literal = null
                    });
                else
                {
                    Domain dom = mod == null ? null : mod.FindDomain(allExpr.VariableDomain);
                    variables[allExpr.VariableName].Add(new VariableRef()
                    {
                        DomainName = allExpr.VariableDomain,
                        Module = mod,
                        VariableName = allExpr.VariableName,
                        ParamNumber = 0,
                        DomainType = dom == null ? DomainType.VARIABLE : dom.Type,
                        Literal = null
                    });
            }
        }
            return allExpr.Next.Accept(this);
        }
    }

    public class VariableRef {
        public Module Module { get; set; }
        public Literal Literal { get; set; }
        public int ParamNumber { get; set; }
        public String VariableName { get; set; }
        public String DomainName { get; set; }
        public DomainType DomainType { get; set; }
        public Boolean isBound()
        {
            return (Literal == null) && (ParamNumber == 0);
        }
    }
}
