﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.CodeDom;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    class ExpressionAnalysis : Expression.Visitor<Object>
    {
        private Module StartingModule = null;
        private Module CurrentModule = null;
        private IDictionary<string, Module> Modules = null;
        private ICollection<InvalidDefinition> Errors = null;
        private IDictionary<string, ICollection<VariableRef>> Variables = null;
        private ICollection<string> DuplicatedModules = null;

        public ExpressionAnalysis(Module startingModule, IDictionary<string, Module> modules)
            : this(startingModule, modules, new List<string>())
        {
        }

        public ExpressionAnalysis(Module startingModule, IDictionary<string, Module> modules, ICollection<string> duplicatedModules)
        {
            this.StartingModule = startingModule;
            this.Modules = modules;
            this.DuplicatedModules = duplicatedModules;
        }

        /// <summary>
        /// analyze the expression, return variables appearing in it
        /// </summary>
        /// <param name="expr"></param>
        /// <param name="errors"></param>
        /// <param name="checkRelatedVariableTypes"></param>
        /// <returns></returns>
        public IDictionary<string, ICollection<VariableRef>> run(Expression expr, ICollection<InvalidDefinition> errors = null, bool checkRelatedVariableTypes = true)
        {
            Errors = errors;
            CurrentModule = StartingModule;
            Variables = new Dictionary<string, ICollection<VariableRef>>();

            try
            {
                //analyze the expression
                expr.Accept(this);

                //check the gathered variables
                if (checkRelatedVariableTypes)
                {
                    foreach (var variableEntry in Variables)
                    {
                        checkVariableTypes(variableEntry.Value, expr, errors);
                    }
                }

                return Variables;
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Error while checking the expression: {0}", expr.Accept(new ExprPrinter("", false))).ToString(), ex);
            }
        }


        private void checkVariableTypes(ICollection<VariableRef> variableEntry, Expression expr, ICollection<InvalidDefinition> errors = null)
        {
            
            // get an iterator for variableRefs with the same name
            var variableIter = variableEntry.GetEnumerator();
            if (!variableIter.MoveNext())
            {
                return;
            }

            var firstVariable = variableIter.Current;
            var hasLocalModuleDomainRef = !firstVariable.Module.IsInternalExternal;
            while (variableIter.MoveNext())
            {   // check if all variableRefs having the same name have the same primitive type
                var nextVariable = variableIter.Current;
                if (firstVariable.DomainType != nextVariable.DomainType)
                {
                    if (firstVariable.isBound()) // different message for bound variables
                        AnalysisUtils.pushError(expr, String.Format("Bound variable '{0}' is defined as type {1} but in '{2}' (param no. {3}) from module '{4}' it has type {5}", firstVariable.VariableName, DomainTypeUtils.getName(firstVariable.DomainType), nextVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), nextVariable.ParamNumber, nextVariable.Module.Name, DomainTypeUtils.getName(nextVariable.DomainType)), errors);
                    else
                        AnalysisUtils.pushError(expr, String.Format("Variable '{0}' is defined by two or more primitive types - {1} in {2} (param no. {3}) from module '{4}' and {5} in {6} (param no. {7}) from module '{8}'", firstVariable.VariableName, DomainTypeUtils.getName(firstVariable.DomainType), firstVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), firstVariable.ParamNumber, firstVariable.Module.Name, DomainTypeUtils.getName(nextVariable.DomainType), nextVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), nextVariable.ParamNumber, nextVariable.Module.Name), errors);
                }
                hasLocalModuleDomainRef = hasLocalModuleDomainRef || !nextVariable.Module.IsInternalExternal;
            }

            if (!hasLocalModuleDomainRef)
            {
                AnalysisUtils.pushError(expr, String.Format("Variable '{0}' is defined by proper types, but the variable's context refers to internal/external modules only that do not hold domain values and thus no value will be generated for them (sample: {1} in {2} (param no. {3}) from module '{4}')", firstVariable.VariableName, DomainTypeUtils.getName(firstVariable.DomainType), firstVariable.Literal.Accept(new ExprPrinter("", false)).ToString(), firstVariable.ParamNumber, firstVariable.Module.Name), errors);
            }
        }

        /// <summary>
        /// Adds variables from the literal to current environment
        /// </summary>
        /// <param name="literal"></param>
        private void pushVariables(Literal literal)
        {
            var variables = ExprVariableGatherer.findVariablesOfLiteral(CurrentModule, literal);
            pushVariables(variables);
        }

        private void pushVariables(IDictionary<string, ICollection<VariableRef>> variables)
        {
            foreach (var variableRefs in variables)
            {
                if (Variables.ContainsKey(variableRefs.Key))
                {
                    var savedVariables = Variables[variableRefs.Key];
                    foreach (var variable in variableRefs.Value)
                    {
                        savedVariables.Add(variable);
                    }
                }
                else
                {
                    Variables.Add(variableRefs.Key, variableRefs.Value);
                }
            }
        }

        #region Visitor<object> Members

        public object visitLiteral(Literal literal)
        {
            if (CurrentModule == null)
            {
                AnalysisUtils.pushError(literal, String.Format("Module in which the expression '{0}' would be executed is non-existent or ambiguous.", literal.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else
            {
                int previousErrors = Errors != null ? Errors.Count : 0;
                LiteralAnalysis.checkLiteral(CurrentModule, literal, Errors, true);
                if (Errors == null || Errors.Count == previousErrors)
                {
                    pushVariables(literal);
                }
            }

            return null;
        }

        public object visitNeg(NegExpression negExpr)
        {
            return negExpr.Next.Accept(this);
        }

        public object visitModule(ModuleExpression moduleExpr)
        {
            if (DuplicatedModules.Contains(moduleExpr.ModuleName))
            {
                AnalysisUtils.pushError(moduleExpr, String.Format("Module '{0}' used in expression '{1}' is ambiguous.", moduleExpr.ModuleName, moduleExpr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else if (!Modules.ContainsKey(moduleExpr.ModuleName))
            {
                AnalysisUtils.pushError(moduleExpr, String.Format("Module '{0}' used in expression '{1}' is non-existent.", moduleExpr.ModuleName, moduleExpr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else if (StartingModule != null && moduleExpr.ModuleName.Equals(StartingModule.Name))
            {
               // AnalysisUtils.pushError(moduleExpr, String.Format("Module '{0}' used in expression '{1}' is also a rule module and thus it cannot be used in the rule.", moduleExpr.ModuleName, moduleExpr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }

            var module = CurrentModule;
            CurrentModule = Modules.ContainsKey(moduleExpr.ModuleName) ? Modules[moduleExpr.ModuleName] : null;

            moduleExpr.Next.Accept(this);

            CurrentModule = module;
            return null;
        }

        public object visitAnd(AndExpression andExpr)
        {
            foreach (var next in andExpr.Nexts)
            {
                next.Accept(this);
            }
            return null;
        }

        public object visitOr(OrExpression orExpr)
        {
            foreach (var next in orExpr.Nexts)
            {
                next.Accept(this);
            }
            return null;
        }

        private void visitQuantifier(QantifierExpression expr)
        {
            var variablesCache = Variables;

            // create local variable environment
            Variables = new Dictionary<string, ICollection<VariableRef>>();

            Variables.Add(expr.VariableName, new List<VariableRef>());
            // get the domain module

            var mod = CurrentModule;

            var moduleName = expr.DomainModule;
            if (moduleName == null && CurrentModule == null)
                AnalysisUtils.pushError(expr,
                    String.Format("Module of domain {0} used in expression '{1}' is undefined.", expr.VariableDomain,
                        expr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            else
            {

                moduleName = moduleName ?? CurrentModule.Name;

            }

            if (moduleName != null && DuplicatedModules.Contains(moduleName))
            {
                AnalysisUtils.pushError(expr, String.Format("Module '{0}' of domain {1} used in expression '{2}' is ambiguous.", moduleName, expr.VariableDomain, expr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else if (moduleName != null && !Modules.ContainsKey(moduleName))
            {
                AnalysisUtils.pushError(expr, String.Format("Module '{0}' of domain {1} used in expression '{2}' is non-existent.", moduleName, expr.VariableDomain, expr.Accept(new ExprPrinter("", false)).ToString()), Errors);
            }
            else
            {
                mod = Modules[moduleName];

                // check if quantifier uses a primitive domain
                DomainType t;
                if (DomainType.TryParse(expr.VariableDomain, true, out t))
                {
                    // push new variable ref for variable under the quantifier
                    Variables[expr.VariableName].Add(new VariableRef() { VariableName = expr.VariableName, Module = mod, DomainName = expr.VariableDomain, DomainType = t, Literal = null, ParamNumber = 0 });

                }
                else // it must be a defined type
                {

                    // check if domain exists
                    Domain dom = mod.FindDomain(expr.VariableDomain);
                    if (dom == null)
                        AnalysisUtils.pushError(expr, String.Format("Domain '{0}' used in expression '{1}' is non-existent in module {2}.", expr.VariableDomain, expr.Accept(new ExprPrinter("", false)).ToString(), CurrentModule.Name), Errors);
                    // push new variable ref for variable under the quantifier
                    Variables[expr.VariableName].Add(new VariableRef() { VariableName = expr.VariableName, DomainName = expr.VariableDomain, DomainType = dom == null? DomainType.VARIABLE : dom.Type, Literal = null, Module = mod, ParamNumber = 0 });

                }
            }

            // analyze the expression under the quantifier
            expr.Next.Accept(this);

            checkVariableTypes(Variables[expr.VariableName], expr, Errors);

            // remove the variable, as it should be invisible outside the scope of this quantifier
            Variables.Remove(expr.VariableName);

            var variablesCache2 = Variables;
            Variables = variablesCache;

            //push local variables to the global environment
            pushVariables(variablesCache2);
            
           
        }


        public object visitExists(ExistsExpression exExpr)
        {
            visitQuantifier(exExpr);
            return null;
        }

        public object visitAll(AllExpression allExpr)
        {
            visitQuantifier(allExpr);
            return null;
        }

        #endregion
    }
}
