﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    class LiteralAnalysis
    {
        public static void checkLiteral(Module module, Literal literal, ICollection<InvalidDefinition> errors, bool allowVariables = false, bool canBeDomain=true)
        {
            // Find the relation definition for a literal.
            var relation = module.FindRelation(literal.RelationName);

            // Check facts - if they represent facts of some relations.
            if (relation == null)
            {
                AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - relation '{2}' does not exist in the module.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, literal.RelationName), errors);
                return;
            }
            // Check if domain literals are allowed
            if (module.DomainRelations.Contains(relation) && !canBeDomain)
            {
                AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - cannot use domain '{2}' as a relation in this context.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, literal.RelationName), errors);
                return;
            }

            // Check if the definition of the relation is not broken in any sense.
            if (module.InvalidDefinitions.DuplicatedRelations.ContainsKey(literal.RelationName))
            {
                // The relation is a duplicate - avoid checking it.
                AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - relation '{2}' is defined more than once in the module and due to the errors cannot be used here.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, literal.RelationName), errors);
                return;
            }
            else if (module.InvalidDefinitions.InvalidRelations.ContainsKey(relation))
            {
                // The relation's definition was invalid - avoid checking it.
                AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - relation '{2}' was detected as invalid and due to that it cannot be used here.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, literal.RelationName), errors);
                return;
            }

            // Check literal - if their parameter counts agree with respective relation parameter counts.
            if (relation.paramCount() != literal.Values.Count)
            {
                AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - base relation '{2}' has {3} arguments, but the literal has {4}.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, literal.RelationName, relation.paramCount(), literal.Values.Count), errors);
                return;
            }

            var variablePrimitiveDomains = new Dictionary<string, DomainType>();
            var variableDefinedDomains = new Dictionary<string, string>();
            
            for (int i = 0; i < literal.Values.Count; ++i)
            {
                // Check facts - if their parameters are not variables.
                var paramValue = literal.Values.ElementAt<ValueHolder>(i);
                if (paramValue.Type == DomainType.VARIABLE)
                {
                    if (!allowVariables)
                    {
                        // Variables are not allowed - throw an exception.
                        AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - facts can only have constant values as parameters while parameter no. {2} in the definition is a variable '{3}'.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, i, paramValue.valueAs<string>()), errors);
                        continue;
                    }

                    // Grab the variable name and types.
                    var variable = paramValue.valueAs<string>();
                    var definedType = relation.DefinedDomains.ContainsKey(i) ? relation.DefinedDomains[i] : null;
                    var primitiveType = definedType == null ? relation.PrimitiveDomains[i] : module.FindDomain(definedType).Type;
                    
                    if (variablePrimitiveDomains.ContainsKey(variable))
                    {
                        // We already met the variable in the literal - we have to check if its "new type" is identical to its previous type.
                        var previousPrimitiveType = variablePrimitiveDomains[variable];
                        var previousDefinedType = variableDefinedDomains.ContainsKey(variable) ? variableDefinedDomains[variable] : null;

                        if (primitiveType != previousPrimitiveType)
                        {
                            // Variable encounters do not match on primitive types - throw an exception.
                            AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - fact uses two different primitive types (previously {2}, now {3}) for variable '{4}'.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, DomainTypeUtils.getName(previousPrimitiveType), DomainTypeUtils.getName(primitiveType), variable), errors);
                        }
                        else if (definedType != null)
                        {
                            if (module.InvalidDefinitions.InvalidDomains.ContainsKey(definedType))
                            {
                                // The domain has an error - avoid checking it.
                                AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - domain '{2}' due to the errors cannot be used here.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, definedType), errors);
                                return;
                            }
                            if (false)
                            {
                                // Disabled: Inter4ql only cares about primitive types and thus the defined types do not have to be checked.
                                if (previousDefinedType == null)
                                {
                                    // Previously it was only a primitive type - just add the defined type for the variable.
                                    variableDefinedDomains.Add(variable, definedType);
                                }
                                else if (!previousDefinedType.Equals(definedType))
                                {
                                    // Variable encounters do not match on defined types - throw an exception.
                                    AnalysisUtils.pushError(literal,
                                        String.Format(
                                            "Invalid definition of literal '{0}' in module '{1}' - fact uses two different defined types (previously {2}, now {3}) for variable '{4}'.",
                                            new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name,
                                            previousDefinedType, definedType, variable), errors);
                                }
                            }
                        }
                    }
                    else
                    {
                        // First encounter of the variable - remember it.
                        variablePrimitiveDomains.Add(variable, primitiveType);

                        if (definedType != null)
                        {
                            variableDefinedDomains.Add(variable, definedType);
                        }
                    }
                }
                else if (relation.PrimitiveDomains.ContainsKey(i))
                {
                    // Check facts - if their values match primitive parameters from relation definitions.
                    var relationParamType = relation.PrimitiveDomains[i];
                    if (paramValue.Type != relationParamType)
                    {
                        AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - in the relation definition param no. {2} (Value='{3}') is of type '{4}' while in the literal it is '{5}'.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, i, ValuePrinter.ToValueString(paramValue), DomainTypeUtils.getName(relationParamType), DomainTypeUtils.getName(paramValue.Type)), errors);
                    }
                }
                else
                {
                    // Check facts - if their values match defined domain parameters from relation definitions.
                    var relationParam = relation.DefinedDomains[i];
                    var relationParamType = module.FindDomain(relationParam).Type;
                    if (paramValue.Type != relationParamType)
                    {
                        AnalysisUtils.pushError(literal, String.Format("Invalid definition of literal '{0}' in module '{1}' - in the relation definition param no. {2} (Value='{3}') is a part of domain '{4}' (type '{5}') while in the literal it is '{6}'.", new ProgramPrinter().appendLiteralNoEnd(literal).Done(), module.Name, i, ValuePrinter.ToValueString(paramValue), relationParam, DomainTypeUtils.getName(relationParamType), DomainTypeUtils.getName(paramValue.Type)), errors);
                    }
                }
            }
        }
    }
}
