﻿using FourqlParser.Structures;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    /// <summary>
    /// Finds module refences in an expression.
    /// The GetReferences method returns a dictionary of referenced module names and relations within those modules.
    /// 
    /// </summary>
    public class ModuleReferenceFinder : Expression.Visitor<IDictionary<String, ICollection<string>>>
    {
        public string CurrentModule { get; set; }
        public string ExecModule { get; set; }
        public IDictionary<string, ICollection<string>> ModuleReferences { get; set; }

        public ModuleReferenceFinder()
        {
            ModuleReferences = new Dictionary<string, ICollection<string>>();
        }

        public IDictionary<string, ICollection<string>> GetReferences(string execModule, Expression expr)
        {
            ExecModule = execModule ?? "";
            CurrentModule = ExecModule;
            ModuleReferences = new Dictionary<string, ICollection<string>>();
            return expr.Accept(this);
        }

        #region Visitor<IDictionary<String, ICollection<string>>> Members

        public IDictionary<String, ICollection<string>> visitLiteral(Literal literal)
        {
            if (!CurrentModule.Equals(ExecModule) && ModuleReferences.ContainsKey(CurrentModule))
            {
                ModuleReferences[CurrentModule].Add(literal.RelationName);
            }
            return ModuleReferences;
        }

        public IDictionary<String, ICollection<string>> visitNeg(NegExpression negExpr)
        {
            return negExpr.Next.Accept(this);
        }

        public IDictionary<String, ICollection<string>> visitModule(ModuleExpression moduleExpr)
        {
            if (!moduleExpr.ModuleName.Equals(ExecModule) && !ModuleReferences.ContainsKey(moduleExpr.ModuleName))
            {
                ModuleReferences.Add(moduleExpr.ModuleName, new HashSet<string>());
            }
            var currModule = CurrentModule;
            CurrentModule = moduleExpr.ModuleName;
            var retValue = moduleExpr.Next.Accept(this);
            CurrentModule = currModule;
            return retValue;
        }

        public IDictionary<String, ICollection<string>> visitAnd(AndExpression andExpr)
        {
            foreach (var next in andExpr.Nexts) {
                next.Accept(this);
            }
            return ModuleReferences;
        }

        public IDictionary<String, ICollection<string>> visitOr(OrExpression orExpr)
        {
            foreach (var next in orExpr.Nexts)
            {
                next.Accept(this);
            }
            return ModuleReferences;
        }




        public IDictionary<string, ICollection<string>> visitExists(ExistsExpression exExpr)
        {
            if (exExpr.DomainModule != null)
            {
                if (!exExpr.DomainModule.Equals(CurrentModule) && !ModuleReferences.ContainsKey(exExpr.DomainModule))
                {
                    ModuleReferences.Add(exExpr.DomainModule, new HashSet<string>() {exExpr.VariableDomain});
                }
            }
            return exExpr.Next.Accept(this);
        }

        public IDictionary<string, ICollection<string>> visitAll(AllExpression allExpr)
        {
            if (allExpr.DomainModule != null)
            {
                if (!allExpr.DomainModule.Equals(CurrentModule) && !ModuleReferences.ContainsKey(allExpr.DomainModule))
                {
                    ModuleReferences.Add(allExpr.DomainModule, new HashSet<string>() { allExpr.VariableDomain });
                }
            }
            return allExpr.Next.Accept(this);

        }

        #endregion
    }
}
