﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    public class ProgramAnalysisGatherErrors
    {

        //TODO ADD IMPORTED DOMAINS CHCECK
        //


        #region Invalid Domain, Relation and Fact checks

        public void checkForInvalidDefinitions(Module module)
        {
            checkForInvalidDomains(module);
            checkForInvalidRelations(module);
            checkForInvalidFacts(module);
        }


        public void checkDomainImports(Module module, Dictionary<string, Module> moduleMap)
        {
            foreach (var domain in module.Domains)
            {
                // Check if domain was imported from different module
                if (domain.Imports != null)
                {
                    foreach (var import in domain.Imports)
                    {
                        String origin = import.Item1;
                        // check if original module exists
                        if (!moduleMap.ContainsKey(origin))
                            InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidDomains, domain.Name, new InvalidDefinition() { ErrorSource = domain, Message = String.Format("Module '{0}' from definition of domain '{1}' in module '{2}' doesn't exist.", origin, domain.Name, module.Name) });

                        else if (origin.Equals(module.Name))
                            InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidDomains, domain.Name, new InvalidDefinition() { ErrorSource = domain, Message = String.Format("Error in definition of domain '{0}' in module '{1}' : Cannot import values from domain in the same module!", origin, module.Name) });
                        else 
                        {
                            // check if original domain exists
                            Domain source = moduleMap[origin].FindDomain(import.Item2);
                            if (source == null)
                                InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidDomains, domain.Name, new InvalidDefinition() { ErrorSource = domain, Message = String.Format("Imported domain '{0}' doesn't exist in module '{1}'.", import.Item2, origin) });
                            else
                                if (!domain.Type.Equals(source.Type))
                                    InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidDomains, domain.Name, new InvalidDefinition() { ErrorSource = domain, Message = String.Format("Type mismatch: imported domain '{0}.{1}' has type {2}, expected {3}.", import.Item1, import.Item2, source.Type, domain.Type) });
                        }
                    }
                }
            }
            
        }

        public void checkForInvalidDomains(Module module)
        {
            var foundDomains = new Dictionary<string, Domain>();
            foreach (var domain in module.Domains)
            {
                // Check domains - if their names are unique within their module.
                if (foundDomains.ContainsKey(domain.Name))
                {
                    // We found a duplicate - add it.
                    InvalidDefinitionUtils.addErrorsWithDesc(module.InvalidDefinitions.DuplicatedDomains, domain.Name, foundDomains[domain.Name], domain);
                    //throw new Exception(String.Format("Duplicated definition of domain '{0}' in module '{1}'.", domain.Name, module.Name));
                }
                else
                {
                    foundDomains.Add(domain.Name, domain);
                }
            }
        }

        public void checkForInvalidRelations(Module module)
        {
            var domainNames = new HashSet<string>();
            foreach (var domain in module.Domains)
            {
                domainNames.Add(domain.Name);
            }

            var foundRelations = new Dictionary<string, Relation>();
            foreach (var relation in module.Relations)
            {
                // Check relations - if their names are unique within their module.
                if (foundRelations.ContainsKey(relation.Name))
                {
                    // We found a duplicate - add it.
                    InvalidDefinitionUtils.addErrorsWithDesc(module.InvalidDefinitions.DuplicatedRelations, relation.Name, foundRelations[relation.Name], relation);
                    //throw new Exception(String.Format("Duplicated definition of relation '{0}' in module '{1}'.", relation.Name, module.Name));
                }
                else if (domainNames.Contains(relation.Name))
                {
                    InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidRelations, relation, new InvalidDefinition() { ErrorSource = relation, Message = String.Format("Invalid definition of relation '{0}' in module '{1}' - Domain with the same name exists in the module.", relation.Name, module.Name) });
                }
                else
                {
                    // Add the relation to the dictionary of found relations.
                    foundRelations.Add(relation.Name, relation);
                    // Check relations in terms of their parameters that are of defined types.
                    foreach (var definedArg in relation.DefinedDomains)
                    {
                        // Check if the argument has a valid defined domain.
                        if (!domainNames.Contains(definedArg.Value))
                        {
                            // We found a relation that has an invalid defined domain.
                            InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidRelations, relation, new InvalidDefinition() { ErrorSource = relation, Message = String.Format("Invalid definition of relation '{0}' in module '{1}' - domain '{2}' (argument no. {3}) does not exist in the module.", relation.Name, module.Name, definedArg.Value, definedArg.Key) });
                            //throw new Exception(String.Format("Invalid definition of relation '{0}' in module '{1}' - domain '{2}' does not exist in the module.", relation.Name, module.Name, domainName));
                        }
                        else if (module.InvalidDefinitions.DuplicatedDomains.ContainsKey(definedArg.Value))
                        {
                            // This relation uses a defined domain that has a duplicated definition.
                            InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidRelations, relation, new InvalidDefinition() { ErrorSource = relation, Message = String.Format("Invalid definition of relation '{0}' in module '{1}' - domain '{2}' (argument no. {3}) was considered a duplicate and thus cannot be used here until fixed.", relation.Name, module.Name, definedArg.Value, definedArg.Key) });
                        }
                        else if (module.InvalidDefinitions.InvalidDomains.ContainsKey(definedArg.Value))
                        {
                            InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidRelations, relation, new InvalidDefinition() { ErrorSource = relation, Message = String.Format("Invalid definition of relation '{0}' in module '{1}' - domain '{2}' (argument no. {3}) has errors and thus cannot be used here until fixed.", relation.Name, module.Name, definedArg.Value, definedArg.Key) });
                        }
                    }
                }
            }
        }

        public void checkForInvalidFacts(Module module)
        {
            foreach (var fact in module.Facts)
            {
                var errors = new List<InvalidDefinition>();
                LiteralAnalysis.checkLiteral(module, fact, errors);

                if (errors.Count > 0)
                {
                    InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidFacts, fact, errors);
                }
            }
        }

        #endregion

        #region Invalid Expressions check

        public void checkExpressions(IDictionary<string, Module> moduleMap, ICollection<string> duplicatedModules)
        {
            foreach (var module in moduleMap.Values)
            {
                foreach (var rule in module.Rules)
                {
                    var errors = new List<InvalidDefinition>();
                    new RuleAnalysis(module, moduleMap, duplicatedModules).run(rule, errors);

                    if (errors.Count > 0)
                    {
                        InvalidDefinitionUtils.addErrors(module.InvalidDefinitions.InvalidRules, rule, errors);
                    }
                }
            }
        }

        #endregion

        #region Invalid Module Hierarchy check

        private void checkModuleCycles(IDictionary<string, Module> moduleMap)
        {
            var stronglyConnectedComponents = new TarjansAlgorithm().run(moduleMap);
            var builder = new StringBuilder("Found cycles (Strongly Connected Components) that consist of modules:\n");
            var cycleNo = 0;

            foreach (var component in stronglyConnectedComponents)
            {
                if (component.Count > 1)
                {
                    builder.Append(String.Format("\t{0}. ", cycleNo++));
                    var first = true;

                    foreach (var item in component)
                    {
                        if (first)
                        {
                            first = false;
                        }
                        else
                        {
                            builder.Append(", ");
                        }

                        builder.Append(item.Name);
                    }
                }
            }

            if (cycleNo > 0)
            {
                throw new Exception(builder.ToString());
            }
        }

        #endregion
    }
}
