﻿using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    class RuleAnalysis
    {
        private Module StartingModule = null;
        private Module CurrentModule = null;
        private IDictionary<string, Module> Modules = null;
        private Rule Rule = null;
        private ICollection<InvalidDefinition> Errors = null;
        private IDictionary<string, ICollection<VariableRef>> CurrentVariables = null;
        private ICollection<IDictionary<string, ICollection<VariableRef>>> Variables = new List<IDictionary<string, ICollection<VariableRef>>>();
        private ICollection<string> DuplicatedModules = null;

        public RuleAnalysis(Module startingModule, IDictionary<string, Module> modules)
            : this(startingModule, modules, new List<string>())
        {
        }

        public RuleAnalysis(Module startingModule, IDictionary<string, Module> modules, ICollection<string> duplicatedModules)
        {
            this.StartingModule = startingModule;
            this.Modules = modules;
            this.DuplicatedModules = duplicatedModules;
        }

        public void run(Rule rule, ICollection<InvalidDefinition> errors = null)
        {
            Errors = errors;
            Rule = rule;
            CurrentModule = StartingModule;

            try
            {
                // Remember the number of errors we have in the error list.
                int errorNo = errors != null ? errors.Count : 0;
                // Check the literal.
                LiteralAnalysis.checkLiteral(StartingModule, rule.Head, errors, true, false);
                // See if our error number changed.
                if (errors != null && errors.Count != errorNo)
                {
                    // It did - ignore checking the body.
                    return;
                }

                try
                {
                    // Remember the head variables.
                    var headVariables = ExprVariableGatherer.findVariablesOfLiteral(StartingModule, rule.Head);


                    // create new expression analyzer object
                    var exprAnalysis = new ExpressionAnalysis(StartingModule, Modules);
                    // Analyze the expression and grab its variables.
                    var bodyVariables = exprAnalysis.run(rule.Body, errors, true);
                    // Check the head variables against the quantified variables -> they cannot have the same name
                    checkBoundVariableNames(headVariables, ExprVariableGatherer.findBoundVariables(StartingModule, rule.Body, Modules), errors);
                
                    // Check if variables from the head appear in the body
                    checkRuleVariableExistence(headVariables, bodyVariables, errors);
                   }
                catch (Exception ex)
                {
                    throw new Exception(String.Format("Error while checking the body of expression: {0}", rule.Body.Accept(new ExprPrinter("", false))).ToString(), ex);
                }
            }
            catch (Exception ex)
            {
                throw new Exception(String.Format("Error while checking rule: {0} of module '{1}'", new ProgramPrinter().append(rule, "", false).Done(), StartingModule.Name), ex);
            }
        }


        private void checkBoundVariableNames(IDictionary<string, ICollection<VariableRef>> headVariables, IDictionary<string, ICollection<VariableRef>> boundVariables, ICollection<InvalidDefinition> errors = null)
        {
            foreach (var boundVar in boundVariables.Values)
            {
                var variable = boundVar.ElementAt(0);
                if (headVariables.ContainsKey(variable.VariableName))
                {
                    AnalysisUtils.pushError(Rule.Body, String.Format("Bound variable '{0}' from rule '{1}' cannot appear in the head of the rule! ", variable.VariableName, new ProgramPrinter(){_module = StartingModule}.append(Rule, "", false).Done()), errors);
                }
                if (variable.Module == null)
                {
                    AnalysisUtils.pushError(Rule.Body, String.Format("Domain of bound variable '{0}' from rule '{1}' comes from a non-existant module! ", variable.VariableName, new ProgramPrinter() { _module = StartingModule }.append(Rule, "", false).Done()), errors);
                }
            } 
        }

        private void checkRuleVariableExistence(IDictionary<string, ICollection<VariableRef>> headVariables, IDictionary<string, ICollection<VariableRef>> exprVariables, ICollection<InvalidDefinition> errors = null)
        {
            foreach (var variableList in headVariables.Values)
            {
                var variable = variableList.ElementAt(0);
                if (!exprVariables.ContainsKey(variable.VariableName))
                {
                    AnalysisUtils.pushError(Rule.Body, String.Format("There is no variable '{0}' in the body of the rule: {1}", variable.VariableName, new ProgramPrinter() { _module = StartingModule }.append(Rule, "", false).Done()), errors);
                }
                else
                {
                    try
                    {
                        checkVariable(variable, exprVariables[variable.VariableName].ElementAt(0));
                    }
                    catch (Exception ex)
                    {
                        if (errors == null)
                        {
                            throw new Exception(String.Format("Definitions of variable '{0}' in the head and body of the rule: {1} differ", variable.VariableName, new ProgramPrinter() { _module = StartingModule }.append(Rule, "", false).Done()), ex);
                        }

                        AnalysisUtils.pushError(Rule, String.Format("Definitions of variable '{0}' in the head and body of the rule: {1} differ: {2}", variable.VariableName, new ProgramPrinter() { _module = StartingModule }.append(Rule, "", false).Done(), ex.Message), errors);

                    }
                }
            }
        }

        private void checkVariable(VariableRef first, VariableRef second)
        {
            if (first.DomainType != second.DomainType)
            {
                throw new Exception(String.Format("Variable '{0}' is defined by two or more primitive types - {1} in {2} (param no. {3}) from module '{4}' and {5} in {6} (param no. {7}) from module '{8}'", first.VariableName, DomainTypeUtils.getName(first.DomainType), first.Literal.Accept(new ExprPrinter("", false)).ToString(), first.ParamNumber, first.Module.Name, DomainTypeUtils.getName(second.DomainType), second.Literal.Accept(new ExprPrinter("", false)).ToString(), second.ParamNumber, second.Module.Name));
            }
        }
    }


}
