﻿using FourqlParser.Structures;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.StaticAnalysis
{
    public class TarjansAlgorithm
    {
        private ICollection<ICollection<Module>> StronglyConnectedComponents = new List<ICollection<Module>>();
        private ICollection<string> VerticesOnStack = new HashSet<string>();
        private Stack<Vertex> VertexStack = new Stack<Vertex>();
        private int Index = 0;

        public ICollection<ICollection<Module>> run(IDictionary<string, Module> moduleMap)
        {
            var vertices = new List<Vertex>();
            var moduleToVertex = new Dictionary<string, Vertex>();

            // Create a vertex for every module.
            foreach (var module in moduleMap.Values)
            {
                var vertex = new Vertex() { Module = module };
                moduleToVertex[module.Name] = vertex;
                vertices.Add(vertex);
            }

            foreach (var module in moduleMap.Values)
            {
                // Calculate references from modules to modules as edges.
                var edges = new HashSet<string>();
                foreach (var rule in module.Rules)
                {
                    var ruleReferences = new ModuleReferenceFinder().GetReferences(module.Name, rule.Body);
                    foreach (var edge in ruleReferences.Keys)
                    {
                        edges.Add(edge);
                    }
                }
                // Calculate references created by importing domains
                foreach (var domain in module.Domains)
                {
                    if (domain.Imports != null)
                        foreach (var import in domain.Imports)
                        {
                            edges.Add(import.Item1);
                        }
                }

                // Remove self-reference.
                edges.Remove(module.Name);

                // Add references as edges to respective vertices.
                var vertex = moduleToVertex[module.Name];
                foreach (var edge in edges)
                {
                    if (moduleToVertex.ContainsKey(edge))
                    {
                        vertex.EdgesTo.Add(moduleToVertex[edge]);
                    }
                }
            }

            // Traverse the graph to find strongly connected components.
            foreach (var vertex in vertices)
            {
                if (!vertex.Index.HasValue)
                {
                    runFor(vertex);
                }
            }

            return StronglyConnectedComponents;
        }

        private void runFor(Vertex vertex)
        {
            vertex.Index = Index;
            vertex.LowLink = Index++;
            VertexStack.Push(vertex);
            VerticesOnStack.Add(vertex.Module.Name);

            foreach (var next in vertex.EdgesTo)
            {
                if (!next.Index.HasValue)
                {
                    runFor(next);
                    vertex.LowLink = Math.Min(vertex.LowLink, next.LowLink);
                }
                else if (VerticesOnStack.Contains(next.Module.Name))
                {
                    vertex.LowLink = Math.Min(vertex.LowLink, next.Index.Value);
                }
            }

            if (vertex.LowLink == vertex.Index.Value)
            {
                var component = new List<Module>();
                Vertex componentVertex;
                do
                {
                    componentVertex = VertexStack.Pop();
                    component.Add(componentVertex.Module);
                    VerticesOnStack.Remove(componentVertex.Module.Name);
                } while (vertex != componentVertex);

                StronglyConnectedComponents.Add(component);
            }
        }
    }

    class Vertex
    {
        public Module Module { get; set; }
        public int? Index { get; set; }
        public int LowLink { get; set; }
        public ICollection<Vertex> EdgesTo { get; set; }

        public Vertex()
        {
            LowLink = Int32.MaxValue;
            EdgesTo = new List<Vertex>();
        }
    }
}
