﻿using Antlr4.Runtime;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Structures.Utils;
using FourqlParser.Structures.Print;

namespace FourqlParser.Structures
{
    public abstract class Expression : OutputSource
    {
        public interface Visitor<R>
        {
            R visitLiteral(Literal literal);
            R visitNeg(NegExpression negExpr);
            R visitModule(ModuleExpression moduleExpr);
            R visitAnd(AndExpression andExpr);
            R visitOr(OrExpression orExpr);

            R visitExists(ExistsExpression exExpr);

            R visitAll(AllExpression allExpr);
        }

        public abstract class AbstractVisitor<R> : Visitor<R>
        {
            #region Visitor<R> Members

            public virtual R visitLiteral(Literal literal)
            {
                return defaultValue();
            }

            public virtual R visitNeg(NegExpression negExpr)
            {
                return defaultValue();
            }

            public virtual R visitModule(ModuleExpression moduleExpr)
            {
                return defaultValue();
            }

            public virtual R visitAnd(AndExpression andExpr)
            {
                return defaultValue();
            }

            public virtual R visitOr(OrExpression orExpr)
            {
                return defaultValue();
            }

            public virtual R visitExists(ExistsExpression exExpr)
            {
                return defaultValue();
            }

            public virtual R visitAll(AllExpression allExpr)
            {
                return defaultValue();
            }

            #endregion

            protected abstract R defaultValue();
        }
        
        public abstract R Accept<R>(Visitor<R> visitor);
    }

    public class Literal : Expression
    {
        public String RelationName { get; set; }
        public IList<ValueHolder> Values { get; set; }

        public Literal()
        {
            Values = new List<ValueHolder>();
        }

        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitLiteral(this);
        }

        public override bool Equals(object obj)
        {
            var that = obj as Literal;
            if (that == null)
            {
                return false;
            }

            if (this == that)
            {
                return true;
            }

            if (!ComparatorUtils.BothNullsOrNot(this.RelationName, that.RelationName))
            {
                return false;
            }

            if (this.RelationName != null && !this.RelationName.Equals(that.RelationName))
            {
                return false;
            }

            if (this.Values.Count != that.Values.Count)
            {
                return false;
            }

            var thisIter = this.Values.GetEnumerator();
            var thatIter = that.Values.GetEnumerator();
            while (thisIter.MoveNext() && thatIter.MoveNext())
            {
                if (!thisIter.Current.Equals(thatIter.Current))
                {
                    return false;
                }
            }

            return true;
        }

        public override int GetHashCode()
        {
            return new ExprPrinter("", false).ConvertExpr(this).ToString().GetHashCode();
        }

        public int UniqueHash()
        {
            return base.GetHashCode();
        }
    }

    public abstract class UnaryExpression : Expression
    {
        public Expression Next { get; set; }
    }

    public class NegExpression : UnaryExpression
    {
        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitNeg(this);
        }
    }

    public class ModuleExpression : UnaryExpression
    {
        public String ModuleName { get; set; }
        public Boolean IsNegative { get; set; }
        public ICollection<LogicValue> InValues { get; set; }

        public ModuleExpression()
        {
            InValues = new List<LogicValue>();
        }

        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitModule(this);
        }

    }

    public abstract class QantifierExpression : UnaryExpression
    {
        public String VariableName { get; set; }

        public String VariableDomain { get; set; }

        public String DomainModule { get; set; }

    }

    public class ExistsExpression : QantifierExpression
    {
        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitExists(this);
        }
    }

    public class AllExpression : QantifierExpression
    {
        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitAll(this);
        }
    }
    
    public abstract class MultiExpression : Expression
    {
        public ICollection<Expression> Nexts { get; set; }

        public MultiExpression()
        {
            Nexts = new List<Expression>();
        }
    }

    

    public class AndExpression : MultiExpression
    {
        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitAnd(this);
        }
    }

    public class OrExpression : MultiExpression
    {
        public override R Accept<R>(Visitor<R> visitor)
        {
            return visitor.visitOr(this);
        }
    }

    public class ValueHolder
    {
        public Object Value { get; set; }

        public DomainType Type { get; set; }

        public ValueHolder(Object value, DomainType type)
        {
            Value = value;
            Type = type;
        }

        public T valueAs<T>()
        {
            return (T)Value;
        }

        public override bool Equals(object obj)
        {
            var that = obj as ValueHolder;
            if (that == null)
            {
                return false;
            }
            
            if (this == that)
            {
                return true;
            }

            if (this.Type != that.Type)
            {
                return false;
            }

            if (!ComparatorUtils.BothNullsOrNot(this.Value, that.Value))
            {
                return false;
            }

            if (this.Value != null && !this.Value.Equals(that.Value))
            {
                return false;
            }

            return true;
        }

        public override int GetHashCode()
        {
            unchecked
            {
                int hash = 17;
                hash = hash * 23 + Value.GetHashCode();
                hash = hash * 23 + Type.GetHashCode();
                return hash;
            }
        }

        public override string ToString()
        {
            return Value.ToString();
        }

    }
    
    public enum LogicValue : int
    {
        FALSE,
        UNKNOWN,
        INCONSISTENT,
        TRUE
    }

    public class ValueFactory
    {
        public static object getDefaultValue(DomainType type)
        {
            switch (type)
            {
                case DomainType.LOGIC:
                    return LogicValue.FALSE;
                case DomainType.LITERAL:
                    return "lit";
                case DomainType.STRING:
                    return "str";
                case DomainType.INTEGER:
                    return 0;
                case DomainType.REAL:
                    return 0.0f;
                case DomainType.DATE:
                case DomainType.DATE_TIME:
                    return DateTime.Now;
                default:
                    return null;
            }
        }
    }
}
