﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.Structures.Optimization
{
    /// <summary>
    /// Optimizes logical expression
    /// </summary>
    class ExpressionOptimizer : Expression.Visitor<Expression>
    {
        private bool OptimizeNeg { get; set; }

        /// <summary>
        /// </summary>
        /// <param name="optimizeNeg">should multiple negations be optimized</param>
        public ExpressionOptimizer(bool optimizeNeg = false)
        {
            OptimizeNeg = optimizeNeg;
        }

        public Expression optimize(Expression expr)
        {
            return expr.Accept(this);
        }

        #region Visitor<Expression> Members

        public Expression visitLiteral(Literal literal)
        {
            return literal;
        }

        public Expression visitNeg(NegExpression negExpr)
        {
            var next = negExpr.Next.Accept(this);
            if (OptimizeNeg)
            { // delete double negation
                if (next.Accept(new IsNegExpr()))
                {
                    return ((NegExpression)next).Next;
                }
            }
            return new NegExpression() { Next = next, ParserContext = negExpr.ParserContext };
        }

        public Expression visitModule(ModuleExpression moduleExpr)
        {
            return new ModuleExpression() { IsNegative = moduleExpr.IsNegative, ModuleName = moduleExpr.ModuleName, InValues = moduleExpr.InValues, Next = moduleExpr.Next.Accept(this), ParserContext = moduleExpr.ParserContext };
        }

        public Expression visitAnd(AndExpression andExpr)
        {
            var optimizedAnd = new AndExpression() { ParserContext = andExpr.ParserContext };
            foreach (var next in andExpr.Nexts)
            {
                var optimizedNext = next.Accept(this);
                if (optimizedNext.Accept(new IsAndExpr()))
                {
                    foreach (var innerItem in ((AndExpression) optimizedNext).Nexts)
                    {
                        optimizedAnd.Nexts.Add(innerItem);
                    }
                }
                else
                {
                    optimizedAnd.Nexts.Add(optimizedNext);
                }
            }
            return optimizedAnd.Nexts.Count == 1 ? optimizedAnd.Nexts.First<Expression>() : optimizedAnd;
        }

        public Expression visitOr(OrExpression orExpr)
        {
            var optimizedOr = new OrExpression() { ParserContext = orExpr.ParserContext };
            foreach (var next in orExpr.Nexts)
            {
                var optimizedNext = next.Accept(this);
                if (optimizedNext.Accept(new IsOrExpr()))
                {
                    foreach (var innerItem in ((OrExpression)optimizedNext).Nexts)
                    {
                        optimizedOr.Nexts.Add(innerItem);
                    }
                }
                else
                {
                    optimizedOr.Nexts.Add(optimizedNext);
                }
            }
            return optimizedOr.Nexts.Count == 1 ? optimizedOr.Nexts.First<Expression>() : optimizedOr;
        }


        public Expression visitExists(ExistsExpression exExpr)
        {
            var next = exExpr.Next.Accept(this);

            return new ExistsExpression() { DomainModule = exExpr.DomainModule, VariableDomain=exExpr.VariableDomain, VariableName=exExpr.VariableName, Next = next, ParserContext = exExpr.ParserContext };

        }

        public Expression visitAll(AllExpression allExpr)
        {
            var next = allExpr.Next.Accept(this);

            return new AllExpression() { DomainModule = allExpr.DomainModule, VariableDomain = allExpr.VariableDomain, VariableName = allExpr.VariableName, Next = next, ParserContext = allExpr.ParserContext };

        }
        #endregion
    }

    public class IsNegExpr : Expression.AbstractVisitor<bool>
    {
        public override bool visitNeg(NegExpression negExpr)
        {
            return true;
        }

        protected override bool defaultValue()
        {
            return false;
        }
    }

    public class IsAndExpr : Expression.AbstractVisitor<bool>
    {
        public override bool visitAnd(AndExpression andExpr)
        {
            return true;
        }

        protected override bool defaultValue()
        {
            return false;
        }
    }

    public class IsOrExpr : Expression.AbstractVisitor<bool>
    {
        public override bool visitOr(OrExpression orExpr)
        {
            return true;
        }

        protected override bool defaultValue()
        {
            return false;
        }
    }
}
