﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace FourqlParser.Structures.Print
{
    public class ExprPrinter : Printer, Expression.Visitor<StringBuilder>
    {
        private String Indent;
        private String OrSign = "OR";
        private String AndSign = "AND";
        private String NotSign = "NOT ";
        private String ExistsSign = "EXISTS ";
        private String AllSign = "ALL ";

        public string Module { get; set; }

        public ExprPrinter(String indent = "", Boolean printNewLines = true, String orSign = "OR", String andSign = "AND", String notSign = "NOT")
            : base(printNewLines)
        {
            Indent = indent;
            /*OrSign = orSign;
            AndSign = andSign;
            NotSign = notSign;*/
        }

        public StringBuilder ConvertExpr(Expression expr, bool appendSemicolon=true)
        {
            Append(Indent);
            expr.Accept(this);
            if (appendSemicolon)
                AppendLine(";");
            return Builder;
        }

        #region Visitor<StringBuilder> Members

        public StringBuilder visitLiteral(Literal literal)
        {
            Append(String.Format("{0}(", literal.RelationName));

            var first = true;
            foreach (var value in literal.Values) {
                if (first)
                {
                    first = false;
                }
                else
                {
                    Append(", ");
                }

                Append(ValuePrinter.ToValueString(value));
            }

            Append(")");
            return Builder;
        }

        public StringBuilder visitNeg(NegExpression negExpr)
        {
            Append(NotSign);
            appendWithBrackets(negExpr.Next, ExprType.Not);
            return Builder;
        }
        public StringBuilder visitExists(ExistsExpression exExpr)
        {
            Append(ExistsSign);
            if (exExpr.DomainModule == null || exExpr.DomainModule.Equals(Module))
                Append(String.Format("{0} : {1} ", exExpr.VariableName, exExpr.VariableDomain));
            else
            {
                Append(String.Format("{0} : {1}.{2} ", exExpr.VariableName, exExpr.DomainModule, exExpr.VariableDomain));
            }
            appendWithBrackets(exExpr.Next, ExprType.Quantifier);
            return Builder;
        }

        public StringBuilder visitAll(AllExpression allExpr)
        {
            Append(AllSign);
            if (allExpr.DomainModule == null || allExpr.DomainModule.Equals(Module))
                Append(String.Format("{0} : {1} ", allExpr.VariableName, allExpr.VariableDomain));
            else
            {
                Append(String.Format("{0} : {1}.{2} ", allExpr.VariableName, allExpr.DomainModule, allExpr.VariableDomain));
            } 
            appendWithBrackets(allExpr.Next, ExprType.Quantifier);
            return Builder;
        }
        
        public StringBuilder visitModule(ModuleExpression moduleExpr)
        {
            if (moduleExpr.IsNegative)
            {
                Builder.Append(NotSign);
            }
            var modName = Module;
            Module = moduleExpr.ModuleName;

            Append(moduleExpr.ModuleName);
            Append(".");
            appendWithBrackets(moduleExpr.Next, ExprType.Module);

            if (moduleExpr.InValues.Count > 0)
            {
                Append(" in { ");

                var first = true;
                foreach (var value in moduleExpr.InValues)
                {
                    if (first)
                    {
                        first = false;
                    }
                    else
                    {
                        Append(", ");
                    }

                    Append(value.ToString().ToLower());
                }

                Append(" }");
            }
            Module = modName;
            return Builder;
        }

        public StringBuilder visitAnd(AndExpression andExpr)
        {
            var first = true;
            foreach (var expr in andExpr.Nexts)
            {
                if (first)
                {
                    first = false;
                }
                else
                {
                    Append(String.Format(" {0} ", AndSign));
                }

                appendWithBrackets(expr, ExprType.And);
            }
            return Builder;
        }

        public StringBuilder visitOr(OrExpression orExpr)
        {
            var first = true;
            foreach (var expr in orExpr.Nexts) {
                if (first)
                {
                    first = false;
                    expr.Accept(this);
                }
                else
                {
                    AppendLine();
                    Append(String.Format("{0}{1} ", PrintNewLines ? Indent : "", OrSign));
                    Append(expr.Accept(new ExprPrinter(Indent + "  ", PrintNewLines, OrSign, AndSign, NotSign)).ToString());
                }
            }
            return Builder;
        }


        #endregion

        public StringBuilder appendWithBrackets(Expression expr, ExprType currExpr)
        {
            var requiresBrackets = expr.Accept(new ExprRequiresBrackets(currExpr));
            if (requiresBrackets)
            {
                Append("(");
            }

            expr.Accept(this);
            if (requiresBrackets)
            {
                Append(")");
            }

            return Builder;
        }
    }
}
