﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using FourqlParser.Parsers;
using FourqlParser.Structures;
using FourqlParser.Structures.Print;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace FourqlParserTests
{
    [TestClass]
    public class CorrectExprTests : BaseModuleTests
    {
        [TestMethod]
        public void SimpleLiteral()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "plany.spotkanie(D, M, G).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual(strExpr, expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleLiteralWithIns()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "plany.spotkanie(D, M, G) in { true, false, incons, inconsistent, unknown }.";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual("plany.spotkanie(D, M, G) in { true, false, inconsistent, inconsistent, unknown }.", expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleLiteralWithNegation()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "NOT plany.spotkanie(D, M, G).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual(strExpr, expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleLiteralWithInsAndNegation()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "NOT plany.spotkanie(D, M, G) in { true, false, incons, inconsistent, unknown }.";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual("NOT plany.spotkanie(D, M, G) in { true, false, inconsistent, inconsistent, unknown }.", expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleLiteralWithNegation2()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "plany.(NOT spotkanie(D, M, G)).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual(strExpr, expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleLiteralWithNegation3()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "NOT (plany.spotkanie(D, M, G)).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual("NOT plany.spotkanie(D, M, G).", expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleConjunction()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "plany.spotkanie(D, M, G) AND plany.spotkanie(D, M, G).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual(strExpr, expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleAlternative()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "plany.spotkanie(D, M, G) OR plany.spotkanie(D, M, G).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual(strExpr, expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleAlternativeAndConjunctionWithBrackets()
        {
            var errors = new ParserDispatcher.Errors();
            var progContext = ParserDispatcher.parseProgram(GetTestContent("kalendarz.4ql"), errors, null);
            var prog = ParserDispatcher.toProgram(progContext);

            var strExpr = "(plany.spotkanie(D, M, G) OR plany.spotkanie(D, M, G)) AND plany.spotkanie(D, M, G).";
            var exprContext = ParserDispatcher.parseExpr(strExpr, errors);
            var expr = ParserDispatcher.toExpr(exprContext);
            ParserDispatcher.checkExpr(expr, prog.Modules, null);

            Assert.AreEqual(strExpr, expr.Accept(new ExprPrinter("", false)).ToString() + ".");
        }

        [TestMethod]
        public void SimpleExistsExpression()
        {

        }
    }
}
